Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import DomType.
Require Import DomBasic.
Require Import DomMap.

Class abs A B `{setoid A, poset B} : Type := Abs {
  abs_relation : forall (a:A) (b:B), Prop;
  abs_relation_proper : Proper (eq ==> le ==> le) abs_relation
}.
Notation "a '\abs' b" := (abs_relation a b) (at level 70).

Add Parametric Morphism (A B : Type) `{setoid A, poset B, @abs A B _ _ _} : (@abs_relation A B _ _ _ _) with
  signature eq ==> le ==> le as abs_relation_mor.
Proof. exact: abs_relation_proper. Qed.

Add Parametric Morphism (A B : Type) `{setoid A, poset B, @abs A B _ _ _} : (@abs_relation A B _ _ _ _) with
  signature eq ==> eq ==> eq as abs_relation_eq_mor.
Proof. 
  move=> x y Hxy z w Hzw.
  apply: le_antisym=> Habs.
  eapply abs_relation_mor; eauto; exact/DomType.eq_le.
  eapply abs_relation_mor; eauto.
  exact/eq_sym.
  exact/DomType.eq_le/eq_sym.
Qed.


Section Options.
  Context {A B : Type} `{abs A B}.

  Global Instance option_abs : abs (option A) B := {
    abs_relation oa b :=
    if oa is Some a
      then a \abs b
      else True
  }.
  Proof.
    move=> [x|] [y|] // Hxy z w Hzw.
    rewrite/eq/= in Hxy.
    rewrite -> Hxy; rewrite -> Hzw.
    exact/DomType.eq_le/eq_refl.
  Defined.
End Options.

Section Lists.
  Context {A B : Type} `{abs A B}.

  Global Instance list_abs : @abs (list A) (list B) _ _ list_poset_pointwise := {
    abs_relation :=
    (fix rel (al : list A) (bl : list B) :=
      match al, bl with
        | nil, nil => True
        | a::al', b::bl' => a \abs b /\ (nosimpl rel al' bl')
        | _, _ => False
      end)
  }.
  Proof.
    elim=> [|x xs IH] [|y ys] Hxy // [|z zs] [|w ws] Hzw //.
    move: Hxy Hzw => [Hxy Hxys] [Hzw Hzws].
    split.
    rewrite <- Hxy; rewrite <- Hzw.
    by case: H3.
    eapply IH; eauto.
    by case: H3.
  Defined.
End Lists.

Section Sums.
  Context {A B C D : Type} `{abs A B, abs C D}.

  Global Instance sum_abs : @abs (A + C)%type (B * D)%type _ _ pair_poset_pointwise :=
  {
    abs_relation x y :=
      match x with
        | inl x' => x' \abs y.1
        | inr x' => x' \abs y.2
      end  
  }.
  Proof.
    move=> [x|y] [z|w] Hxyzw [a b] [c d] [Hac Hbd] Habs; rewrite/eq//= in Hxyzw.
    by rewrite <- Hxyzw; rewrite <- Hac.
    by rewrite <- Hxyzw; rewrite <- Hbd.
  Defined.
End Sums.


Section Pairs.
  Context {A B C D : Type} `{abs A B, abs C D}.

  Global Instance pair_abs
         {A B C D : Type} `{abs A B, abs C D} :
    @abs (A * C)%type (B * D)%type _ _ pair_poset_pointwise := {
    abs_relation x y := x.1 \abs y.1 /\ x.2 \abs y.2
  }.
  Proof.
    move=> [x y] [z w] [Hxz Hyw] [a b] [c d] [Hac Hbd].
    move=> [Hyp Hyp']; split.
    by rewrite <- Hxz; rewrite <- Hac.
    by rewrite <- Hyw; rewrite <- Hbd.
  Defined.

  Context `{@setoidDec A _, @poset A _, @posetDec A _ _ _, @DomType.order A _ _}.
  Context `{@hasBot D _ _}.
  
  Local Program Instance pair_abs_map :
    @abs (A * C)%type Map![A, D] _ _ _ := {
    abs_relation x y := x.2 \abs y[x.1]!
  }.
  Next Obligation.
    move=> [x y] [z w] [Hxz Hyw] a b Hab Habs.
    rewrite <- Hxz; rewrite <- Hyw.
    by eapply abs_relation_proper; eauto; reflexivity.
  Qed.    
End Pairs.
