Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Classical_Prop.

Require Import Utf8.
Require Import Basics.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section Props.
  Global Program Instance Prop_setoid : setoid Prop := {
    eq_ := iff
  }.
  
  Global Program Instance Prop_poset : @poset Prop _ := {
    le_ := impl
  }.

  Global Program Instance Prop_hasBot : @hasBot Prop _ _ := {
    bot_ := False
  }.

  Global Program Instance Prop_hasTop : @hasTop Prop _ _ := {
    top_ := True
  }.
  
  Global Program Instance Prop_order : @order Prop _ _ := {
  }.
  Next Obligation.
    by case: (classic x); case: (classic y); firstorder.
  Qed.

  Global Program Instance Prop_hasJoin : @hasJoin Prop _ _ := {
    join_ := or
  }.
  Next Obligation. by firstorder. Qed.
  Next Obligation. by firstorder. Qed.
  Next Obligation. by firstorder. Qed.
End Props.


Section Morphisms.
  Context {A B : Type} `{setoid A, setoid B}.

  Definition morphism : Type := {functionT : A -> B | Proper (eq ==> eq) functionT}.

  Global Program Instance morphism_setoid : setoid morphism := {
    eq_ x y := forall z, (proj1_sig x) z ≡ (proj1_sig y) z
  }.
  Next Obligation.
    move=> [x xp] y /=.
    exact/eq_refl.
  Qed.
  Next Obligation.
    move=> [x xp] [y yp] /= Hxy z.
    exact/eq_sym/Hxy.
  Qed.
  Next Obligation.
    move=> [x xp] [y yp] [z zp] /= Hxy Hyz w.
    by apply/eq_trans; [apply Hxy | apply Hyz].
  Qed.

  Context `{@poset B _}.

  Global Program Instance morphism_poset : @poset morphism _ := {
    le_ x y := forall z, le ((proj1_sig x) z) ((proj1_sig y) z)
  }.
  Next Obligation.
    move=> [x xp] [y yp] /= Hxy z.
    exact/eq_le.
  Qed.
  Next Obligation.
    move=> [x xp] [y yp] /= Hxy Hyx z.
    by apply/le_antisym; [apply Hxy | apply Hyx].
  Qed.
  Next Obligation.
    move=> [x xp] [y yp] [z zp] /= Hxy Hyz w.
    by apply/le_trans; [apply Hxy | apply Hyz].
  Qed.
  
  Global Program Instance morphism_hasJoin `{@hasJoin B _ _} : @hasJoin morphism _ _ := {
    join_ x y := fun z => join ((proj1_sig x) z) ((proj1_sig y) z)
  }.
  Next Obligation.
    move: x y => [x xp] [y yp] /= z w Hzw.
    by rewrite -> Hzw; apply/eq_refl.
  Qed.
  Next Obligation.
    move: x y => [x xp] [y yp] /= z.
    exact: join_lhs.
  Qed.
  Next Obligation.
    move: x y => [x xp] [y yp] /= z.
    exact: join_rhs.
  Qed.
  Next Obligation.
    move: x y ub Hx Hy => [x xp] [y yp] [z zp] /= Hxz Hyz w.
    exact: join_lub.
  Qed.

  Global Program Instance morphism_hasTop `{@hasTop B _ _} : @hasTop morphism _ _ := {
    top_ := fun _ => top
  }.
  Next Obligation.
    by move=> ?; apply/top_prop.
  Qed.

  Global Program Instance morphism_hasBot `{@hasBot B _ _} : @hasBot morphism _ _ := {
    bot_ := fun _ => bot
  }.
  Next Obligation.
    by move=> ?; apply/bot_prop.
  Qed.

  (* Morphism as a function map *)
  Section Maplike.
    Context `{@setoidDec A _, @hasBot B _ _}.

    Global Program Instance morphism_maplike : maplike (Key:=A) (Elt:=B) (M:=morphism) := {
      mlookup key m := (proj1_sig m) key;
      madd k v m := fun (k':A) =>
                      if eqb k' k
                      then v
                      else (proj1_sig m) k';
      mremove k m := fun (k':A) =>
                       if eqb k' k
                       then bot
                       else (proj1_sig m) k'
    }.
    Next Obligation.
      move: m => [x xp] y z Hyz; rewrite -> Hyz.
      case: (eqb z k) => /=.
      exact: eq_refl.
      by rewrite -> Hyz; apply: eq_refl.
    Qed.
    Next Obligation.
      move: m => [x xp] y z Hyz; rewrite -> Hyz.
      case: (eqb z k) => /=.
      exact: eq_refl.
      by rewrite -> Hyz; apply: eq_refl.
    Qed.
    Next Obligation.
      move=> x y Hxy [z zp] [w wp] /= Hzw.
      by rewrite -> Hxy.
    Qed.
    Next Obligation.
      move=> x y Hxy z w Hzw [a ap] [b bp] Hab c.
      rewrite/proj1_sig.
      by rewrite -> Hxy; case: (eqb c y).
    Qed.
    Next Obligation.
      move=> x y Hxy [a ap] [b bp] Hab c.
      rewrite/proj1_sig.
      rewrite -> Hxy; case: (eqb c y).
      apply eq_le; exact/eq_refl.
      unfold le in Hab; simpl in Hab; exact/Hab.
    Qed.

    Lemma morphism_mlookup_spec (a:A) (m:morphism) :
      (proj1_sig m) a = m[a]!.
    Proof. done. Qed.
    
    Lemma morphism_madd_spec (m : morphism) (a a' : A) (b : B) :
      (m[a <- b]!)[a']! =
      if eqb a' a
      then b
      else m[a']!.
    Proof. by rewrite/madd/mlookup/morphism_maplike/proj1_sig. Qed.

    Lemma morphism_join_spec `{@hasJoin B _ _} (x y : morphism) (a : A) :
      (join x y)[a]! = join x[a]! y[a]!.
    Proof. done. Qed.

    Lemma if_madd (test:bool) (x y:morphism) (k:A) (v:B) :
      (if test then x else y)[k <- v]! = (if test then x[k <- v]! else y[k <- v]!).
    Proof. by case: test. Qed.

    Lemma if_mlookup (test:bool) (x y:morphism) (k:A) :
      (if test then x else y)[k]! = (if test then x[k]! else y[k]!).
    Proof. by case: test. Qed.
  End Maplike.
End Morphisms.
Notation "A '->>' B" := (@morphism A B _ _) (at level 15, right associativity).


Section Powersets.
  Context {A : Type} `{setoid A} `{@setoidDec A _}.
  Definition powerset : Type := A ->> bool.

  Section Setlike.
    Global Program Instance powerset_setlike : setlike (Elt:=A) (S:=powerset) := {
      sempty := fun _ => false;
      ssingleton elt := fun elt' => eqb elt elt';
      smem elt s := (proj1_sig s) elt;
      sadd elt s := fun elt' => if eqb elt elt'
                                then true
                                else (proj1_sig s) elt';
      sremove elt s := fun elt' => if eqb elt elt'
                                   then false
                                   else (proj1_sig s) elt'
    }.
    Next Obligation. by move=> x y ->. Qed.
    Next Obligation. 
    Proof.
      move=> x y Hxy /=. apply congr_if; try done.
      rewrite Hxy; reflexivity.
      unfold proj1_sig; destruct s.
      rewrite Hxy; reflexivity.
    Qed.
    Next Obligation. 
    Proof.
      move=> x y Hxy /=. apply congr_if; try done.
      rewrite Hxy; reflexivity.
      unfold proj1_sig; destruct s.
      rewrite Hxy; reflexivity. 
    Qed.
    Next Obligation.
      move=> x y Hxy z //=.
      by rewrite -> Hxy.
    Qed.
    Next Obligation.
      move=> x y Hxy z w Hzw.
      case: z Hzw => z zp; case: w => w wp Hzw //.
      by rewrite -> Hzw; rewrite -> Hxy.
    Qed.
    Next Obligation.
      move=> x y Hxy z w Hzw. intro; simpl.
      apply congr_if; try reflexivity.
      rewrite Hxy; reflexivity.
      rewrite Hzw; reflexivity.
    Qed.
    Next Obligation.
      move=> x y Hxy z w Hzw. intro; simpl.
      apply congr_if; try reflexivity.
      rewrite Hxy; reflexivity.
      rewrite Hzw; reflexivity.
    Qed.
    Next Obligation.
      remember (eqb e e') as Beqb; destruct Beqb; symmetry in HeqBeqb.
      {
        move/eqbP:HeqBeqb => HeqBeqb. left; split.
        apply HeqBeqb.
        symmetry in HeqBeqb. move/eqbP:HeqBeqb; apply.
      }
      {
        move/eqbP:HeqBeqb => HeqBeqb. right; split.
        apply HeqBeqb.
        assert (Heqsym:¬e' ≡ e). intro. apply HeqBeqb. symmetry; assumption.
        move/eqbP:Heqsym; apply.
      }
    Qed.
    Next Obligation.
      remember (eqb e' e) as Beqb; destruct Beqb; symmetry in HeqBeqb.
      {
        left; split; last auto.
        move/eqbP:HeqBeqb => HeqBeqb. symmetry; tauto.
      }        
      {
        right; split; last reflexivity.
        move/eqbP:HeqBeqb => HeqBeqb. 
        intro. apply HeqBeqb. symmetry; assumption.
      }
    Qed.
    Next Obligation.
      remember (eqb e' e) as Beqb; destruct Beqb; symmetry in HeqBeqb.
      {
        left; split; last auto.
        move/eqbP:HeqBeqb => HeqBeqb. symmetry; tauto.
      }        
      {
        right; split; last reflexivity.
        move/eqbP:HeqBeqb => HeqBeqb. 
        intro. apply HeqBeqb. symmetry; assumption.
      }
    Qed.

    Lemma powerset_smem_spec (a:A) (m:powerset) :
      (proj1_sig m) a = smem a m.
    Proof. done. Qed.
  End Setlike.

  Program Definition powerset_neg (s : powerset) : powerset :=
    (fun x => ~~ (s x)).
  Next Obligation.
    move: s=> [s sp] x y Hxy.
    red; rewrite -/eq/negb/proj1_sig.
    rewrite -> Hxy.
    by case: (s y).
  Qed.

  Global Add Parametric Morphism : powerset_neg with
      signature eq ==> eq as powerset_neg_mor.
  Proof.
    move=> [x xp] [y yp] Hxy z.
    rewrite/eq/= in Hxy.
    rewrite/powerset_neg/proj1_sig.
    by rewrite -> Hxy.
  Qed.

  Lemma powerset_neg_spec (s : powerset) (x : A) :
    x !\in (powerset_neg s) = ~~ (x !\in s).
  Proof. by rewrite/powerset_neg/=. Qed.
  
  Definition subset (s s': powerset) : Prop := 
    forall (x:A) (Hxin:smem x s), smem x s'.
  
  Lemma subset_neg : forall (l l': powerset) (Hsubset: subset l l'),
      subset (powerset_neg l') (powerset_neg l).
  Proof.
    unfold subset; simpl; intros.
    assert (proj1_sig l x → proj1_sig l' x); first by apply Hsubset.
    destruct (proj1_sig l x); auto.
    destruct (proj1_sig l' x).
    by apply Hxin.
    by apply H1; auto.
  Qed.

  Program Definition union (s s': powerset) : powerset := 
    (fun x => (smem x s) || (smem x s')).
  Next Obligation.  
    move => x y Hxy.
    rewrite/eq; apply/orP.
    case yins:(y !\in s); case yins':(y !\in s'); rewrite/orb.
    by rewrite {} Hxy; firstorder. 
    by rewrite {} Hxy; firstorder. 
    by rewrite {} Hxy; firstorder. 
    rewrite {} Hxy; rewrite {} yins yins'. by firstorder.
  Qed.

  Lemma union_lhs : forall (s s':powerset), subset s (union s s').
  Proof.
    move => [s sp] [s' sp'] x /= => Hxin.
    rewrite {} Hxin; by firstorder.
  Qed.

  Lemma union_rhs : forall (s s':powerset), subset s (union s' s).
  Proof.
    move => [s sp] [s' sp'] x /= => Hxin.
    rewrite {} Hxin. apply/orP. by firstorder.
  Qed.

  Lemma union_lub : forall (s s' t:powerset) (Hs: subset s t) (Hs': subset s' t),
    subset (union s s') t.
  Proof.
    move => [s sp] [s' sp'] [t tp]. 
    rewrite/subset /= => Hs Hs' x.
    move/orP => [Hx|Hx].
    apply:Hs. by [].
    apply:Hs'. by [].
  Qed.

  Lemma union_commutative : forall (s t:powerset),
    union s t ≡ union t s.
  Proof.
    move => [s sp] [s' sp'] z /=.
    case:(s z); case:(s' z); by firstorder.
  Qed.

  Lemma subset_union : forall (s t:powerset) (Hsub: subset s t),
    union s t ≡ t.
  Proof.
    move => [s sp] [s' sp'] Hsub z /=.
    rewrite/subset in Hsub. simpl in Hsub.
    case s'z:(s' z). 
    case:(s z); by firstorder.
    case sz:(s z); last by firstorder.
    apply Hsub in sz. rewrite s'z in sz. inversion sz.
  Qed.    

End Powersets.
Notation "'\powerset' A" := (@powerset A _) (at level 10).

Class restricting_maplike
      (K E M : Type)
      `{setoid K, @setoidDec K _}
      `{setoid E, @poset E _}
      `{setoid M, @poset M _, @hasBot M _ _,
        @maplike K E M _ _ _ _ _ _} : Type := RestrictingMaplike {
  mrestrict : forall (ks : \powerset K) (m : M), M;
  mrestrict_proper : Proper (eq ==> le ==> le) mrestrict;

  mrestrict_lookup_spec : forall (s: \powerset K) (m: M) (loc: K)
    (Hloc:loc !\in s),
    eq m[loc]! (mrestrict s m)[loc]!;
       
  mrestrict_update_spec : forall (s: \powerset K) (m: M) (loc: K) (v: E)
    (Hloc:loc !\in s),
    eq (mrestrict s m[loc <- v]!) (mrestrict s m)[loc <- v]!;

  mrestrict_double : forall l (m:M),
    mrestrict l (mrestrict l m) ≡ mrestrict l m;

  mrestrict_commutative : forall l l' (m:M),
    mrestrict l (mrestrict l' m) ≡ mrestrict l' (mrestrict l m);

  mrestrict_bot : forall l, mrestrict l (bot:M) ≡ bot;

  mrestrict_subset : forall l l' (m:M) (Hsubset: subset l l'),
    mrestrict l' (mrestrict l m) ≡ mrestrict l m;

  mrestrict_subset' : forall l l' (m:M) (Hsubset: subset l l'),
    le (mrestrict l m) (mrestrict l' m);
  
  mrestrict_le : forall l (m:M),
    le (mrestrict l m) m
}.

Global Add Parametric Morphism `{restricting_maplike} : mrestrict with
    signature eq ==> le ==> le as mrestrict_mor.
Proof. exact: mrestrict_proper. Qed.

Global Add Parametric Morphism `{restricting_maplike} : mrestrict with
    signature eq ==> eq ==> eq as mrestrict_eq_mor.
Proof.
  move=> x y Hxy z w Hzw.
  apply: le_antisym; apply mrestrict_proper.
  done.
  exact/DomType.eq_le.
  exact/eq_sym.
  exact/DomType.eq_le/eq_sym.
Qed.

Section Mrestrict_join.
  Context {K E M: Type} `{restricting_maplike K E M, @hasJoin E _ _, @hasJoin M _ _}.

  Definition mrestrict_join (s : \powerset K) (m n : M) :=
    join (mrestrict s m) (mrestrict (powerset_neg s) n).

  Global Add Parametric Morphism : mrestrict_join with
      signature (eq ==> le ==> le ==> le) as mrestrict_join_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    rewrite/mrestrict_join.
    apply join_le_mor.
    by rewrite -> Hxy; rewrite -> Hzw; apply/DomType.eq_le/eq_refl.
    by rewrite -> Hxy; rewrite -> Hab; apply/DomType.eq_le/eq_refl.
  Qed.

  Global Add Parametric Morphism : mrestrict_join with
      signature (eq ==> eq ==> eq ==> eq) as mrestrict_join_eq_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    apply: le_antisym.
    by rewrite <- Hxy; rewrite -> Hzw; rewrite -> Hab; apply/DomType.eq_le/eq_refl.
    by rewrite -> Hxy; rewrite <- Hzw; rewrite <- Hab; apply/DomType.eq_le/eq_refl.
  Qed.
  
  Class restricting_join_maplike : Type := RestrictingJoinMaplike {
    mrestrict_distributive : forall l (m m':M),
      mrestrict l (m ⊔ m') ≡ mrestrict l m ⊔ mrestrict l m';

    mrestrict_join_powerset_neg : forall l (m:M), 
      mrestrict l m ⊔ mrestrict (powerset_neg l) m ≡ m;
  
    mrestrict_merge : forall l l' (m: M),
      mrestrict (union l l') m ≡ mrestrict l m ⊔ mrestrict l' m;

    mrestrict_join_update : forall (s: \powerset K) (m n: M) (loc: K) (v: E),
      eq
        (mrestrict_join s m n)[loc <- v]!
        (mrestrict_join
           s
           (if loc !\in s then m[loc <- v]! else m)
           (if ~~ (loc !\in s) then n[loc <- v]! else n));

    mrestrict_join_merge : forall l l' (m m': M),
      mrestrict_join l m (mrestrict_join l' m m') 
      ≡ mrestrict_join (union l l') m m'
  }.

  Context `{restricting_join_maplike}.

  Lemma mrestrict_join_imply_mrestrict (m m':M) l :
    mrestrict_join l m m' ≡ mrestrict_join l (mrestrict l m) m'.
  Proof.
    rewrite/mrestrict_join.
    by erewrite mrestrict_double; reflexivity.
  Qed.

  Lemma mrestrict_join_subset l l' (m m':M) (Hsubset:subset l' l) :
    mrestrict_join l' m (mrestrict_join l m m') ≡ mrestrict_join l m m'.
  Proof.
    rewrite/mrestrict_join.
    rewrite -> mrestrict_distributive.
    apply: le_antisym; do ? apply: join_lub.
    { eapply le_trans; last apply join_lhs.
        by apply mrestrict_subset'; auto.
    }
    { eapply le_trans; last apply join_lhs.
        by apply mrestrict_le; auto.
    }
    { eapply le_trans; last apply join_rhs.
        by apply mrestrict_le; auto.
    }
    { rewrite -{1}(mrestrict_join_powerset_neg l' (mrestrict l m)).
      apply: join_lub.
      eapply le_trans; last apply join_lhs.
      rewrite -> mrestrict_commutative.
      rewrite -> mrestrict_subset; auto.
      exact/DomType.eq_le/eq_refl.
      eapply le_trans; last apply join_rhs.
      exact: join_lhs.
    }
    { eapply le_trans; last apply join_rhs.
      eapply le_trans; last apply join_rhs.
      rewrite -> mrestrict_subset.
      exact/DomType.eq_le/eq_refl.
      exact: subset_neg.
    }
  Qed.
End Mrestrict_join.


Global Program Instance morphism_restricting_maplike
       A B
       `{setoid A, @setoidDec A _}
       `{setoid B, @poset B _, @hasBot B _ _} :
  @restricting_maplike A B (A ->> B) _ _ _ _ _ _ _ _ := {
  mrestrict s m :=
    fun (x:A) =>
      if x !\in s
        then m[x]!
        else bot
  }.
Next Obligation.
  move=> x y Hxy.
  red; rewrite -/eq.
  rewrite -> Hxy.
  case: (y !\in s).
  by rewrite -> Hxy; apply: eq_refl.
  exact: eq_refl.
Qed.
Next Obligation.
  move=> x y Hxy z w Hzw a.
  rewrite/proj1_sig.
  rewrite -> Hxy.
  case: (a !\in y).
  exact: Hzw.
  exact/DomType.eq_le/eq_refl.
Qed.
Next Obligation.
  simpl in *. rewrite Hloc.
  exact: eq_refl.
Qed.  
Next Obligation.
  move: s m Hloc => [s sp] [m mp] Hloc x.
  rewrite/=/proj1_sig.
  remember (eqb x loc) as Heq; destruct Heq; last by apply eq_refl.
  remember (s x) as Hsx; destruct Hsx; first by apply eq_refl.
  symmetry in HeqHeq. move/eqbP:HeqHeq => HeqHeq. 
  assert (Hsx:s x). rewrite HeqHeq; auto.
  rewrite <- HeqHsx in Hsx. inversion Hsx.
Qed.  
Next Obligation.
  move: l=> [l lp].
  move=> loc; rewrite/=/proj1_sig.
  by case: (l loc); apply: eq_refl.
Qed.
Next Obligation.
  move: l l'=> [l lp] [l' lp'].
  move=> loc; rewrite/=/proj1_sig.
  case: (l loc); first exact: eq_refl.
  by case: (l' loc); apply: eq_refl.
Qed.
Next Obligation.
  move: l=> [l lp].
  move=> loc; rewrite/=/proj1_sig.
  by case: (l loc); apply: eq_refl.
Qed.
Next Obligation.
  move: l l' Hsubset=> [l lp] [l' lp'] Hsubset.
  rewrite/subset/= in Hsubset.
  move=> loc; rewrite/=/proj1_sig.
  case Hl: (l loc)=> /=.
  by rewrite Hsubset; [apply: eq_refl | auto].
  by case: (l' loc); apply: eq_refl.
Qed.
Next Obligation.
  move: l l' Hsubset m=> [l lp] [l' lp'] Hsubset [m mp].
  rewrite/subset/= in Hsubset.
  move=> loc; rewrite/=/proj1_sig.
  case Hl: (l loc); last exact: bot_prop.
  rewrite Hsubset; last done.
  exact/DomType.eq_le/eq_refl.
Qed.    
Next Obligation.
  move: l m=> [l lp] [m mp].
  move=> loc; rewrite/=/proj1_sig.
  case: (l loc)=> /=.
  exact/DomType.eq_le/eq_refl.
  exact/bot_prop.
Qed.

Lemma morphism_mrestrict_spec
  A B
  `{setoid A, @setoidDec A _}
  `{setoid B, @poset B _, @hasBot B _ _}
  (s : \powerset A) (m : A ->> B) (a : A) :
  (mrestrict s m)[a]! = if a !\in s then m[a]! else bot.
Proof. done. Qed.

Lemma morphism_mrestrict_madd_spec
  A B
  `{setoid A, @setoidDec A _}
  `{setoid B, @poset B _, @hasBot B _ _}
  (s : \powerset A) (m : A ->> B) (a : A) (Ha : a !\in s) (b : B) :
  eq (mrestrict s m)[a <- b]! (mrestrict s m[a <- b]!).
Proof.
  move: s m Ha=> [s sp] [m mp] /=.
  move=> Ha loc /=.
  case Heqb: (eqb loc a); last exact: eq_refl.
  move/eqbP in Heqb; rewrite Heqb Ha.
  exact: eq_refl.
Qed.

Global Program Instance morphism_restricting_join_maplike
       A B
       `{setoid A, @setoidDec A _}
       `{setoid B, @poset B _, @hasBot B _ _, @hasJoin B _ _} :
  @restricting_join_maplike A B (A ->> B) _ _ _ _ _ _ _ _ _ _ := {
  }.
Next Obligation.
  move: l=> [l lp].
  move=> loc; rewrite/=/proj1_sig.
  case: (l loc); first exact: eq_refl.
  by rewrite join_bot; apply: eq_refl.
Qed.  
Next Obligation.
  move: l m=> [l lp] [m mp].
  move=> loc; rewrite/=/proj1_sig.
  case: (l loc)=> /=.
  exact: join_bot.
  by rewrite join_commutative join_bot; apply: eq_refl.
Qed.
Next Obligation.
  move: l l'=> [l lp] [l' lp'].
  move=> loc; rewrite/=/proj1_sig.
  case Hl: (l loc); case Hl': (l' loc)=> /=.
  by rewrite join_convolutive; apply: eq_refl.
  by rewrite join_bot; apply: eq_refl.
  by rewrite join_commutative join_bot; apply: eq_refl.
  by rewrite join_bot; apply: eq_refl.
Qed.
Next Obligation.
  move: s m n=> [s sp] [m mp] [n np] loc'.
  rewrite/proj1_sig/=.
  case Heqb: (eqb loc' loc); move/eqbP in Heqb.
  { rewrite -> Heqb.
    case: (s loc)=> /=; rewrite -> Heqb; rewrite eqb_refl.
    exact/eq_sym/join_bot.
    rewrite join_commutative.
    exact/eq_sym/join_bot.
  }    
  { case Hloc: (s loc); case Hloc': (s loc'); rewrite/proj1_sig/=.
    case Heqb': (eqb loc' loc).
    by contradict Heqb; apply/eqbP.
    exact: eq_refl.
    exact: eq_refl.
    exact: eq_refl.
    case Heqb': (eqb loc' loc).
    by contradict Heqb; apply/eqbP.
    exact: eq_refl.
  }
Qed.
Next Obligation.
  move: l l' m m'=> [l lp] [l' lp'] [m mp] [m' mp'] loc.
  rewrite/proj1_sig/=.
  case: (l loc); case: (l' loc)=> /=; try reflexivity.
  rewrite join_bot join_commutative join_bot; reflexivity.
  rewrite join_commutative join_bot join_commutative join_bot; reflexivity.
Qed.

Section Options.
  Context {A : Type} `{setoid A}.

  Definition option_eq :=
    (fun lhs y =>
       match lhs, y with
         | None, None => True
         | Some lhs', Some y' => lhs' ≡ y'
         | _, _ => False
       end).

  Global Program Instance option_setoid : setoid (option A) := {
    eq_ := option_eq
  }.
  Next Obligation.
    move=> [x |]; last by [].
    exact/eq_refl.
  Qed.
  Next Obligation.
    move=> [x |] [y |] //=.
    exact/eq_sym.
  Qed.
  Next Obligation.
    move=> [x |] [y |] [z |] //=.
    exact/eq_trans.
  Qed.

  Global Program Instance option_usual_setoid `{@usual_setoid A _} : @usual_setoid (option A) _ := {
  }.
  Next Obligation.
    move: x y H1 => [x|] [y|] //=.
    rewrite/option_eq => Heq.
    by congr Some; apply/usual_eqP'.
  Qed.
  
  Context `{@poset A _}.
  Definition option_le :=
    (fun x y =>
       match x, y with
         | None, _ => True
         | Some x', Some y' => le x' y'
         | _, _ => False
       end).

  Global Program Instance option_poset : @poset (option A) _ := {
    le_ := option_le
  }.
  Next Obligation.
    move=> [x |] [y |] //= Hxy.
    exact/eq_le.
  Qed.
  Next Obligation.
    move=> [x |] [y |] //= Hxy Hyx.
    rewrite/eq; exact/le_antisym.
  Qed.
  Next Obligation.
    move=> [x |] [y |] [z |] //= Hxy Hyz.
    by apply/le_trans; eauto.
  Qed.

  Global Program Instance option_hasBot : @hasBot (option A) _ _ := {
    bot_ := None
  }.

  Global Program Instance option_hasTop `{@hasTop A _ _} : @hasTop (option A) _ _ := {
    top_ := Some top
  }.
  Next Obligation.
    case: x => //=.
    exact/top_prop.
  Qed.
  
  Global Program Instance option_order `{@order A _ _} : @order (option A) _ _ := {
  }.
  Next Obligation.
    by case: x; case: y => //=; firstorder.
  Qed.

  Definition option_join `{@hasJoin A _ _} :=
    (fun x y =>
       match x, y with
         | None, _ => y
         | _, None => x
         | Some x', Some y' => Some (join x' y')
       end).

  Global Program Instance option_hasJoin `{@hasJoin A _ _} : @hasJoin (option A) _ _ := {
    join_ := option_join
  }.
  Next Obligation.
    case: x => [x |]; case: y => [y |] //=.
    exact/join_lhs.
    exact/eq_le/eq_refl.
  Qed.
  Next Obligation.
    case: x => [x |]; case: y => [y |] //=.
    exact/join_rhs.
    exact/eq_le/eq_refl.
  Qed.
  Next Obligation.
    move: x y ub Hx Hy => [x |] [y |] [z |] Hx Hy //=.
    exact/join_lub.
  Qed.

  Section Dec.
    Context `{@setoidDec A _}.
    Definition option_eqb :=
      (fun x y =>
         match x, y with
           | Some x', Some y' => eqb x' y'
           | None, None => true
           | _, _ => false
         end).

    Global Program Instance option_setoidDec : @setoidDec (option A) _ := {
      eqb_ := option_eqb
    }.
    Next Obligation.
      rewrite/option_eqb; move: x y => [x |] [y |]; try constructor => //=.
      exact/eqbP.
    Qed.

    Context `{@posetDec A _ _ _}.
    Definition option_leb :=
      (fun x y =>
         match x, y with
           | None, _ => true
           | Some x', Some y' => leb x' y'
           | _, _ => false
         end).

    Global Program Instance option_posetDec : @posetDec (option A) _ _ _ := {
      leb_ := option_leb
    }.
    Next Obligation.
      rewrite /option_leb; move: x y => [x |] [y |]; try constructor => //=.
      exact/lebP.
    Qed.
  End Dec.
End Options.


Section Withtops.
  Context {A : Type} `{setoid A}.

  Inductive withtop A :=
  | Elt of A
  | Top.
  Arguments Top {A}.

  Definition withtop_eq :=
    (fun x y =>
       match x, y with
         | Top, Top => True
         | Elt x', Elt y' => x' ≡ y'
         | _, _ => False
       end).

  Global Program Instance withtop_setoid : setoid (withtop A) := {
    eq_ := withtop_eq
  }.
  Next Obligation.
    move=> [x |]; last by [].
    exact/eq_refl.
  Qed.
  Next Obligation.
    move=> [x |] [y |] //=.
    exact/eq_sym.
  Qed.
  Next Obligation.
    move=> [x |] [y |] [z |] //=.
    exact/eq_trans.
  Qed.

  Global Add Parametric Morphism : (Elt (A:=A)) with
    signature eq ==> eq as Elt_mor.
  Proof. done. Qed.
  
  Global Program Instance withtop_usual_setoid `{@usual_setoid A _} : @usual_setoid (withtop A) _ := {
  }.
  Next Obligation.
    move: x y H1 => [x|] [y|] //=.
    rewrite/withtop_eq => Heq.
    by congr Elt; apply/usual_eqP'.
  Qed.
  
  Context `{@poset A _}.
  Definition withtop_le :=
    (fun x y =>
       match x, y with
         | _, Top => True
         | Elt x', Elt y' => le x' y'
         | _, _ => False
       end).

  Global Program Instance withtop_poset : @poset (withtop A) _ := {
    le_ := withtop_le
  }.
  Next Obligation.
    move=> [x |] [y |] //= Hxy.
    exact/eq_le.
  Qed.
  Next Obligation.
    move=> [x |] [y |] //= Hxy Hyx.
    rewrite/eq.
    exact/le_antisym.
  Qed.
  Next Obligation.
    move=> [x |] [y |] [z |] //= Hxy Hyz.
    by apply/le_trans; eauto.
  Qed.
  
  Global Program Instance withtop_hasBot `{@hasBot A _ _} : @hasBot (withtop A) _ _ := {
    bot_ := Elt bot
  }.
  Next Obligation.
    case: x => //=.
    exact/bot_prop.
  Qed.

  Global Program Instance withtop_hasTop : @hasTop (withtop A) _ _ := {
    top_ := Top
  }.
  Next Obligation.
    by case: x.
  Qed.

  Lemma Top_top : Top = top.
  Proof. done. Qed.

  Global Program Instance withtop_order `{@order A _ _} : @order (withtop A) _ _ := {
  }.
  Next Obligation.
    by move: x y => [x |] [y |]; firstorder.
  Qed.

  Definition withtop_join `{@hasJoin A _ _} :=
    (fun x y =>
       match x, y with
         | Top, _ => Top
         | _, Top => Top
         | Elt x', Elt y' => Elt (join x' y')
       end).

  Global Program Instance withtop_hasJoin `{@hasJoin A _ _} : @hasJoin (withtop A) _ _ := {
    join_ := withtop_join
  }.
  Next Obligation.
    move: x y => [x |] [y |] //=.
    exact/join_lhs.
  Qed.
  Next Obligation.
    move: x y => [x |] [y |] //=.
    exact/join_rhs.
  Qed.
  Next Obligation.
    move: x y ub Hx Hy => [x |] [y |] [z |] Hx Hy //=.
    exact/join_lub.
  Qed.

  Section Dec.
    Context `{@setoidDec A _}.
    Definition withtop_eqb :=
      (fun x y =>
         match x, y with
           | Elt x', Elt y' => eqb x' y'
           | Top, Top => true
           | _, _ => false
         end).

    Global Program Instance withtop_setoidDec : @setoidDec (withtop A) _ := {
      eqb_ := withtop_eqb
    }.
    Next Obligation.
      rewrite /withtop_eqb; move: x y => [x |] [y |]; try constructor => //=.
      exact/eqbP.
    Qed.

    Context `{@posetDec A _ _ _}.
    Definition withtop_leb :=
      (fun x y =>
         match x, y with
           | _, Top => true
           | Elt x', Elt y' => leb x' y'
           | _, _ => false
         end).

    Global Program Instance withtop_posetDec : @posetDec (withtop A) _ _ _ := {
      leb_ := withtop_leb
    }.
    Next Obligation.
      rewrite /withtop_leb; move: x y => [x |] [y |]; try constructor => //=.
      exact/lebP.
    Qed.
  End Dec.

  Global Program Instance withtop_maplike (K E : Type)
    `{setoid K, @setoidDec K _}
    `{setoid E, @poset E _, @hasTop E _ _}
    `{@maplike K E A _ _ _ _ _ _} : @maplike K E (withtop A) _ _ _ _ _ _ := {
    mlookup k m := if m is Elt m' then mlookup k m' else top;
    madd k e m := if m is Elt m' then Elt (madd k e m') else Top;
    mremove k m := if m is Elt m' then Elt (mremove k m') else Top
  }.
  Next Obligation.
    move=> x y Hxy [z|] [w|] Hzw //; last (apply eq_le;exact/eq_refl).
    rewrite/eq/eq_/withtop_setoid/withtop_eq in Hzw.
    apply mlookup_le_mor; auto.
    apply top_prop.
  Qed.
  Next Obligation.
    move=> x y Hxy z w Hzw [a|] [b|] Hab //.
    rewrite/le/le_/withtop_poset/withtop_le in Hab *.
    by apply madd_le_mor; auto.
  Qed.
  Next Obligation.
    move=> x y Hxy [z|] [w|] Hzw //.
    rewrite/eq/eq_/withtop_setoid/withtop_eq in Hzw *.
    apply mremove_le_proper; auto.
  Qed.

  Global Program Instance withtop_restricting_maplike (K E : Type)
    `{setoid K, @setoidDec K _}
    `{setoid E, @poset E _, @hasTop E _ _}
    `{@hasBot A _ _}
    `{@maplike K E A _ _ _ _ _ _, 
      @restricting_maplike K E A _ _ _ _ _ _ _ _} : 
    @restricting_maplike K E (withtop A) _ _ _ _ _ _ _ _ := {
    mrestrict s m :=
      if m is Elt m'
      then Elt (mrestrict s m')
      else Top
  }.
  Next Obligation.
    move=> x y Hxy [z|] [w|] Hzw //.
    rewrite/le/= in Hzw *.
    rewrite -> Hxy; rewrite -> Hzw.
    exact/DomType.eq_le/eq_refl.
  Qed.
  Next Obligation.
    move: m=> [m|]; last exact: eq_refl.
    rewrite/mrestrict/=.
    destruct H8.
    exact: mrestrict_lookup_spec0.
  Qed.
  Next Obligation.
    move: m=> [m|]; last exact: eq_refl.
    rewrite/mrestrict/=.
    destruct H8.
    exact: mrestrict_update_spec0.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_double.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_commutative.
  Qed.
  Next Obligation.
    rewrite/bot/eq/=.
    exact: mrestrict_bot.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_subset.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_subset'.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/le/=.
    exact: mrestrict_le.
  Qed.
  
  Global Program Instance withtop_restricting_join_maplike (K E : Type)
    `{setoid K, @setoidDec K _}
    `{setoid E, @poset E _, @hasTop E _ _}
    `{@hasJoin A _ _, @hasBot A _ _,
      @maplike K E A _ _ _ _ _ _, 
      @restricting_maplike K E A _ _ _ _ _ _ _ _, 
      @restricting_join_maplike K E A _ _ _ _ _ _ _ _ _ _} : 
    @restricting_join_maplike K E (withtop A) _ _ _ _ _ _ _ _
    (@withtop_restricting_maplike _ _ _ _ _ _ _ _ _ _) _ := {
  }.
  Next Obligation.
    move: m m'=> [m|] [m'|] //; rewrite/eq/=.
    exact: mrestrict_distributive.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_join_powerset_neg.
  Qed.
  Next Obligation.
    move: m=> [m|] //; rewrite/eq/=.
    exact: mrestrict_merge.
  Qed.
  Next Obligation.
    rewrite/mrestrict_join.
    move: m=> [m|].
    move: n=> [n|].
    { assert (J := (mrestrict_join_update s m n loc v)).
      case Heqb: (loc !\in s); rewrite/eq/=; rewrite -> J; rewrite -> Heqb.
      exact: eq_refl.
      exact: eq_refl.
    }
    { rewrite [mrestrict _ Top]/=.
      rewrite join_commutative.
      rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
      rewrite [madd _ _ Top]/madd/withtop_maplike.
      case: (loc !\in s).
      { rewrite /=.
        rewrite join_commutative.
        rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
        exact: eq_refl.
      }
      { rewrite /=.
        rewrite join_commutative.
        rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
        exact: eq_refl.
      }
    }
    { rewrite [mrestrict _ Top]/=.
      rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
      rewrite [madd _ _ Top]/madd/withtop_maplike.
      case: (loc !\in s).
      { rewrite /=.
        rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
        exact: eq_refl.
      }
      { rewrite /=.
        rewrite [join Top _]/join/join_/withtop_hasJoin/withtop_join.
        exact: eq_refl.
      }
    }
  Qed.
  Next Obligation.
    move: m m'=> [m|] [m'|] //; rewrite/eq/=.
    exact: mrestrict_join_merge.
  Qed.

  Lemma if_Elt test (x y:A) :
    (if test then Elt x else Elt y) = Elt (if test then x else y).
  Proof. by case: test. Qed.  
End Withtops.


Section Pairs.
  Context {A B : Type} `{setoid A, setoid B}.

  Program Definition pair_eq :=
    (fun (x y:A*B) =>
       (fst x) ≡ (fst y) /\ (snd x) ≡ (snd y)).

  Global Program Instance pair_setoid : setoid (A * B) := {
    eq_ := pair_eq
  }.
  Next Obligation.
    rewrite/pair_eq; move=> [x y] //=.
    by split; apply/eq_refl.
  Qed.
  Next Obligation.
    rewrite/pair_eq; move=> [x y] [z w] //= [Hxz Hyw].
    by split; apply/eq_sym.
  Qed.
  Next Obligation.
    rewrite/pair_eq; move=> [x y] [z w] [a b] //= [Hxz Hyw] [Hza Hwb].
    by split; apply/eq_trans; eauto.
  Qed.

  Global Program Instance pair_usual_setoid `{@usual_setoid A _, @usual_setoid B _} : @usual_setoid (A * B) _ := {
  }.
  Next Obligation.
    by inversion H3; congr pair; apply/usual_eqP'.
  Qed.

  Context `{@poset A _, @poset B _}.
  Definition pair_le :=
    (fun (x:A*B) (y:A*B) =>
       (fst x) ⊑! (fst y) \/
       ((fst x) ≡ (fst y) /\ (snd x) ⊑ (snd y))).

  Global Program Instance pair_poset : @poset (A * B) _ := {
    le_ := pair_le
  }.
  Next Obligation.
    rewrite/pair_eq/pair_le; move=> [x y] [z w] //= [Hxz Hyw].
    right; split; first by [].
    exact/eq_le.
  Qed.
  Next Obligation.
    rewrite/pair_eq/pair_le; move=> [x y] [z w] //= [[Hxz Hxz'] | [Hxz Hyw]] [[Hzx Hzx'] | [Hzx Hwy]].
    by case Hxz'; apply/le_antisym.
    by case Hxz'; apply/eq_sym.
    by case Hzx'; apply/eq_sym.
    split; first by [].
    exact/le_antisym.
  Qed.
  Next Obligation.
    rewrite/pair_eq/pair_le; move=> [x y] [z w] [a b] //=.
    move=> [Hxz | [Hxz Hyw]] [Hza | [Hza Hwb]].
    by left; rewrite <- Hza.
    by left; rewrite <- Hza.
    by left; rewrite -> Hxz.
    right; split; first by rewrite -> Hxz.
    by rewrite -> Hyw.
  Qed.    

  Global Program Instance pair_hasBot `{@hasBot A _ _, @hasBot B _ _} : @hasBot (A * B) _ _ := {
    bot_ := (bot, bot)
  }.
  Next Obligation.
    case: (classic (a ≡ bot)) => Ha //=; [right | left]; split.
    exact/eq_sym.
    exact/bot_prop.
    exact/bot_prop.
    by rewrite/neq; contradict Ha; firstorder.
  Qed.
  
  Global Program Instance pair_hasTop `{@hasTop A _ _, @hasTop B _ _} : @hasTop (A * B) _ _ := {
    top_ := (top, top)
  }.
  Next Obligation.
    case: (classic (a ≡ top)) => Ha //=; [right | left]; split=> //=.
    exact/top_prop.
    exact/top_prop.
  Qed.

  Global Program Instance pair_order `{@order A _ _, @order B _ _} : @order (A * B) _ _ := {
  }.
  Next Obligation.
    case: (classic (a0 ≡ a))=> Ha; [case: (le_total b0 b)=> Hb | case: (le_total a0 a)=> Ha'].
    by left; right.
    by right; right; split; first by apply/eq_sym.
    left; left; split; first by [].
    by rewrite/neq; contradict Ha; apply/eq_sym; firstorder.
    right; left; split; first by [].
    by rewrite/neq; contradict Ha; firstorder.
  Qed.    

  Section Dec.
    Context `{@setoidDec A _, @setoidDec B _}.
    Definition pair_eqb :=
      (fun (x y:A*B) =>
         andb (eqb (fst x) (fst y)) (eqb (snd x) (snd y))).

    Global Program Instance pair_setoidDec : @setoidDec (A * B) _ :=
      {
        eqb_ := pair_eqb
      }.
    Next Obligation.
      rewrite/pair_eq/pair_eqb => //=.
      apply/equivP; first apply andP.
      by constructor; move=> [Ha Hb]; split; apply/eqbP.
    Qed.

    Context `{@posetDec A _ _ _, @posetDec B _ _ _}.
    Definition pair_leb :=
      (fun (x y:A*B) =>
         orb
           (ltb (fst x) (fst y))
           (andb (eqb (fst x) (fst y)) (leb (snd x) (snd y)))).

    Global Program Instance pair_posetDec : @posetDec (A * B) _ _ _ := {
      leb_ := pair_leb
    }.
    Next Obligation.
      rewrite/pair_le/pair_leb => //=.
      apply/equivP; first by apply orP.
      constructor; move=> [Ha | Hb].
      by left; apply/ltP.
      by right; split; [apply/eqbP | apply/lebP]; apply Bool.andb_true_iff in Hb; firstorder.
      by left; apply/ltP.
      by right; apply/andP; split; [apply/eqbP | apply/lebP]; firstorder.
    Qed.
  End Dec.

  Add Parametric Morphism : (@fst A B) with
    signature eq ==> eq as fst_mor.
  Proof. by move=> [x y] [z w] [Hxz Hyw]. Qed.

  Add Parametric Morphism : (@snd A B) with
    signature eq ==> eq as snd_mor.
  Proof. by move=> [x y] [z w] [Hxz Hyw]. Qed.

  Section Pointwises.
    Local Program Instance pair_poset_pointwise : @poset (A * B) _ := {
      le_ x y := (le x.1 y.1) /\ (le x.2 y.2)
    }.
    Next Obligation.
      move=> [x y] [z w] [Hxz Hyw].
      by split; apply/DomType.eq_le.
    Qed.
    Next Obligation.
      move=> [x y] [z w] [Hxz Hyw] [Hzx Hwy].
      by split; apply/le_antisym.
    Qed.
    Next Obligation.
      move=> [x y] [z w] [a b] [Hxz Hyw] [Hza Hwb].
      by split; eapply le_trans; eauto.
    Qed.

    Add Parametric Morphism : (@fst A B) with
      signature le ==> le as fst_le_mor.
    Proof. by move=> [x y] [z w] [Hxz Hyw]. Qed.

    Add Parametric Morphism : (@snd A B) with
      signature le ==> le as snd_le_mor.
    Proof. by move=> [x y] [z w] [Hxz Hyw]. Qed.

    Global Program Instance pair_hasBot_pointwise `{@hasBot A _ _, @hasBot B _ _} : @hasBot (A * B) _ _ := {
      bot_ := (bot, bot)
    }.
    Next Obligation. by split; apply/bot_prop. Qed.
      
    Global Program Instance pair_hasTop_pointwise `{@hasTop A _ _, @hasTop B _ _} : @hasTop (A * B) _ _ := {
      top_ := (top, top)
    }.
    Next Obligation. by split; apply/top_prop. Qed.

    Global Program Instance pair_hasJoin_pointwise `{@hasJoin A _ _, @hasJoin B _ _} : @hasJoin (A * B) _ _ := {
      join_ x y := (join x.1 y.1, join x.2 y.2)
    }.
    Next Obligation. by split; apply/join_lhs. Qed.
    Next Obligation. by split; apply/join_rhs. Qed.
    Next Obligation.
      move: Hx Hy=> [Hx1 Hx2] [Hy1 Hy2].
      by split; apply/join_lub.
    Qed.

    Global Program Instance pair_posetDec_pointwise `{@setoidDec A _, @setoidDec B _, @posetDec A _ _ _, @posetDec B _ _ _} : @posetDec (A * B) _ _ _ := {
      leb_ x y := (leb x.1 y.1) && (leb x.2 y.2)
    }.
    Next Obligation.
      apply/equivP; first apply andP.
      by constructor; move=> [Ha Hb]; split; apply/lebP.
    Qed.
  End Pointwises.
End Pairs.


Section Sums.
  Context {A B : Type} `{setoid A, setoid B}.

  Definition sum_eq :=
    (fun (x y:A+B) =>
       match x, y with
         | inl x', inl y' => x' ≡ y'
         | inr x', inr y' => x' ≡ y'
         | _, _ => False
       end).

  Global Program Instance sum_setoid : setoid (A + B) := {
    eq_ := sum_eq
  }.
  Next Obligation.
    by move=> [x | y]; apply/eq_refl.
  Qed.
  Next Obligation.
    by move=> [x | y] [z | w] Hxz //=; apply/eq_sym.
  Qed.
  Next Obligation.
    by move=> [x | y] [z | w] [a | b] Hxz Hza //=; apply/eq_trans; eauto.
  Qed.

  Global Program Instance sum_usual_setoid `{@usual_setoid A _, @usual_setoid B _} : @usual_setoid (A + B) _ := {
  }.
  Next Obligation.
    move: x y H3 => [x|x] [y|y] //=; rewrite/sum_eq => Heq.
    by congr inl; apply/usual_eqP'.
    by congr inr; apply/usual_eqP'.
  Qed.    

  Context `{@poset A _, @poset B _}.
  Definition sum_le :=
    (fun (x y:A+B) =>
       match x, y with
         | inl x', inl y' => x' ⊑ y'
         | inl _, inr _ => True
         | inr _, inl _ => False
         | inr x', inr y' => x' ⊑ y'
       end).

  Global Program Instance sum_poset : @poset (A + B) _ := {
    le_ := sum_le
  }.
  Next Obligation.
    by move=> [x | y] [z | w] Hxz //=; apply/eq_le.
  Qed.
  Next Obligation.
    by move=> [x | y] [z | w] Hxz //=; apply/le_antisym.
  Qed.
  Next Obligation.
    by move=> [x | y] [z | w] [a | b] Hxz Hza //=; apply/le_trans; eauto.
  Qed.

  Global Program Instance sum_order `{@order A _ _, @order B _ _} : @order (A + B) _ _ := {
  }.
  Next Obligation.
    by move: x y => [x | y] [z | w]; firstorder.
  Qed.

  Section Dec.
    Context `{@setoidDec A _, @setoidDec B _}.
    Definition sum_eqb :=
      (fun (x y:A+B) =>
         match x, y with
           | inl x', inl y' => eqb x' y'
           | inr x', inr y' => eqb x' y'
           | _, _ => false
         end).

    Global Program Instance sum_setoidDec : @setoidDec (A + B) _ := {
      eqb_ := sum_eqb
    }.
    Next Obligation.
      move: x y => [x | y] [z | w] //=; rewrite/eq; try exact/eqbP.
      by constructor.
      by constructor.
    Qed.

    Context `{@posetDec A _ _ _, @posetDec B _ _ _}.
    Definition sum_leb :=
      (fun (x y:A+B) =>
         match x, y with
           | inl x', inl y' => leb x' y'
           | inl _, inr _ => true
           | inr _, inl _ => false
           | inr x', inr y' => leb x' y'
         end).

    Global Program Instance sum_posetDec : @posetDec (A + B) _ _ _ := {
      leb_ := sum_leb
    }.
    Next Obligation.
      move: x y => [x | y] [z | w] //=; rewrite/le; try exact/lebP.
      by constructor.
      by constructor.
    Qed.
  End Dec.
End Sums.


Section Lists.
  Context {A : Type} `{setoid A}.

  Definition list_eq :=
    (fix list_eq (x y : list A) :=
       match x, y with
         | nil, nil => True
         | lelt::x', relt::y' => lelt ≡ relt /\ nosimpl list_eq x' y'
         | _, _ => False
       end).

  Global Program Instance list_setoid : setoid (list A) := {
    eq_ := list_eq
  }.
  Next Obligation.
    by elim=> //= elt lst Heq; split; first by apply/eq_refl.
  Qed.
  Next Obligation.
    elim=> //=; first by case.
    move=> elt lst IH [] //= elt' lst' [Helt Hlst].
    split; first by apply/eq_sym.
    exact/IH.
  Qed.
  Next Obligation.
    elim=> //=; first by case.
    move=> elt lst IH [] //= elt' lst' lst'' [Helt Hlst].
    case: lst'' => //= elt'' lst'' [Helt' Hlst'].
    split; first by apply/eq_trans; eauto.
    by apply/IH; eauto.
  Qed.

  Global Program Instance list_usual_setoid `{@usual_setoid A _} : @usual_setoid (list A) _ := {
  }.
  Next Obligation.
    move: x y H1; elim=> [|l x IH] [|r y] //= [Hh Ht].
    congr cons; first exact/usual_eqP'.
    exact/IH.
  Qed.
  
  Context `{@poset A _}.
  Definition list_le :=
    (fix list_le (x y : list A) :=
       match x, y with
         | nil, _ => True
         | lelt::x', relt::y' => (lelt ⊑! relt) \/ (lelt ≡ relt /\ nosimpl list_le x' y')
         | _, _ => False
       end).
  
  Global Program Instance list_poset : @poset (list A) _ := {
    le_ := list_le
  }.
  Next Obligation.
    elim; first by case.
    move=> x xs IH [| y ys] //= [Hh Ht].
    right; split; first by [].
    exact/IH.
  Qed.
  Next Obligation.
    elim; first by case.
    move=> x xs IH [| y ys] //= [[Hxy Hxy'] | [Hxy Hxys]] [[Hyx Hyx'] | [Hyx Hyxs]].
    by case Hxy'; exact/le_antisym.
    by case Hxy'; exact/eq_sym.
    by case Hyx'; exact/eq_sym.
    split; first by [].
    exact/IH.
  Qed.
  Next Obligation.
    elim; first by case.
    move=> x xs IH [| y ys] [| z zs] //=.
    move=> [Hxy | [Hxy Hxys]] [Hyz | [Hyz Hyzs]].
    by left; rewrite <- Hyz.
    by left; rewrite <- Hyz.
    by left; rewrite Hxy.
    right; split; first by apply/eq_trans; eauto.
    by apply/IH; eauto.
  Qed.

  Section Pointwises.
    Definition list_le_pointwise_ :=
      (fix list_le (x y : list A) :=
        match x, y with
          | nil, nil => True
          | lelt::x', relt::y' => (lelt ⊑ relt) /\ (nosimpl list_le x' y')
          | _, _ => False
        end).

    Local Program Instance list_poset_pointwise : @poset (list A) _ := {
      le_ := list_le_pointwise_
    }.
    Next Obligation.
      elim=> [|x xs IH] [|y ys] Hxy //.
      inversion Hxy.
      by split; [apply/eq_le | apply/IH].
    Qed.
    Next Obligation.
      elim=> [|x xs IH] [|y ys] Hxy Hyx //.
      inversion Hxy; inversion Hyx.
      by split; [apply/le_antisym | apply/IH].
    Qed.
    Next Obligation.
      elim=> [|x xs IH] [|y ys] [|z zs] Hxy Hyz //.
      inversion Hxy; inversion Hyz.
      by split; [apply/le_trans | apply/IH]; eauto.
    Qed.
  End Pointwises.
  
  Global Program Instance list_hasBot : @hasBot (list A) _ _ := {
    bot_ := nil
  }.

  Global Program Instance list_order `{@order A _ _} : @order (list A) _ _ := {
  }.
  Next Obligation.
    move: x y; elim; first by case; firstorder.
    move=> x xs IH [| y ys]; first by right.
    case: (classic (x ≡ y)) => Hxy; [case: (IH ys) | case: (le_total x y)] => Hxy'.
    by left; right.
    by right; right; split; first by apply/eq_sym.
    by left; left.
    by right; left; split; last by rewrite/neq => Hyx; case Hxy; apply/eq_sym.
  Qed.    

  Section Dec.
    Context `{@setoidDec A _}.
    Definition list_eqb :=
      (fix list_eqb (x y : list A) :=
         match x, y with
           | nil, nil => true
           | lelt::x', relt::y' => andb (eqb lelt relt) (nosimpl list_eqb x' y')
           | _, _ => false
         end).

    Global Program Instance list_setoidDec : @setoidDec (list A) _ := {
      eqb_ := list_eqb
    }.
    Next Obligation.
      move: x y; elim; first by case; constructor.
      move=> x xs IH [| y ys]; first by constructor.
      apply/equivP; first by apply andP.
      by constructor; move=> [Hxy Hxys]; do [split; [apply/eqbP | apply/IH]].
    Qed.

    Context `{@posetDec A _ _ _}.
    Definition list_leb :=
      (fix list_leb (x y : list A) :=
         match x, y with
           | nil, _ => true
           | lelt::x', relt::y' =>
             if ltb lelt relt
             then true
             else if eqb lelt relt
                  then nosimpl list_leb x' y'
                  else false
           | _, _ => false
         end).
  
    Global Program Instance list_posetDec : @posetDec (list A) _ _ _ := {
      leb_ := list_leb
    }.
    Next Obligation.
      move: x y; elim; first by case; constructor.
      move=> x xs IH [| y ys]; first by constructor.
      apply/equivP; first by apply orP.
      constructor; do [move=> [Hxy | Hxys]; first by left; apply/ltP].
      by move/andP: Hxys; right; split; [apply/eqbP | apply/IH]; firstorder.
      by right; apply/andP; split; [apply/eqbP | apply/IH]; firstorder.
    Qed.
  End Dec.
End Lists.


Section Nats.
  Global Program Instance nat_setoid : setoid nat := {
    eq_ := Logic.eq
  }.

  Global Program Instance nat_usual_setoid : @usual_setoid nat _ := {
  }.
  
  Global Program Instance nat_poset : @poset nat _ := {
    le_ x y := x ≤ y
  }.
  Next Obligation.
    move=> x y ->.
    exact/Le.le_refl.
  Qed.
  Next Obligation.
    move=> //= x y Hlr Hrl.
    exact/Le.le_antisym.
  Qed.
  Next Obligation.
    move=> //= x mhs y Hlm Hmr.
    by apply/Le.le_trans; eauto.
  Qed.

  Global Program Instance nat_hasBot : @hasBot nat _ _ := {
    bot_ := 0
  }.
  Next Obligation.
    exact/Le.le_0_n.
  Qed.

  Global Program Instance nat_order : @order nat _ _.
  Next Obligation.
    case: (Lt.le_or_lt x y); first by left.
    by right; apply/Lt.lt_le_weak.
  Qed.

  Global Program Instance nat_hasJoin : @hasJoin nat _ _ := {
    join_ := max
  }.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite max_r.
    by rewrite max_l.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite max_r.
    by rewrite max_l.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite max_r.
    by rewrite max_l.
  Qed.

  Require Import Arith.
  
  Global Program Instance nat_setoidDec : @setoidDec nat _ := {
    eqb_ x y := if eq_nat_dec x y is left _ then true else false
  }.
  Next Obligation.
    by case: (eq_nat_dec x y); constructor.
  Qed.

  Global Program Instance nat_posetDec : @posetDec nat _ _ _ := {
    leb_ x y := if le_dec x y is left _ then true else false
  }.
  Next Obligation.
    by case: (le_dec x y); constructor.
  Qed.
End Nats.


Section Positives.
  Require Import PArith.
  
  Global Program Instance positive_setoid : setoid positive := {
    eq_ := Logic.eq
  }.

  Global Program Instance positive_usual_setoid : @usual_setoid positive _ := {
  }.
  
  Global Program Instance positive_poset : @poset positive _ := {
    le_ x y := Pos.le x y
  }.
  Next Obligation.
    move=> x y ->.
    exact/Pos.le_refl.
  Qed.
  Next Obligation.
    move=> /= x y Hxy Hyx.
    exact/Pos.le_antisym.
  Qed.

  Global Program Instance positive_order : @order positive _ _.
  Next Obligation.
    case Hcmp: (x ?= y)%positive.
    by left; apply/eq_le/Pos.compare_eq_iff.
    by left; apply/Pos.lt_le_incl/Pos.compare_lt_iff.
    by right; apply/Pos.lt_le_incl/Pos.compare_gt_iff.
  Qed.

  Global Program Instance positive_hasBot : @hasBot positive _ _ := {
    bot_ := 1%positive
  }.
  Next Obligation.
    exact/Pos.le_1_l.
  Qed.
  
  Global Program Instance positive_hasJoin : @hasJoin positive _ _ := {
    join_ := Pos.max
  }.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite Pos.max_r.
    rewrite Pos.max_l; last by [].
    exact/eq_le/eq_refl.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    rewrite Pos.max_r; last by [].
    exact/eq_le/eq_refl.
    by rewrite Pos.max_l.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite Pos.max_r.
    by rewrite Pos.max_l.
  Qed.

  Global Program Instance positive_setoidDec : @setoidDec positive _ := {
    eqb_ := Pos.eqb
  }.
  Next Obligation.
    apply/equivP; last by apply Pos.eqb_eq.
    by case: ((x =? y)%positive); constructor.
  Qed.

  Global Program Instance positive_posetDec : @posetDec positive _ _ _ := {
    leb_ := Pos.leb
  }.
  Next Obligation.
    apply/equivP; last by apply Pos.leb_le.
    by case: ((x <=? y)%positive); constructor.
  Qed.
End Positives.

Definition idT := positive.
Definition allocT := idT.
Definition bidT := idT.
Definition fidT := idT.

Definition varT := (idT + idT)%type.
Instance varT_setoid : setoid varT := _.
Instance varT_usual_setoid : @usual_setoid varT _ := _.
Instance varT_setoidDec : @setoidDec varT _ := _.
Instance varT_poset : @poset varT _ := _.
Instance varT_posetDec : @posetDec varT _ _ _ := _. 

Inductive op1 := op_not | op_bnot.

Definition op1_eqb (lhs rhs : op1) :=
  match lhs, rhs with
    | op_not, op_not => true
    | op_bnot, op_bnot => true
    | _, _ => false
  end.

Program Instance op1_setoid : setoid op1 := {}.
Program Instance op1_usual_setoid : @usual_setoid op1 _ := {}.
Program Instance op1_setoidDec : @setoidDec op1 _ := {
  eqb_ := op1_eqb
}.
Next Obligation.
  by case: x; case: y; constructor.
Qed.
  
Inductive op2 := 
| op_add 
| op_sub
| op_mul 
| op_div
| op_mod
| op_lt
| op_le 
| op_gt
| op_ge
| op_eq 
| op_neq 
| op_and 
| op_or 
| op_band 
| op_bor 
| op_bxor 
| op_lshift 
| op_rshift.

Definition op2_eqb (lhs rhs : op2) :=
  match lhs, rhs with
    | op_add, op_add => true
    | op_sub, op_sub => true
    | op_mul, op_mul => true
    | op_div, op_div => true
    | op_mod, op_mod => true
    | op_lt, op_lt => true
    | op_le, op_le => true
    | op_gt, op_gt => true
    | op_ge, op_ge => true
    | op_eq, op_eq => true
    | op_neq, op_neq => true
    | op_and, op_and => true
    | op_or, op_or => true
    | op_band, op_band => true
    | op_bor, op_bor => true
    | op_bxor, op_bxor => true
    | op_lshift, op_lshift => true
    | op_rshift, op_rshift => true
    | _, _ => false
  end.

Program Instance op2_setoid : setoid op2 := {}.
Program Instance op2_usual_setoid : @usual_setoid op2 _ := {}.
Program Instance op2_setoidDec : @setoidDec op2 _ := {
  eqb_ := op2_eqb
}.
Next Obligation.
  by case: x; case: y; constructor.
Qed.

Obligation Tactic := obligation_tactic.
