Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomBridge.

Require Export Containers.Maps.

Set Implicit Arguments.

Section Maps.
  Context (Key Elt : Type).
  Context `{setoidKey : setoid Key,
    posetKey : @poset Key _,
    orderKey : @DomType.order Key _ _,
    setoidDecKey : @setoidDec Key _,
    posetDecKey : @posetDec Key _ _ _}.
  
  Definition map_forall (test : Key -> Elt -> bool) (m : Map[Key, Elt]) : bool :=
    fold
      (fun k v acc => acc && (test k v))
      m
      true.                       

  Lemma map_forallP'
        (test : Key -> Elt -> bool)
        (Hproper : Proper (eq ==> Logic.eq ==> eq) test)
        (kes : list (Key * Elt))
        (HNoDup : NoDupA MF.K.eqk kes) :
    reflect
      (forall key, if findA (MF.eqb key) kes is Some elt then (test key elt):Prop else True)
      (fold_right (fun ke acc => acc && test ke.1 ke.2) true kes).
  Proof.
    elim: kes HNoDup => [| [key elt] kes IH] HNoDup //=.
    constructor=> //.
    apply/equivP; first apply andP.
    rewrite/MF.eqb/is_compare/compare/_cmp /=.
    inversion HNoDup; subst; clear HNoDup.
    split.
    { move=> [IHfold Htest] key'.
      case Hk'k: (leb key' key); move: Hk'k; move/lebP => Hk'k;
      case Hkk': (leb key key'); move: Hkk'; move/lebP => Hkk' //; try exact/IH.
      erewrite -> Hproper; eauto.
      exact/le_antisym.
    }
    { move=> IHeq; split.
      { apply/IH => //.
        move=> key'; have := (IHeq key').
        case Hk'k: (leb key' key); move: Hk'k; move/lebP => Hk'k;
        case Hkk': (leb key key'); move: Hkk'; move/lebP => Hkk' //.
        move=> Htest.
        case HfindA: (findA (MF.eqb key') kes) => //.
        contradict H1.
        rewrite <- MF.find_NoDup in HfindA => //.
        eapply MF.InA_eqke_eqk with (k1:=key'); eauto.
        exact/le_antisym.
      }
      { have := (IHeq key).
        case Hkk: (leb key key); move: Hkk; move/lebP => Hkk //.
        contradict Hkk; apply/DomType.eq_le/eq_refl.
      }
    }
  Qed.

  Lemma map_forallP
        (test : Key -> Elt -> bool)
        (Hproper : Proper (eq ==> Logic.eq ==> eq) test)
        (m : Map[Key, Elt]) :
    reflect (forall key elt (HMapsTo : MapsTo key elt m), test key elt) (map_forall test m).
  Proof.
    rewrite /map_forall fold_1; rewrite <- fold_left_rev_right.
    apply/equivP; first apply map_forallP' => //;
      first by apply NoDupA_rev; [apply MF.K.eqk_Equiv | apply elements_3w].
    apply forall_iff => key.
    rewrite MF.findA_rev; last by apply NoDupA_rev; [apply MF.K.eqk_Equiv | apply elements_3w].
    rewrite rev_involutive.
    rewrite <- MF.elements_o.
    split=> H.
    { move=> elt HMapsTo.
      apply MF.find_mapsto_iff in HMapsTo.
      by rewrite -> HMapsTo in H.
    }
    { case HMapsTo: m[key] => //.
      apply: H.
      by apply MF.find_mapsto_iff.
    }
  Qed.

  Definition map_elt_compare
             (test : option Elt -> option Elt -> bool)
             (HNone : test None None)
             (x y : Map[Key, Elt]) : bool :=
    is_empty
      (map2
         (fun x y => if test x y then None else Some tt)
         x
         y).

  Lemma map_elt_compareP
        (test : option Elt -> option Elt -> bool)
        (HNone : test None None)
        (x y : Map[Key, Elt]) :
    reflect (forall key, test x[key] y[key])
            (map_elt_compare test HNone x y).
  Proof.
    rewrite/map_elt_compare.
    apply/equivP; first apply idP.
    rewrite <- is_true_iff.
    rewrite <- MF.is_empty_iff.
    split=> Heq key.
    { apply/negbNE/negP.
      have := (Heq key tt) => Hc; contradict Hc.
      apply MF.find_mapsto_iff.
      rewrite -> MF.map2_1bis; last by move: HNone; case: (test None None).
      by case Ht: (test x[key] y[key]) Hc.
    }
    { move=> tt Hc.
      apply MF.find_mapsto_iff in Hc.
      rewrite -> MF.map2_1bis in Hc; last by move: HNone; case: (test None None).
      case Ht: (test x[key] y[key]) Hc => //.
      contradict Ht.
      exact/Bool.not_false_iff_true/Heq.
    }                                    
  Qed.
End Maps.

Section DomMaps.
  Context (Key Elt : Type).
  Context `{setoidKey : setoid Key,
    posetKey : @poset Key _,
    orderKey : @DomType.order Key _ _,
    setoidDecKey : @setoidDec Key _,
    posetDecKey : @posetDec Key _ _ _}.
  Context `{setoidElt : setoid Elt,
    posetElt : poset Elt,
    hasBotElt : hasBot Elt}.

  Notation "'Map!' '[' A ',' B ']'" := (@dict A order_OrderedType _ B) (at level 0, no associativity).

  Definition coerce (oe : option Elt) := if oe is Some e then e else bot.
  Global Add Parametric Morphism : coerce with
      signature le ==> le as coerce_mor.
  Proof.
    move=> [x|] [y|] Hle //=.
    exact/bot_prop.
    exact/bot_prop.
  Qed.

  Global Add Parametric Morphism : coerce with
      signature eq ==> eq as coerce_eq_mor.
  Proof.
    intros. apply le_antisym.
    apply coerce_mor; apply DomType.eq_le; auto.
    apply coerce_mor; apply DomType.eq_le; apply eq_sym; auto.
  Qed.

  Global Program Instance map_setoid : setoid Map![Key, Elt] := {
    eq_ x y :=
      forall key, coerce x[key] ≡ coerce y[key]
  }.
  Next Obligation.
    by move=> m key; apply/eq_refl.
  Qed.
  Next Obligation.
    move=> //= x y Heq key.
    exact/eq_sym/Heq.
  Qed.
  Next Obligation.
    move=> //= x mhs y Hlm Hmr key.
    by rewrite -> Hlm.
  Qed.

  Global Program Instance map_poset : @poset Map![Key, Elt] _ := {
    le_ x y := forall key, coerce x[key] ⊑ coerce y[key]
  }.
  Next Obligation.
    move=> x y Heq key.
    exact/DomType.eq_le.
  Qed.
  Next Obligation.
    move=> //= x y Hlr Hrl key.
    by apply/le_antisym; [apply Hlr | apply Hrl].
  Qed.
  Next Obligation.
    move=> //= x mhs y Hlm Hmr key.
    by rewrite -> Hlm.
  Qed.

  Global Program Instance map_maplike : 
    maplike (Key:=Key) (Elt:=Elt) (M:=Map![Key, Elt]) := {
    mlookup key m := coerce m[key];
    madd := add;
    mremove := remove
  }.
  Next Obligation.
    move=> x y Hxy z w Hzw.
    rewrite -> (Hzw x).
    rewrite/coerce.
    erewrite -> (MF.find_o w); last by eauto.
    apply DomType.eq_le; exact/eq_refl.
  Qed.
  Next Obligation.
    move=> x y Hxy z w Hzw a b Hab c.
    rewrite/coerce.
    rewrite 2! MF.add_o.
    rewrite/is_compare/compare/_cmp /=.
    do 2! rewrite -> Hxy.
    by case: (leb y c); case: (leb c y) => //; apply/Hab.
  Qed.
  Next Obligation.
    rewrite/coerce.
    move=> x y Hxy z w Hzw c.
    rewrite 2! MF.remove_o.
    rewrite/is_compare/compare/_cmp /=.
    do 2! rewrite -> Hxy.
    case: (leb y c); case: (leb c y) => //.
    apply DomType.eq_le; exact/eq_refl.
  Qed.

  Lemma map_mlookup_spec (key : Key) (m : Map![Key, Elt]) :
    coerce m[key] = m[key]!.
  Proof. done. Qed.

  Lemma map_madd_spec (key : Key) (elt : Elt) (m : Map![Key, Elt]) (key' : Key) :
    eq (m[key <- elt]!)[key']! (if eqb key key' then elt else m[key']!).
  Proof.
    rewrite/mlookup/madd/map_maplike/coerce.
    rewrite MF.add_o is_compare_eqb.
    by case: (eqb key key'); apply: eq_refl.
  Qed.

  Global Program Instance map_hasBot : @hasBot Map![Key, Elt] _ _ := {
    bot_ := empty _
  }.
  Next Obligation.
    move=> key; rewrite/mlookup /= /coerce.
    rewrite MF.empty_o //=.
    exact/bot_prop.
  Qed.

  Context `{setoidDecElt : @setoidDec Elt _}.
  Global Program Instance map_setoidDec : @setoidDec Map![Key, Elt] _ := {
    eqb_ := map_elt_compare (fun x y => eqb (coerce x) (coerce y)) _
  }.
  Next Obligation. exact/eqbP/eq_refl. Qed.
  Next Obligation.
    apply/equivP; first apply map_elt_compareP.
    apply/forall_iff => key.
    apply iff_sym; apply rwP; apply eqbP.
  Qed.

  Context `{posetDecElt : @posetDec Elt _ _ _}.
  Global Program Instance map_posetDec : @posetDec Map![Key, Elt] _ _ _ := {
    leb_ := map_elt_compare (fun x y => leb (coerce x) (coerce y)) _
  }.
  Next Obligation. exact/lebP/DomType.eq_le/eq_refl. Qed.
  Next Obligation.
    apply/equivP; first apply map_elt_compareP.
    apply/forall_iff => key.
    apply iff_sym; apply rwP; apply lebP.
  Qed.

  Global Program Instance map_hasJoin `{hasJoinElt : @hasJoin Elt _ _} : @hasJoin Map![Key, Elt] _ _ := {
    join_ x y :=
      map2
        (fun x y =>
           match x, y with
             | Some x', Some y' => Some (join x' y')
             | None, _ => y
             | _, None => x
           end)
        x y
  }.
  Next Obligation.
    move=> key; rewrite/mlookup /= /coerce.
    rewrite -> MF.map2_1bis => //.
    case Hx: (x[key]); last exact/bot_prop.
    case Hy: (y[key]); last exact/DomType.eq_le/eq_refl.
    exact/join_lhs.
  Qed.
  Next Obligation.
    move=> key; rewrite/mlookup /= /coerce.
    rewrite -> MF.map2_1bis => //.
    case Hy: (y[key]); last exact/bot_prop.
    case Hx: (x[key]); last exact/DomType.eq_le/eq_refl.
    exact/join_rhs.
  Qed.
  Next Obligation.
    move=> key.
    move: (Hx key) (Hy key) => {Hx Hy}.
    rewrite/mlookup /= /coerce.
    rewrite -> MF.map2_1bis => //.
    case Hx: (x[key]); case Hy: (y[key]) => //.
    move=> Hx' Hy'; exact/join_lub.
  Qed.

  Lemma if_madd (test:bool) (x y:Map![Key,Elt]) (k:Key) (v:Elt) :
    (if test then x else y)[k <- v]! = (if test then x[k <- v]! else y[k <- v]!).
  Proof. by case: test. Qed.

  Lemma if_mlookup (test:bool) (x y:Map![Key,Elt]) (k:Key) :
    (if test then x else y)[k]! = (if test then x[k]! else y[k]!).
  Proof. by case: test. Qed.

  Lemma map_join_spec `{hasJoinElt : @hasJoin Elt _ _} (m n : Map![Key, Elt]) (key : Key) :
    eq ((join m n)[key]!) (join m[key]! n[key]!).
  Proof.
    rewrite {1}/join/join_/map_hasJoin.
    rewrite/mlookup/map_maplike/coerce.
    rewrite MF.map2_1bis; last done.
    case: (m[key])=> [melt|]; case: (n[key])=> [nelt|].
    exact: eq_refl.
    by rewrite join_bot; apply: eq_refl.
    by rewrite join_commutative join_bot; apply: eq_refl.
    by rewrite join_bot; apply: eq_refl.
  Qed.

  Lemma withtop_map_join_spec `{hasJoinElt : @hasJoin Elt _ _, 
    hasTopElt: @hasTop Elt _ _} (m n : withtop Map![Key, Elt]) (key : Key) :
    eq ((join m n)[key]!) (join m[key]! n[key]!).
  Proof.
    rewrite {1}/join/join_/withtop_hasJoin/withtop_join.
    case: m. move => a. case: n. move => b.
    rewrite/mlookup/withtop_maplike; by apply:map_join_spec.
    rewrite/mlookup/withtop_maplike. rewrite join_top_rhs. reflexivity.
    rewrite/mlookup/withtop_maplike. rewrite join_top_lhs. reflexivity.
  Qed.

  Lemma map_join_update_spec `{hasJoinElt : @hasJoin Elt _ _} (m n : Map![Key, Elt]) (key : Key) (elt : Elt) :
    eq (join m n)[key <- elt]! (join m[key <- elt]! n[key <- elt]!).
  Proof.
    move=> key'.
    rewrite 2! map_mlookup_spec.
    rewrite map_madd_spec.
    rewrite [X in eq _ X]map_join_spec.
    rewrite 2! map_madd_spec.
    case: (eqb key key').
    exact/eq_sym/join_convolutive.
    exact: map_join_spec.
  Qed.

  Lemma smem_proper_keys: forall (keys: \powerset Key),
    Proper (_eq ==> Logic.eq ==> Logic.eq)
    (λ (key : Key) (_ : Elt), smem (Elt:=Key) key keys).
  Proof.
    intros keys kx ky Hk _ _ _.
    unfold _eq in Hk. unfold order_OrderedType in Hk.
    rewrite Hk. reflexivity.
  Qed.

  Lemma filter_preserve_not_in: forall (loc:Key) (s: \powerset Key) m,
    ¬In loc m ->
    ¬In loc (MF.filter (λ (key : Key) (_ : Elt), key !\in s) m).
  Proof.
    move => loc s m Hin.
    intro Hin_filter. elim Hin.
    apply MF.mem_in_iff in Hin_filter.
    rewrite MF.mem_find_b in Hin_filter.
    remember ((MF.filter (λ (key : Key) (_ : Elt), key !\in s) m) [loc]) as 
      v; destruct v as [v|]; last by inversion Hin_filter.
    symmetry in Heqv. apply MF.find_mapsto_iff in Heqv.
    apply MF.filter_iff in Heqv; last by apply smem_proper_keys.
    destruct Heqv as [Hmapsto _].
    apply MF.find_mapsto_iff in Hmapsto.
    apply MF.in_find_iff; rewrite Hmapsto; discriminate.
  Qed.

  Lemma filter_preserve_in: forall (loc:Key) (s: \powerset Key) m,
    In loc m ->
    loc !\in s ->
    In loc (MF.filter (λ (key : Key) (_ : Elt), key !\in s) m).
  Proof.
    intros loc s m Hinm Hins.
    apply MF.in_find_iff in Hinm. apply MF.in_find_iff.
    remember ((m) [loc]) as v; destruct v as [v|]; last tauto.
    assert (Hfilter:
      (MF.filter (λ (key : Key) (_ : Elt), key !\in s) m) [loc] = Some v).
    {
      apply MF.find_mapsto_iff.
      apply MF.filter_iff; first by apply smem_proper_keys.
      split; last by auto.
      symmetry in Heqv; apply MF.find_mapsto_iff; by auto.
    }
    rewrite Hfilter. tauto.
  Qed.

  Context `{hasJoinElt : @hasJoin Elt _ _}.

  Definition map_mrestrict (s : \powerset Key) (m : Map![Key, Elt]) := MF.filter (fun key _ => key !\in s) m.

  Lemma map_mrestrict_spec' 
    (s : \powerset Key) (m : Map![Key, Elt]) (key : Key) :
    eq (map_mrestrict s m)[key]! (if key !\in s then m[key]! else bot).
  Proof.
    rewrite/map_mrestrict.
    rewrite/mlookup/map_maplike/coerce.
    case Hlookup: ((MF.filter (λ (key0 : Key) (_ : Elt), key0 !\in s) m) [key])=> [elt|].
    { apply MF.find_mapsto_iff in Hlookup.
      apply MF.filter_iff in Hlookup.
      move: Hlookup=> [Hmapsto Hs].
      rewrite -> Hs.
      apply MF.find_mapsto_iff in Hmapsto; rewrite -> Hmapsto.
      exact: eq_refl.
      move=> x y Hxy z w Hzw; subst.
      by rewrite -> Hxy.
    }
    { apply MF.not_find_in_iff in Hlookup.
      case Hs: (key !\in s); last exact: eq_refl.
      case Hlookup': m[key]=> [elt|]; last exact: eq_refl.
      contradict Hlookup.
      apply: filter_preserve_in; last done.
      apply MF.in_find_iff.
      by rewrite -> Hlookup'.
    }
  Qed.
  
  Lemma mlookup_fold : forall (m:Map![Key,Elt]) (k:Key),
    coerce m [k] = m [k]!.
  Proof. intros. unfold mlookup, map_maplike; auto. Qed.

  Global Instance map_restricting_maplike :
    @restricting_maplike Key Elt Map![Key, Elt] _ _ _ _ _ _ _ _ := {
    mrestrict := map_mrestrict
  }.
  Proof.
    { move=> x y Hxy z w Hzw a.
      rewrite 2! mlookup_fold.
      rewrite 2! map_mrestrict_spec'.
      rewrite -> Hxy.
      case: (a !\in y); last exact/DomType.eq_le/eq_refl.
      exact: Hzw.
    }
    { move=> s m loc Hloc.
      rewrite map_mrestrict_spec' Hloc.
      exact: eq_refl.
    }
    { move=> s m loc v Hloc loc'.
      rewrite 2! map_mlookup_spec.
      rewrite map_madd_spec.
      rewrite map_mrestrict_spec'.
      case Hloc': (loc' !\in s).
      rewrite map_madd_spec.
      case: (eqb loc loc'); first exact: eq_refl.
      by rewrite map_mrestrict_spec' Hloc'; apply: eq_refl.
      case Heqb: (eqb loc loc').
      move/eqbP in Heqb; rewrite -> Heqb in Hloc.
      by rewrite -> Hloc in Hloc'.
      by rewrite map_mrestrict_spec' Hloc'; apply: eq_refl.
    }
    { move=> l m loc.
      rewrite 2! map_mlookup_spec.
      case Hloc: (loc !\in l).
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      exact: eq_refl.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      exact: eq_refl.
    }
    { move=> l l' m loc.
      rewrite 2! map_mlookup_spec.
      case Hloc: (loc !\in l); case Hloc': (loc !\in l');
      do ? [rewrite map_mrestrict_spec';
        try rewrite -> Hloc;
          try rewrite -> Hloc'];
      try exact: eq_refl.
    }
    { move=> l loc.
      rewrite map_mlookup_spec.
      exact: eq_refl.
    }
    { move=> l l' m ll' loc.
      rewrite 2! map_mlookup_spec.
      rewrite map_mrestrict_spec'.
      rewrite [X in eq _ X]map_mrestrict_spec'.
      case Hloc: (loc !\in l).
      rewrite ll'; auto.
      rewrite map_mrestrict_spec' Hloc.
      exact: eq_refl.
      case Hloc': (loc !\in l'); last exact: eq_refl.
      by rewrite map_mrestrict_spec' Hloc; apply: eq_refl.
    }
    { move=> l l' m ll' loc.
      rewrite 2! mlookup_fold.
      rewrite 2! map_mrestrict_spec'.
      case Hloc: (loc !\in l).
      rewrite ll'; auto.
      exact/DomType.eq_le/eq_refl.
      exact: bot_prop.
    }
    { move=> l m loc.
      rewrite mlookup_fold.
      rewrite map_mrestrict_spec'.
      case: (loc !\in l); first exact/DomType.eq_le/eq_refl.
      exact: bot_prop.
    }
  Defined.
  
  Global Instance map_restricting_join_maplike :
    @restricting_join_maplike Key Elt Map![Key, Elt] _ _ _ _ _ _ _ _ _ _ := {
  }.
  Proof.
    { move=> l m m' loc.
      rewrite 2! map_mlookup_spec.
      case Hloc: (loc !\in l).
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite 2! map_join_spec.
      rewrite 2! map_mrestrict_spec'; rewrite -> Hloc.
      exact: eq_refl.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite map_join_spec.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite map_mrestrict_spec'; rewrite -> Hloc.
      rewrite join_convolutive; reflexivity.
    }
    { move=> l m loc.
      rewrite map_mlookup_spec.
      rewrite map_join_spec.
      rewrite 2! map_mrestrict_spec'.
      case Hloc: (loc !\in l).
      rewrite powerset_neg_spec Hloc join_bot.
      exact: eq_refl.
      rewrite powerset_neg_spec Hloc join_commutative join_bot.
      exact: eq_refl.
    }
    { move=> l l' m loc.
      rewrite 2! map_mlookup_spec.
      rewrite map_join_spec.
      rewrite 3! map_mrestrict_spec'.
      case Hloc: (loc !\in l).
      { rewrite union_lhs; auto.
        case Hloc': (loc !\in l').
        by rewrite join_convolutive; reflexivity.
        by rewrite join_bot; reflexivity.
      }
      { case Hloc': (loc !\in l').
        rewrite union_rhs; auto.
        by rewrite join_commutative join_bot; reflexivity.
        move: l l' Hloc Hloc'=> [l lp] [l' lp'] Hloc Hloc'.
        rewrite/union/=/proj1_sig in Hloc Hloc' *.
        rewrite Hloc Hloc' /=.
        by rewrite join_convolutive; reflexivity.
      }
    }
    { move=> s m n loc v.
      rewrite/mrestrict_join.
      rewrite map_join_update_spec.
      case Hloc: (loc !\in s); rewrite/negb.
      { rewrite mrestrict_update_spec; last done.
        move=> loc'.
        rewrite 2! map_mlookup_spec.
        rewrite 2! map_join_spec.
        rewrite 2! map_madd_spec.
        case Heqb: (eqb loc loc'); last exact: eq_refl.
        rewrite map_mrestrict_spec' powerset_neg_spec.
        move/eqbP in Heqb; rewrite <- Heqb; rewrite Hloc; rewrite/negb.
        rewrite join_bot.
        exact/join_convolutive.
      }
      { rewrite mrestrict_update_spec; last by rewrite powerset_neg_spec Hloc.
        move=> loc'.
        rewrite 2! map_mlookup_spec.
        rewrite 2! map_join_spec.
        rewrite 2! map_madd_spec.
        case Heqb: (eqb loc loc'); last exact: eq_refl.
        rewrite map_mrestrict_spec'.
        move/eqbP in Heqb; rewrite <- Heqb; rewrite Hloc; rewrite/negb.
        rewrite join_commutative join_bot.
        exact/join_convolutive.
      }
    }
    { move=> l l' m m' loc.
      rewrite 2! map_mlookup_spec /mrestrict_join.
      rewrite 2! map_join_spec.
      rewrite 4! map_mrestrict_spec'.
      rewrite 2! powerset_neg_spec.
      case Hloc: (loc !\in l); rewrite/negb.
      { rewrite union_lhs; auto.
        exact: eq_refl.
      }
      { rewrite map_join_spec.
        case Hloc': (loc !\in l'); rewrite /negb.
        rewrite union_rhs; auto.
        rewrite 2! map_mrestrict_spec'.
        rewrite powerset_neg_spec Hloc' /negb.
        rewrite join_commutative join_bot.
        reflexivity.
        rewrite 2! map_mrestrict_spec'.
        rewrite powerset_neg_spec Hloc' /negb.
        move: l l' Hloc Hloc'=> [l lp] [l' lp'] Hloc Hloc'.
        rewrite/union/=/proj1_sig in Hloc Hloc' *.
        rewrite Hloc Hloc' /=.
        rewrite join_commutative join_bot.
        reflexivity.
      }
    }
  Defined.

  Lemma map_mrestrict_spec
    (s : \powerset Key) (m : Map![Key, Elt]) (key : Key) :
    eq (mrestrict s m)[key]! (if key !\in s then m[key]! else bot).
  Proof. exact: map_mrestrict_spec'. Qed.

  Lemma map_mrestrict_madd_spec
    (s : \powerset Key) (m : Map![Key, Elt])
    (key : Key) (Hkey : key !\in s) (elt : Elt) :
    (mrestrict s m) [key <- elt]! ≡ mrestrict s (m) [key <- elt]!.
  Proof.
    move=> key'; rewrite ? map_mlookup_spec map_madd_spec.
    rewrite [X in eq _  X]map_mrestrict_spec.
    case Heqb: (eqb key key').
    { move/eqbP in Heqb.
      rewrite <- Heqb; rewrite Hkey map_madd_spec Heqb eqb_refl.
      exact: eq_refl.
    }
    { rewrite map_mrestrict_spec.
      case: (_ !\in _).
      rewrite map_madd_spec Heqb.
      exact: eq_refl.
      exact: eq_refl.
    }
  Qed.
End DomMaps.
Notation "'Map!' '[' A ',' B ']'" := (@dict A order_OrderedType _ B) (at level 0, no associativity).
