Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Basics.
Require Import Setoid.
Require Import Classes.Equivalence.
Require Import Classes.Morphisms.

Require Import LibTactics.

Set Implicit Arguments.

Class setoid A : Type := Setoid {
  eq_ : ∀ (x y : A), Prop;
  eq_refl_ : Reflexive eq_;
  eq_sym_ : Symmetric eq_;
  eq_trans_ : Transitive eq_
}.
Definition eq {A} `{setoid A} := nosimpl (@eq_ _ _).
Notation "x '≡' y" := (eq x y) (at level 70).
Definition neq {A} `{setoid A} x y := ~ (x ≡ y).
Notation "x '!≡' y" := (neq x y) (at level 70).

Lemma eq_refl {A} `{setoid A} : Reflexive eq.
Proof. exact/eq_refl_. Qed.
Lemma eq_sym {A} `{setoid A} : Symmetric eq.
Proof. exact/eq_sym_. Qed.
Lemma eq_trans {A} `{setoid A} : Transitive eq.
Proof. exact/eq_trans_. Qed.

Add Parametric Relation {A} `{setoid A} : A eq
  reflexivity proved by eq_refl
  symmetry proved by eq_sym
  transitivity proved by eq_trans
  as setoid_equiv.
Add Parametric Morphism {A} `{setoid A} : neq with
  signature eq ==> eq ==> iff as neq_mor.
Proof.
  move=> x y Hxy z w Hzw; constructor=> Hneq Heq; case Hneq.
  by rewrite -> Hxy; rewrite -> Heq; apply/eq_sym.
  by rewrite <- Hxy; rewrite -> Heq.
Qed.

Ltac solve_eq :=
  repeat
    (unfold Reflexive, Symmetric, Transitive;
     try match goal with
           | [ H: eq ?a ?b |- _ ] =>
             match goal with
               | [ H: eq b a |- _ ] => fail 1
               | [ |- _ ] => (generalize (eq_sym a b H); intro)
             end
           | [ H: eq ?a ?b, H': eq ?b ?c |- _ ] =>
             match goal with
               | [ |- eq a c] => (eapply eq_trans; eauto)
               | [ H: eq a c |- _ ] => fail 1
               | [ |- _ ] => (generalize (eq_trans a b c H H'); intro)
             end
           | [ |- eq ?a ?a ] => apply eq_refl
           | [ H: ~ eq ?a ?a |- _ ] => contradict H
         end; auto).

Ltac solve_eq_try := try (solve_eq; fail).

Class usual_setoid A `{setoid A} : Type := UsualSetoid {
  usual_eqP' : forall (x y : A), (x ≡ y) -> (x = y)
}.

Definition usual_eqP A `{setoid A, @usual_setoid A _} :
  forall (x y : A), (x ≡ y) <-> (x = y).
Proof.
  split=> Heq; first exact/usual_eqP'.
  by subst; apply/eq_refl.
Qed.

Program Instance usual_eq_eq_subrelation {A} `{usual_setoid A} : subrelation eq Logic.eq.
Next Obligation.
  move=> x y Hxy.
  exact/usual_eqP.
Qed.

Ltac usual_tac H :=
  apply usual_eqP in H; [subst | auto with typeclass_instances].

Class poset A `{setoid A} : Type := Poset {
  le_ : ∀ (x y : A), Prop;
  eq_le_ : subrelation eq le_;
  le_antisym_ : Antisymmetric A eq le_;
  le_trans_ : Transitive le_
}.
Definition le {A} `{poset A} := nosimpl (@le_ _ _ _).
Notation "x '⊑' y" := (le x y) (at level 70).
Definition lt {A} `{poset A} := nosimpl (fun (x y : A) => x ⊑ y /\ x !≡ y).
Notation "x '⊑!' y" := (lt x y) (at level 70).

Lemma eq_le {A} `{poset A} : subrelation eq le.
Proof. exact/eq_le_. Qed.
Lemma le_antisym {A} `{poset A} : Antisymmetric A eq le.
Proof. exact/le_antisym_. Qed.
Lemma le_trans {A} `{poset A} : Transitive le.
Proof. exact/le_trans_. Qed.

Program Instance eq_le_subrelation {A} `{poset A} : subrelation eq le.
Next Obligation.
  move=> x y.
  exact/eq_le.
Qed.
Program Instance eq_inverse_le_subrelation {A} `{poset A} : subrelation eq (inverse le).
Next Obligation.
  rewrite /flip => x y Heq.
  apply/eq_le.
  exact/eq_sym.
Qed.
Add Parametric Morphism {A} `{poset A} : le with
  signature le --> le ++> impl as le_mor.
Proof.
  move=> x y Hyx z w Hzw Hxz.
  apply/le_trans; first by eauto.
  apply/le_trans; first by eauto.
  done.
Qed.
Add Parametric Morphism {A} `{poset A} : le with
  signature eq ==> eq ==> iff as le_mor_eq.
Proof.
  move=> x y Hxy z w Hzw; constructor=> Hle.
  { apply le_trans with x; first by apply/eq_le/eq_sym.
    apply le_trans with z; first by [].
    exact/eq_le.
  }
  { apply le_trans with y; first by apply/eq_le.
    apply le_trans with w; first by [].
    exact/eq_le/eq_sym.
  }
Qed.

Program Instance lt_le_subrelation {A} `{poset A} : subrelation lt le.
Next Obligation.
  by rewrite/lt => x y [Hle Hneq].
Qed.
Add Parametric Morphism {A} `{poset A} : lt with
  signature le --> le ++> impl as lt_mor.
Proof.
  move=> x y Hyx z w Hzw [Hxz Hxz'].
  split; first by rewrite -> Hyx; rewrite -> Hxz.
  move=> Hyw; case Hxz'.
  apply/le_antisym; first by [].
  by rewrite -> Hzw; rewrite <- Hyw.
Qed.
Add Parametric Morphism {A} `{poset A} : lt with
  signature eq ==> eq ==> iff as lt_mor_eq.
Proof.
  move=> x y Hxy z w Hzw.
  constructor=> Hlt.
  by split; rewrite <- Hxy; rewrite <- Hzw; firstorder.
  by split; rewrite -> Hxy; rewrite -> Hzw; firstorder.
Qed.

Class hasBot A `{poset A} : Type := HasBot {
  bot_ : A;
  bot_prop_ : ∀ x, bot_ ⊑ x
}.
Definition bot {A} `{hasBot A} := nosimpl bot_.
Lemma bot_prop A `{hasBot A} : ∀ x, bot ⊑ x.
Proof. exact: bot_prop_. Qed.

Class hasTop A `{poset A} : Type := HasTop {
  top_ : A;
  top_prop_ : ∀ x, x ⊑ top_
}.
Definition top {A} `{hasTop A} := nosimpl top_.
Lemma top_prop A `{hasTop A} : ∀ x, x ⊑ top.
Proof. exact: top_prop_. Qed.

Class order A `{poset A} : Type := Order {
  le_total : ∀ (x y : A), x ⊑ y \/ y ⊑ x
}.

Class hasJoin A `{poset A} : Type := HasJoin {
  join_ : forall (x y : A), A;
  join_lhs_ : ∀ x y, x ⊑ join_ x y;
  join_rhs_ : ∀ x y, y ⊑ join_ x y;
  join_lub_ : ∀ x y ub (Hx : x ⊑ ub) (Hy : y ⊑ ub), (join_ x y) ⊑ ub
}.
Definition join {A} `{hasJoin A} := nosimpl (@join_ _ _ _ _).
Notation "x '⊔' y" := (join x y) (at level 69).
Lemma join_lhs A `{hasJoin A} : ∀ x y, x ⊑ join_ x y.
Proof. exact: join_lhs_. Qed.
Lemma join_rhs A `{hasJoin A} : ∀ x y, y ⊑ join_ x y.
Proof. exact: join_rhs_. Qed.
Lemma join_lub A `{hasJoin A} : ∀ x y ub (Hx : x ⊑ ub) (Hy : y ⊑ ub), (join x y) ⊑ ub.
Proof. exact: join_lub_. Qed.

Add Parametric Morphism {A} `{hasJoin A} : join with
  signature le ==> le ==> le as join_le_mor.
Proof.
  move=> x y Hxy z w Hzw; apply/join_lub; (apply/le_trans; first by eauto).
  exact/join_lhs.
  exact/join_rhs.
Qed.
Add Parametric Morphism {A} `{hasJoin A} : join with
  signature eq ==> eq ==> eq as join_mor.
Proof.
  move=> x y Hxy z w Hzw; apply/le_antisym.
  by rewrite -> Hxy; rewrite -> Hzw; apply/eq_le/eq_refl.
  by rewrite <- Hxy; rewrite <- Hzw; apply/eq_le/eq_refl.
Qed.

Lemma join_bot A {setoidA:setoid A} {posetA:@poset A setoidA} 
  {hasBotA: @hasBot A setoidA posetA} {hasJoinA:@hasJoin A setoidA posetA}: 
  forall (m:A), m ⊔ bot ≡ m.
Proof.
  intros. apply le_antisym.
  {
    apply join_lub.
    apply eq_le; reflexivity.
    apply bot_prop.
  }
  { apply join_lhs. }
Qed.

Lemma join_distributive A `{hasJoin A} : 
  forall (a b c:A), (a ⊔ b) ⊔ c ≡ (a ⊔ c) ⊔ (b ⊔ c).
Proof.
  intros. apply le_antisym.
  {
    repeat apply join_lub.
    { 
      apply le_trans with (a ⊔ c).
      by apply join_lhs.
      by apply join_lhs.
    }
    { 
      apply le_trans with (b ⊔ c).
      by apply join_lhs.
      by apply join_rhs.
    }
    { 
      apply le_trans with (b ⊔ c).
      by apply join_rhs.
      by apply join_rhs.
    }
  }        
  {
    repeat apply join_lub.
    { 
      apply le_trans with (a ⊔ b).
      by apply join_lhs.
      by apply join_lhs.
    }
    { by apply join_rhs. }
    { 
      apply le_trans with (a ⊔ b).
      by apply join_rhs.
      by apply join_lhs.
    }
    { by apply join_rhs. }
  }  
Qed.

Lemma join_associative A `{hasJoin A} : forall (a b c:A), (a ⊔ b) ⊔ c ≡ a ⊔ (b ⊔ c).
Proof.
  intros. apply le_antisym.
  {
    repeat apply join_lub.
    { by apply join_lhs. }
    {
      apply le_trans with (b ⊔ c).
      by apply join_lhs.
      by apply join_rhs.
    }
    {
      apply le_trans with (b ⊔ c).
      by apply join_rhs.
      by apply join_rhs.
    }
  }
  {
    repeat apply join_lub.
    {
      apply le_trans with (a ⊔ b).
      by apply join_lhs.
      by apply join_lhs.
    }
    {
      apply le_trans with (a ⊔ b).
      by apply join_rhs.
      by apply join_lhs.
    }
    { by apply join_rhs. }
  }
Qed.

Lemma join_commutative A `{hasJoin A} : forall (a b:A), a ⊔ b ≡ b ⊔ a.
Proof.
  intros. apply le_antisym.
  {
    repeat apply join_lub.
    by apply join_rhs.
    by apply join_lhs.
  }
  {
    repeat apply join_lub.
    by apply join_rhs.
    by apply join_lhs.
  }
Qed.

Lemma join_convolutive A `{hasJoin A} (a:A) : a ⊔ a ≡ a.
Proof.
  apply: le_antisym.
  by apply: join_lub; apply/eq_le/eq_refl.
  exact: join_lhs.
Qed.

Lemma join_top_lhs A `{setoid A, @poset A _, @hasJoin A _ _, @hasTop A _ _} (a : A) :
  eq (join top a) top.
Proof.
  apply/le_antisym.
  { by apply/join_lub; apply: top_prop. }
  { exact: join_lhs. }
Qed.

Lemma join_top_rhs A `{setoid A, @poset A _, @hasJoin A _ _, @hasTop A _ _} (a : A) :
  eq (join a top) top.
Proof.
  apply/le_antisym.
  { by apply/join_lub; apply: top_prop. }
  { exact: join_rhs. }
Qed.

Class setoidDec A `{setoid A} : Type := SetoidDec {
  eqb_ : ∀ (x y : A), bool;
  eqbP_ : ∀ x y, reflect (x ≡ y) (eqb_ x y)
}.

Definition eqb {A} `{setoidDec A} := nosimpl (@eqb_ _ _ _).
Lemma eqbP {A} `{setoidDec A} : ∀ x y, reflect (x ≡ y) (eqb x y).
Proof. exact: eqbP_. Qed.

Lemma eqb_refl A `{setoid A, @setoidDec A _} (a : A) : eqb a a = true.
Proof.
  exact/eqbP/eq_refl.
Qed.

Lemma eqb_sym A `{setoid A, @setoidDec A _} (a b : A) : eqb a b = eqb b a.
Proof.
  case Hab: (eqb a b); case Hba: (eqb b a); firstorder.
  move/eqbP in Hab; move/eqbP in Hba.
  by contradict Hba; apply eq_sym.
  move/eqbP in Hab; move/eqbP in Hba.
  by contradict Hab; apply eq_sym.
Qed.

Definition eq_dec A `{setoidDec A} (x y : A) : {x ≡ y} + {x !≡ y}.
Proof.
  case Heqb: (eqb x y).
  { by left; apply/eqbP. }
  { by right; apply/eqbP; rewrite -> Heqb. }
Defined.
  
Class posetDec A {setoidA : setoid A} {posetA : @poset A setoidA} {setoidDecA : @setoidDec A setoidA} : Type := PosetDec {
  leb_ : ∀ (x y : A), bool;
  lebP_ : ∀ x y, reflect (x ⊑ y) (leb_ x y)
}.

Definition leb {A} `{posetDec A} := nosimpl (@leb_ _ _ _ _ _).
Lemma lebP {A} `{posetDec A} : ∀ x y, reflect (x ⊑ y) (leb x y).
Proof. exact: lebP_. Qed.

Definition le_dec A `{posetDec A} (x y : A) : {x ⊑ y} + {~ x ⊑ y}.
Proof.
  case Hleb: (leb x y).
  { by left; apply/lebP. }
  { by right; apply/lebP; rewrite -> Hleb. }
Defined.

Definition ltb {A} `{posetDec A} :=
  nosimpl (fun x y : A => andb (leb x y) (~~ (eqb x y))).

Lemma ltP {A} `{posetDec A} (x y : A) : reflect (lt x y) (ltb x y).
Proof.
  rewrite /lt /ltb.
  case (@andP _ _).
  { move=> [Hle Hneq]; constructor; split; first by exact/lebP.
    exact/eqbP.
  }
  { move=> Hlene; constructor; contradict Hlene; case: Hlene => *; split; first by exact/lebP.
    exact/eqbP.
  }
Qed.

Definition lt_dec A `{posetDec A} (x y : A) : {x ⊑! y} + {~ x ⊑! y}.
Proof.
  case Hltb: (ltb x y).
  { by left; apply/ltP. }
  { by right; apply/ltP; rewrite -> Hltb. }
Defined.

Section Bools.
  Global Program Instance bool_setoid : setoid bool := {
    eq_ := Logic.eq
  }.

  Global Program Instance bool_usual_setoid : @usual_setoid bool _ := {
  }.

  Global Program Instance bool_poset : @poset bool _ := {
    le_ := nosimpl (fun x y => x -> y)
  }.
  Next Obligation.
    move=> x y.
    by case: x; case: y; firstorder.
  Qed.
  Next Obligation.
    move=> x y //=.
    case: x; case: y; firstorder.
    by case H.
    by case H0.
  Qed.
  
  Global Program Instance bool_hasBot : @hasBot bool _ _ := {
    bot_ := false
  }.

  Global Program Instance bool_hasTop : @hasTop bool _ _ := {
    top_ := true
  }.
  Next Obligation. by case: x. Qed.

  Global Program Instance bool_order : @order bool _ _ := {
  }.
  Next Obligation.
    by case: x; case: y; firstorder.
  Qed.

  Global Program Instance bool_hasJoin : @hasJoin bool _ _ := {
    join_ := orb
  }.
  Next Obligation.
    by case: x; case: y; firstorder.
  Qed.
  Next Obligation.
    by case: x; case: y; firstorder.
  Qed.
  Next Obligation.
    by case: x Hx; case: y Hy; firstorder.
  Qed.

  Global Program Instance bool_setoidDec : @setoidDec bool _ := {
    eqb_ := Bool.eqb
  }.
  Next Obligation.
    by case: x; case: y; constructor.
  Qed.

  Global Program Instance bool_posetDec : @posetDec bool _ _ _ := {
    leb_ := nosimpl (fun x y => x ==> y)
  }.
  Next Obligation.
    by case: x; case: y; constructor; firstorder.
  Qed.
End Bools.

Add Parametric Morphism {A} `{setoidDec A} : eqb with
  signature eq ==> eq ==> eq as eqb_mor.
Proof.
  move=> x y Hxy z w Hzw; move: Hxy Hzw.
  case E: (eqb x z); case F: (eqb y w) => Hxy Hzw //=.
  by case/eqbP: F; rewrite <- Hxy; rewrite <- Hzw; apply/eqbP.
  by case/eqbP: E; rewrite -> Hxy; rewrite -> Hzw; apply/eqbP.
Qed.

Add Parametric Morphism {A} `{posetDec A} : leb with
  signature le --> le ++> le as leb_mor.
Proof.
  move=> x y Hxy z w Hzw; move: Hxy Hzw.
  case E: (leb x z); case F: (leb y w) => Hxy Hzw //=; firstorder.
  by case/lebP: F; rewrite -> Hxy; rewrite <- Hzw; apply/lebP.
Qed.

Add Parametric Morphism {A} `{posetDec A} : leb with
  signature eq ==> eq ==> eq as leb_eq_mor.
Proof.
  move=> x y Hxy z w Hzw.
  apply/le_antisym.
  by rewrite -> Hxy; rewrite -> Hzw; apply/eq_le/eq_refl.
  by rewrite <- Hxy; rewrite <- Hzw; apply/eq_le/eq_refl.
Qed.

Class maplike (Key Elt M : Type) 
  `{setoid Key, @setoidDec Key _, setoid Elt, setoid M} 
  `{@poset Elt _, @poset M _}
  : Type := Maplike {
  mlookup : forall (key:Key) (m:M), Elt;
  madd : forall (key:Key) (elt:Elt) (m:M), M;
  mremove : forall (key:Key) (m:M), M;

  mlookup_le_proper : Proper (eq ==> le ==> le) mlookup;
  madd_le_proper : Proper (eq ==> le ==> le ==> le) madd;
  mremove_le_proper : Proper (eq ==> le ==> le) mremove
}.
Notation "M '[' key ']!' " := (mlookup key M) (at level 0, no associativity).
Notation "M '[' key '<-' v ']!' " := (madd key v M) (at level 0, no associativity).

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : mlookup with
    signature eq ==> le ==> le as mlookup_le_mor.
Proof. exact/mlookup_le_proper. Qed.

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : mlookup with
    signature eq ==> eq ==> eq as mlookup_mor.
Proof. 
  intros. apply le_antisym.
  apply mlookup_le_mor; [idtac|apply eq_le]; assumption.
  apply mlookup_le_mor; [idtac|apply eq_le]; apply eq_sym; assumption.
Qed.

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : madd with
    signature eq ==> le ==> le ==> le as madd_le_mor.
Proof. exact/madd_le_proper. Qed.

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : madd with
    signature eq ==> eq ==> eq ==> eq as madd_mor.
Proof. 
  intros. apply le_antisym.
  apply madd_le_mor; [idtac|apply eq_le|apply eq_le]; assumption.
  apply madd_le_mor; [idtac|apply eq_le|apply eq_le]; apply eq_sym; assumption.
Qed.

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : mremove with
    signature eq ==> le ==> le as mremove_le_mor.
Proof. exact/mremove_le_proper. Qed.

Global Add Parametric Morphism (Key Elt M : Type) `{maplike Key Elt M} : mremove with
    signature eq ==> eq ==> eq as mremove_mor.
Proof. 
  intros. apply le_antisym.
  apply mremove_le_mor; [idtac|apply eq_le]; assumption.
  apply mremove_le_mor; [idtac|apply eq_le]; apply eq_sym; assumption.
Qed.

Class setlike (Elt S : Type) `{setoid Elt, setoid S} : Type := Setlike {
  sempty : S;
  ssingleton : forall (elt:Elt), S;
  smem : forall (elt:Elt) (s:S), bool;
  sadd : forall (elt:Elt) (s:S), S;
  sremove : forall (elt:Elt) (s:S), S;

  ssingleton_proper : Proper (eq ==> eq) ssingleton;
  smem_proper : Proper (eq ==> eq ==> eq) smem;
  sadd_proper : Proper (eq ==> eq ==> eq) sadd;
  sremove_proper : Proper (eq ==> eq ==> eq) sremove;

  sempty_spec : forall (elt:Elt), ~~ (smem elt sempty);
  ssingleton_spec : forall (e e':Elt),
    (e ≡ e' ∧ smem e (ssingleton e')) ∨
    (not (e ≡ e') ∧ ~~ (smem e (ssingleton e')));
  sadd_spec : forall (e e':Elt) (s:S),
    (e ≡ e' ∧ smem e (sadd e' s)) ∨
    (not (e ≡ e') ∧ smem e (sadd e' s) = smem e s);
  sadd_remove : forall (e e':Elt) (s:S),
    (e ≡ e' ∧ ~~ (smem e (sremove e' s))) ∨
    (not (e ≡ e') ∧ smem e (sremove e' s) = smem e s)
}.
Notation "'!{}'" := (sempty) (at level 0, no associativity).
Notation "'!{' v '}'" := (ssingleton v).
Notation "'!{' v ';' S '}'" := (sadd v S) (v at level 99).
Notation "'!{' S '~' v '}'" := (sremove v S) (S at level 99).
Notation "v '!\in' S" := (smem v S) (at level 70, no associativity).

Global Add Parametric Morphism (Elt S : Type) `{setlike Elt S} : ssingleton with
    signature eq ==> eq as ssingleton_mor.
Proof. exact/ssingleton_proper. Qed.

Global Add Parametric Morphism (Elt S : Type) `{setlike Elt S} : smem with
    signature eq ==> eq ==> eq as smem_mor.
Proof. exact/smem_proper. Qed.

Global Add Parametric Morphism (Elt S : Type) `{setlike Elt S} : sadd with
    signature eq ==> eq ==> eq as sadd_mor.
Proof. exact/sadd_proper. Qed.

Global Add Parametric Morphism (Elt S : Type) `{setlike Elt S} : sremove with
    signature eq ==> eq ==> eq as sremove_mor.
Proof. exact/sremove_proper. Qed.

Lemma congr_if A `{setoid A} :
  forall
    (b b' : bool) (Hb: b = b')
    (t t' : A) (Ht : eq t t')
    (f f' : A) (Hf : eq f f'),
    eq (if b then t else f) (if b' then t' else f').
Proof.
  move=> b b' Hb t t' Ht f f' Hf.
  subst.
  by case: b'.
Qed.
