Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Export ZArith.

Require Import DomType.
Require Import DomBasic.

Open Scope Z.

Section Zs.
  Global Program Instance Z_setoid : setoid Z := {
    eq_ := Logic.eq
  }.

  Global Program Instance Z_usual_setoid : @usual_setoid Z _ := {
  }.
  
  Global Program Instance Z_poset : @poset Z _ := {
    le_ x y := Z.le x y
  }.
  Next Obligation.
    move=> x y ->.
    by case: Z.le_preorder => [Hrefl _]; apply/Hrefl.
  Qed.
  Next Obligation.
    move=> /= x y Hxy Hyx.
    rewrite/eq; simpl.
    exact/Z.le_antisymm.
  Qed.

  Global Program Instance Z_order : @order Z _ _.
  Next Obligation.
    case Hcmp: (x ?= y).
    by left; apply/eq_le/Z.compare_eq_iff.
    by left; apply/Z.lt_le_incl/Z.compare_lt_iff.
    by right; apply/Z.lt_le_incl/Z.compare_gt_iff.
  Qed.
  
  Global Program Instance Z_hasJoin : @hasJoin Z _ _ := {
    join_ := Z.max
  }.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite Z.max_r.
    rewrite Z.max_l; last by [].
    exact/eq_le/eq_refl.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    rewrite Z.max_r; last by [].
    exact/eq_le/eq_refl.
    by rewrite Z.max_l.
  Qed.
  Next Obligation.
    case: (le_total x y) => Hle.
    by rewrite Z.max_r.
    by rewrite Z.max_l.
  Qed.

  Global Program Instance Z_setoidDec : @setoidDec Z _ := {
    eqb_ := Z.eqb
  }.
  Next Obligation.
    apply/equivP; last by apply Z.eqb_eq.
    by case: (x =? y); constructor.
  Qed.

  Global Program Instance Z_posetDec : @posetDec Z _ _ _ := {
    leb_ := Z.leb
  }.
  Next Obligation.
    apply/equivP; last by apply Z.leb_le.
    by case: (x <=? y); constructor.
  Qed.
End Zs.

(*******************************)
(* Ltac for Cdiv-related proof *)
(*******************************)

Ltac destif := 
  match goal with
    | [H: _ |- context[if ?c then _ else _]] => 
      let cond := fresh "cond" in
      let C := fresh "C" in
        remember c as cond eqn:C; symmetry in C; destruct cond
  end.

Ltac inv_eq := 
  match goal with
    | [H: False |- _] => inversion H
    | [H: ?n ≡ ?m |- _] => unfold eq in H; simpl in H; inversion H; clear H
    | [H: ?n ⊑ ?m |- _] => unfold le in H; simpl in H
    | [|- ?n ≡ ?m] => unfold eq; simpl
    | [|- ?n ⊑ ?m] => unfold le; simpl
    | [H: _ /\ _ |- _] => destruct H
  end.

Ltac inv_eq' := 
  match goal with
    | [H: False |- _] => inversion H
    | [H: ?n ≡ ?m |- _] => unfold eq in H; simpl in H; inversion H; clear H
    | [H: ?n ⊑ ?m |- _] => unfold le in H; simpl in H
    | [H: _ /\ _ |- _] => destruct H
  end.

Ltac leb_and_or :=
  repeat
  match goal with
    | [C : leb _ _ && leb _ _ || leb _ _ && leb _ _ = false |- _] =>
      let Ha := fresh "Ha" in
      let Hb := fresh "Hb" in
      apply Bool.orb_false_elim in C; destruct C as [Ha Hb]
    | [C : leb _ _ && leb _ _ || leb _ _ && leb _ _ = true |- _] =>
      let H := fresh "H" in
      let Ha := fresh "Ha" in
      let Hb := fresh "Hb" in
        move/orP:C=>C; elim C; intros H; clear C; move/andP:H=> [Ha Hb];
          move/lebP:Ha => Ha; move/lebP:Hb => Hb; simpl in Ha,Hb
    | [Hab: ?a <= ?b, H: leb ?a ?b && leb _ _ = false |- _] =>
      let Hleb_ab:= fresh "Hleb_ab" in
        assert (Hleb_ab:leb a b); first apply/lebP; simpl; auto;
          rewrite Hleb_ab in H; rewrite Bool.andb_true_l in H
    | [Hab: ?a <= ?b, H: leb _ _ && leb ?a ?b = false |- _] =>
      let Hleb_ab:= fresh "Hleb_ab" in
        assert (Hleb_ab:leb a b); first apply/lebP; simpl; auto;
          rewrite Hleb_ab in H; rewrite Bool.andb_true_r in H
    | [H: ?a ⊑ ?b |- _] => rewrite/le in H; simpl in H
  end.

(*******************)
(* Zopp properties *)
(*******************)

Lemma Zopp_preserve_0: ∀ a, a = 0 -> - a = 0.
Proof. destruct a; firstorder. Qed.

Lemma pos_imply_opp_le : ∀ n, 0 <= n -> - n <= n.
Proof. intros; apply Zle_trans with 0; omega. Qed.

(******************)
(* Zle properties *)
(******************)

Lemma l_z_u_dec : ∀ l u, {l <= 0 ∧ 0 <= u} + {0 < l ∨ u < 0}.
Proof.
  intros l u; destruct (Z_le_dec l 0); first destruct (Z_le_dec 0 u).
  left; auto.
  right; right; omega.
  right; left; omega.
Qed.

(*******************)
(* Zabs properties *)
(*******************)

Lemma Zabs_neq_0: ∀ a, a ≠ 0 -> Z.abs a ≠ 0.
Proof. firstorder using Z.abs_0_iff. Qed.

Lemma zabs_le : ∀ l m u (Hl: l <= m) (Hu: m <= u), 
  Z.abs m <= Z.max (Z.abs l) (Z.abs u).
Proof.
  intros l m u Hlm Hmu; destruct (Z_le_dec 0 m).
  {
    rewrite (Z.abs_eq m); last omega.
    rewrite (Z.abs_eq u); last omega.
    apply Zle_trans with u; auto using Z.le_max_r.
  }
  {
    rewrite (Z.abs_neq m); last omega.
    rewrite (Z.abs_neq l); last omega.
    apply Zle_trans with (-l); [omega| auto using Z.le_max_l].
  }    
Qed.

(************************)
(* Zmin/Zmax properties *)
(************************)

Lemma Zmin_le_Zmax : ∀n m, Z.min n m <= Z.max n m.
Proof. intros; apply Z.min_le_iff; left; apply Z.le_max_l. Qed.

Lemma Zmin_le_Zmax_strong : ∀n m o p, n <= o -> Z.min n m <= Z.max o p.
Proof. 
  intros n m o p H; apply Z.min_le_iff with (n:=n) (m:=m) (p:=Z.max o p); left.
  apply Zle_trans with o; eauto using Z.le_max_l.
Qed.

(*******************)
(* Zdiv properties *)
(*******************)

Lemma Z_div_le_strong: ∀ a b c : Z, 0 <= c → a <= b → a / c <= b / c.
Proof.
  intros a b c Hc Hab.
  apply Z_le_lt_eq_dec in Hc; destruct Hc; first (apply Z_div_le; omega).
  rewrite <- e. repeat rewrite Zdiv_0_r; omega.
Qed.

Lemma Z_div_pos_strong: ∀ a b : Z, 0 <= a -> 0 <= b → 0 <= a / b.
Proof.
  intros a b Hza Hzb; destruct (Z_le_lt_eq_dec 0 b); auto.
  apply Z_div_pos; omega.
  rewrite <- e; rewrite Zdiv_0_r; omega.
Qed.

(**********************************)
(* Cdiv definition and properties *)
(**********************************)

(*
Eval compute in (Z.div 3 2). (* 1 *)
Eval compute in (Z.div 3 (-2)). (* -2, this should be -1 in C *)
Eval compute in (Z.div (-3) 2). (* -2, this should be -1 in C *)
Eval compute in (Z.div (-3) (-2)). (* 1 *)
*)

Definition Cdiv (z1 z2 : Z) := 
  (if (leb 0 z1 && leb 0 z2) || (leb z1 0 && leb z2 0)
    then Z.div (Z.abs z1) (Z.abs z2)
    else Z.opp (Z.div (Z.abs z1) (Z.abs z2)))%Z.

(*
Eval compute in (Cdiv 3 2). (* 1 *)
Eval compute in (Cdiv 3 (-2)). (* -1 *)
Eval compute in (Cdiv (-3) 2). (* -1 *)
Eval compute in (Cdiv (-3) (-2)). (* 1 *)
*)

Lemma Cdiv_0_l: ∀ (a:Z), (Cdiv 0 a) = 0.
Proof. intros; unfold Cdiv; destif; eauto using Zdiv_0_l. Qed.

Lemma Cdiv_0_r: ∀ (a:Z), (Cdiv a 0) = 0.
Proof. intros; unfold Cdiv; destif; eauto using Zdiv_0_r,Zopp_preserve_0. Qed.

Lemma Cdiv_1_r: ∀ (a:Z), (Cdiv a 1) = a.
Proof.
  intros a; unfold Cdiv; destif; simpl Z.abs.
  {
    apply Bool.orb_true_iff in C; destruct C as [C|C];
      apply andb_prop in C; destruct C as [Ca Cb].
    move/lebP:Ca => Ca.
    replace (Z.abs a) with a; firstorder using Zdiv_1_r, Z.abs_eq.
    move/lebP:Cb => Cb.
    rewrite/le in Cb; simpl in Cb; omega.
  }
  {
    apply Bool.orb_false_iff in C; destruct C as [Ca Cb].
    apply Bool.andb_false_iff in Ca; apply Bool.andb_false_iff in Cb. 
    destruct Ca as [Caa|Cab]; [idtac|destruct Cb as [Cba|Cbb]].
    {
      move/lebP:Caa; intros Caa.
      replace (Z.abs a) with (-a).
      rewrite Zdiv_1_r; omega.
      rewrite/le in Caa; simpl in Caa; symmetry; apply Z.abs_neq; omega.
    }
    { move/lebP:Cab; intros Cab; rewrite/le in Cab; simpl in Cab; omega. }
    { move/lebP:Cab; intros Cab; rewrite/le in Cab; simpl in Cab; omega. }
  }
Qed.

Lemma Cdiv_1_l: ∀ (a:Z), 1 < (Z.abs a) -> Cdiv 1 a = 0.
Proof. intros; unfold Cdiv; destif; auto using Zdiv_1_l, Zopp_preserve_0. Qed.

Lemma Cdiv_n_l: ∀ (n a:Z), Z.abs n < Z.abs a -> Cdiv n a = 0.
Proof. 
  intros; unfold Cdiv; destif; 
  auto using Zopp_preserve_0, Zdiv_small, Z.abs_nonneg.
Qed.

Lemma Cdiv_same_full: ∀ a, a ≠ 0 -> Cdiv a a = 1.
Proof.
  intros a H; unfold Cdiv; destif.
  auto using Z_div_same_full, Zabs_neq_0.
  apply Bool.orb_false_iff in C; destruct C as [Ca Cb].
  apply Bool.andb_false_iff in Ca; apply Bool.andb_false_iff in Cb. 
  destruct Ca as [Ca|Ca]; destruct Cb as [Cb|Cb];
    move/lebP:Ca => Ca; move/lebP:Cb => Cb; rewrite/le in Ca Cb; simpl in *; omega.
Qed.  

Lemma Cdiv_mult_full: ∀ a b, b ≠ 0 -> Cdiv (a * b) b = a.
Proof.
  intros a b H; unfold Cdiv; destif.
  {
    apply Bool.orb_true_iff in C; destruct C as [C|C];
      apply andb_prop in C; destruct C as [Ca Cb].
    {
      rewrite Z.abs_mul Z_div_mult_full.
      move/lebP:Ca => Ca; move/lebP:Cb => Cb; rewrite/le in Ca Cb; simpl in *. 
      assert (0 <= a); first by apply Zmult_le_0_reg_r with b; omega.
      apply Z.abs_eq_iff; assumption.
      auto using Zabs_neq_0.
    }
    {
      move/lebP:Ca => Ca; move/lebP:Cb => Cb; rewrite/le in Ca Cb; simpl in *. 
      rewrite Z.abs_mul Z_div_mult_full.
      assert (0 <= a).
      apply Zmult_le_0_reg_r with (-b); first omega.
      rewrite <- Zopp_mult_distr_r; omega.
      eauto using Z.abs_eq.
      auto using Zabs_neq_0.
    }
  }
  {
    apply Bool.orb_false_iff in C; destruct C as [Ca Cb].
    apply Bool.andb_false_iff in Ca; apply Bool.andb_false_iff in Cb. 
    destruct Ca as [Ca|Ca]; destruct Cb as [Cb|Cb];
      move/lebP:Ca => Ca; move/lebP:Cb => Cb; rewrite/le in Ca, Cb; simpl in Ca,Cb; try omega. 
    {
      rewrite Z.abs_mul Z_div_mult_full.
      assert (0 <= -a); first apply Zmult_le_0_reg_r with b; try omega.
      rewrite <- Zopp_mult_distr_l; omega.
      assert (a <= 0); first omega.
      rewrite Z.abs_neq; omega.
      auto using Zabs_neq_0.
    }      
    {
      rewrite Z.abs_mul Z_div_mult_full.
      assert (0 <= -a); first apply Zmult_le_0_reg_r with (-b); try omega.
      rewrite <- Zopp_mult_distr_l, <- Zopp_mult_distr_r; omega.
      assert (a <= 0); first omega.
      rewrite Z.abs_neq; omega.
      auto using Zabs_neq_0.
    }
  }
Qed.

Lemma Cdiv_preserve_le : ∀a b n, a <= b -> 0 <= n -> Cdiv a n <= Cdiv b n.
Proof.
  intros a b n H H0; unfold Cdiv; repeat destif.
  {
    move/orP:C => [Ca|Ca]; move/orP:C0 => [Cb|Cb];
      move/andP:Ca => [Caa Cab]; move/andP:Cb => [Cba Cbb];
        move/lebP:Caa => Caa; move/lebP:Cab => Cab;
          move/lebP:Cba => Cba; move/lebP:Cbb => Cbb;
            rewrite/le in Caa Cab Cba Cbb; simpl in Caa,Cab,Cba,Cbb.
    apply Z_div_le_strong; first apply Z.abs_nonneg.
    rewrite (Z.abs_eq a); auto; rewrite (Z.abs_eq b); auto.
    apply Z_div_le_strong; first apply Z.abs_nonneg.
    rewrite (Z.abs_eq a); auto; rewrite (Z.abs_eq b); omega.
    assert (n = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_r; omega.
    assert (n = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_r; omega.
  }    
  {
    apply Bool.orb_false_iff in C0; destruct C0 as [Cba Cbb].
    move/orP:C => [Ca|Ca];  move/andP:Ca => [Caa Cab].
    {
      rewrite Cab in Cba; rewrite Bool.andb_true_r in Cba.
      move/lebP:Cba => Cba; elim Cba.
      move/lebP:Caa => Caa; eauto using le_trans.
    }
    {
      move/lebP:Cab => Cab; rewrite/le in Cab; simpl in Cab.
      assert (n = 0); first omega; subst.
      repeat rewrite Zdiv_0_r; omega.
    }      
  }
  {
    apply Zle_trans with 0.
    apply Z.opp_nonpos_nonneg; apply Z_div_pos_strong; apply Z.abs_nonneg.
    apply Z_div_pos_strong; apply Z.abs_nonneg.
  }
  {
    apply Bool.orb_false_iff in C; destruct C as [Caa Cab].
    apply Bool.orb_false_iff in C0; destruct C0 as [Cba Cbb].
    apply -> Z.opp_le_mono; apply Z_div_le_strong; first apply Z.abs_nonneg.
    move/lebP:H0 => Hzn; rewrite Hzn in Caa Cba;
      rewrite Bool.andb_true_r in Caa; rewrite Bool.andb_true_r in Cba.
    move/lebP:Caa => Caa; move/lebP:Cba => Cba; rewrite/le in Caa Cba; simpl in Caa,Cba.
    rewrite (Z.abs_neq a); last omega; rewrite (Z.abs_neq b); omega.
  }
Qed.

Lemma Cdiv_opp : ∀n m, Cdiv n (-m) = - Cdiv n m.
Proof.
  intros n m; unfold Cdiv; repeat destif.
  {
    move/orP:C => [Ca|Ca]; move/orP:C0 => [Cb|Cb];
      move/andP:Ca => [Caa Cab]; move/andP:Cb => [Cba Cbb];
        move/lebP:Caa => Caa; move/lebP:Cab => Cab;
          move/lebP:Cba => Cba; move/lebP:Cbb => Cbb;
            rewrite/le in Caa Cab Cba Cbb; simpl in Caa,Cab,Cba,Cbb.
    assert (m = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_r; omega.
    assert (n = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_l; omega.
    assert (n = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_l; omega.
    assert (m = 0); first omega; subst; simpl.
    repeat rewrite Zdiv_0_r; omega.
  }    
  {
    rewrite Z.abs_opp; rewrite Z.opp_involutive; reflexivity.
  }
  {
    rewrite Z.abs_opp; reflexivity.
  }
  {
    rewrite Z.abs_opp; rewrite Z.opp_involutive.
    apply Bool.orb_false_iff in C; destruct C as [Caa Cab].
    apply Bool.orb_false_iff in C0; destruct C0 as [Cba Cbb].
    destruct (Z_le_dec 0 n).
    {
      assert (Hzn: leb 0 n); first by (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hzn in Caa,Cba; 
        rewrite Bool.andb_true_l in Caa;
          rewrite Bool.andb_true_l in Cba.
      move/lebP:Caa => Caa; move/lebP:Cba => Cba; rewrite/le in Caa Cba; simpl in Caa, Cba.
      assert (m = 0); first omega; subst; simpl.
      rewrite Zdiv_0_r; omega.
    }
    {
      assert (Hnz: leb n 0); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hnz in Cab,Cbb; 
        rewrite Bool.andb_true_l in Cab;
          rewrite Bool.andb_true_l in Cbb.
      move/lebP:Cab => Cab; move/lebP:Cbb => Cbb; rewrite/le in Cab Cbb; simpl in Cab, Cbb.
      assert (m = 0); first omega; subst; simpl.
      rewrite Zdiv_0_r; omega.
    }      
  }
Qed.

Lemma Cdiv_opp_l: ∀ n m : Z, Cdiv (-n) m = - Cdiv n m.
Proof.
  intros n m; unfold Cdiv; repeat destif; leb_and_or.
  {
    assert (n = 0); first omega; subst.
    simpl; rewrite Zdiv_0_l; auto.
  }
  {
    assert (m = 0); first omega; subst.
    simpl; repeat rewrite Zdiv_0_r; auto.
  }
  {
    assert (m = 0); first omega; subst.
    simpl; repeat rewrite Zdiv_0_r; auto.
  }
  {
    assert (n = 0); first omega; subst.
    simpl; rewrite Zdiv_0_l; auto.
  }
  { rewrite Z.opp_involutive Z.abs_opp; reflexivity. }    
  { rewrite Z.opp_involutive Z.abs_opp; reflexivity. }    
  { rewrite Z.abs_opp; reflexivity. }    
  { rewrite Z.abs_opp; reflexivity. }    
  { 
    apply Bool.andb_false_iff in Ha; elim Ha; clear Ha; intro Ha;
      apply Bool.andb_false_iff in Hb; elim Hb; clear Hb; intro Hb;
        move/lebP:Ha=>Ha; move/lebP:Hb=>Hb; rewrite/le in Ha Hb; simpl in Ha,Hb.
    { assert (n = 0); first omega; subst; simpl; omega. }
    {
      assert (leb 0 (-n)); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite H in Ha0; rewrite Bool.andb_true_l in Ha0.
      move/lebP:Ha0=>Ha0; rewrite/le in Ha0; simpl in Ha0; omega.
    }
    {
      assert (leb (-n) 0); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite H in Hb0; rewrite Bool.andb_true_l in Hb0.
      move/lebP:Hb0=>Hb0; rewrite/le in Hb0; simpl in Hb0; omega.
    }      
    { assert (m = 0); first omega; subst; simpl; omega. }
  }    
Qed.

Lemma Cdiv_preserve_le_neg : ∀a b n, a <= b -> n <= 0 -> Cdiv b n <= Cdiv a n.
Proof.
  intros; apply Z.opp_le_mono; repeat rewrite <- Cdiv_opp.
  apply Cdiv_preserve_le; omega.
Qed.

Lemma Cdiv_le_compat_l: ∀ p q r : Z, 
  0 <= p → 0 < q -> q <= r → Cdiv p r <= Cdiv p q.
Proof.
  intros; unfold Cdiv; repeat destif.
  {
    apply Z.div_le_compat_l; first apply Z.abs_nonneg.
    split; first (apply Z.abs_pos; omega).
    rewrite (Z.abs_eq q); last omega.
    rewrite (Z.abs_eq r); omega.
  }
  {
    apply Bool.orb_false_iff in C0; destruct C0.
    assert (leb 0 p); first (apply/lebP; rewrite/le; simpl; omega).
    rewrite H4 in H2; rewrite Bool.andb_true_l in H2.
    move/lebP:H2 => H2; rewrite/le in H2; simpl in H2; omega.
  }
  {
    apply Zle_trans with 0.
    apply Z.opp_nonpos_nonneg; apply Z_div_pos_strong; apply Z.abs_nonneg.
    apply Z_div_pos_strong; apply Z.abs_nonneg.
  }
  {
    apply Bool.orb_false_iff in C0; destruct C0.
    assert (leb 0 p); first (apply/lebP; rewrite/le; simpl; omega).
    assert (leb 0 q); first (apply/lebP; rewrite/le; simpl; omega).
    rewrite H4 H5 in H2; inversion H2.
  }    
Qed.  

Lemma Cdiv_pos_pos : ∀n m, 0 <= n -> 0 <= m -> 0 <= Cdiv n m.
Proof.
  intros n m Hn Hm; replace 0 with (Cdiv 0 m); last (rewrite Cdiv_0_l; auto).
  apply Cdiv_preserve_le; omega.
Qed.

Ltac leauto := repeat 
  (match goal with
     | [H: ?n ⊑ ?m |- _ ] => simpl in H
     | [H: ¬ ?n ⊑ ?m |- _] => simpl in H
     | [H: ¬ ?n <= ?m |- _] => 
       assert (m < n); first omega;
       assert (m <= n); first omega; clear H
     | [H: _ |- Z.min _ ?n <= ?n] => apply Zle_min_r
     | [H: _ |- Z.min ?n _ <= ?n] => apply Zle_min_l
     | [H: _ |- ?n <= Z.max ?n _ ] => apply Zle_max_l
     | [H: _ |- ?n <= Z.max _ ?n ] => apply Zle_max_r
     | [H: context[-_ * _] |- _] => rewrite <- Zopp_mult_distr_l in H
     | [H: context[_ * -_] |- _] => rewrite <- Zopp_mult_distr_r in H
     | [H: context[- - _] |- _] => rewrite Z.mul_opp_opp in H
     | [H: ?n <= 0 |- ?n * ?m <= ?n * ?p] =>
       let Hg := fresh "Hg" in
         assert (Hg: - n * p <= - n * m);
           [apply Zmult_le_compat_l |
            repeat rewrite <- Zopp_mult_distr_l in Hg ]
     | [H: 0 <= ?n |- ?n * ?m <= ?n * ?p] =>
       let Hg := fresh "Hg" in
         assert (Hg: n * m <= n * p);
           [apply Zmult_le_compat_l | idtac]
     | [H: ?n <= 0 |- ?m * ?n <= ?p * ?n] =>
       let Hg := fresh "Hg" in
         assert (Hg: p * - n <= m * - n);
           [apply Zmult_le_compat_r |
            repeat rewrite <- Zopp_mult_distr_r in Hg ]
     | [H: 0 <= ?n |- ?m * ?n <= ?p * ?n] =>
       let Hg := fresh "Hg" in
         assert (Hg: m * n <= p * n);
           [apply Zmult_le_compat_r | idtac]
     | [Hab: ?a <= ?b, Hnz: ?n <= 0 |- Cdiv ?b ?n <= Cdiv ?a ?n] =>
       apply Cdiv_preserve_le_neg
     | [Hab: ?a <= ?b, Hzn: 0 <= ?n |- Cdiv ?a ?n <= Cdiv ?b ?n] =>
       apply Cdiv_preserve_le
     | [Hza: 0 <= ?a, Hzn: 0 < ?n, Hnm: ?n <= ?m |- Cdiv ?a ?m <= Cdiv ?a ?n] =>
       apply Cdiv_le_compat_l
     | [Hza: 0 <= ?a, Hnm: ?n <= ?m, Hmz: ?m < 0 |- Cdiv ?a ?m <= Cdiv ?a ?n] =>
       apply Z.opp_le_mono; repeat rewrite <- Cdiv_opp; apply Cdiv_le_compat_l
     | [Hza: ?a <= 0, Hzn: 0 < ?n, Hnm: ?n <= ?m |- Cdiv ?a ?n <= Cdiv ?a ?m] =>
       apply Z.opp_le_mono; repeat rewrite <- Cdiv_opp_l; apply Cdiv_le_compat_l
     | [Hza: ?a <= 0, Hnm: ?n <= ?m, Hmz: ?m < 0 |- Cdiv ?a ?n <= Cdiv ?a ?m] =>
       apply Z.opp_le_mono; repeat rewrite <- Cdiv_opp_l; 
         apply Z.opp_le_mono; repeat rewrite <- Cdiv_opp; 
           apply Cdiv_le_compat_l
   end; try omega); try omega.

(**********************)
(* leb/ltb properties *)
(**********************)

Lemma leb_false_imply_lt : ∀n m, leb n m = false -> m < n.
Proof.
  intros; move/lebP:H => /= H. 
  rewrite/le in H; simpl in H.
  omega.
Qed.

Lemma leb_true_imply_le : ∀n m, leb n m = true -> n <= m.
Proof.
  intros; move/lebP:H => /= H. 
  rewrite/le in H; simpl in H.
  omega.
Qed.

Lemma ltb_true_imply_lt : ∀n m, ltb n m = true -> n < m.
Proof. 
  intros; unfold ltb in H.
  apply Bool.andb_true_iff in H; destruct H.
  apply leb_true_imply_le in H.
  apply Bool.negb_true_iff in H0.
  move/eqbP:H0 => /= H0; rewrite/eq in H0; simpl in H0; omega.
Qed.

Lemma ltb_false_imply_le : ∀n m, ltb n m = false -> m <= n.
Proof. 
  intros; unfold ltb in H.
  destruct (Z_le_dec n m); last omega.
  apply Bool.andb_false_iff in H; elim H; intros.
  apply leb_false_imply_lt in H0; omega.
  move/eqbP:H0 => /= H0; rewrite/eq in H0; simpl in H0; omega.
Qed.

(**********************************)
(* Cmod definition and properties *)
(**********************************)

Definition Cmod n m := n - (Cdiv n m) * m.

Lemma cmod_proper : Proper (eq ==> eq ==> eq) Cmod.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity.
Qed.

Lemma cmod_abs_m : ∀ n m (Hm: m ≠ 0),
  - ((Z.abs m) - 1) <= Cmod n m <= ((Z.abs m) - 1).
Proof.
  intros n m Hm; unfold Cmod, Cdiv; destif.
  {
    move/orP:C => [C|C]; move/andP:C => [C1 C2];
      move/lebP:C1 => /= C1; move/lebP:C2 => /= C2; rewrite/le in C1 C2; simpl in C1,C2.

    rewrite (Z.abs_eq n); last omega; rewrite (Z.abs_eq m); last omega.
    rewrite Z.mul_comm; rewrite <- Z.mod_eq; last omega.
    pose (Hlt:= Z_mod_lt n m); omega.

    rewrite (Z.abs_neq n); last omega; rewrite (Z.abs_neq m); last omega.
    rewrite Z.mul_comm. 
    rewrite <- (Z.add_opp_r n).
    rewrite Zopp_mult_distr_l.
    replace (n + - m * (- n / - m)) with (- (- n - - m * (- n / - m))); 
      last omega.
    rewrite <- Z.mod_eq; last omega.
    pose (Hlt:= Z_mod_lt (-n) (-m)); omega.
  }
  {
    apply Bool.orb_false_iff in C; destruct C as [C1 C2].
    apply Bool.andb_false_iff in C1; apply Bool.andb_false_iff in C2.
    destruct C1 as [C1|C1]; destruct C2 as [C2|C2]; 
      apply leb_false_imply_lt in C1; apply leb_false_imply_lt in C2.
    { omega. }
    { 
      rewrite (Z.abs_eq m); last omega.
      rewrite (Z.abs_neq n); last omega.
      rewrite Z.mul_comm. 
      rewrite <- Zopp_mult_distr_r.
      replace (n - - (m * (- n / m))) with (-(-n - (m * (- n / m))));
        last omega.
      rewrite <- Z.mod_eq; last omega.
      pose (Hlt:= Z_mod_lt (-n) m); omega.
    }
    {
      rewrite (Z.abs_neq m); last omega.
      rewrite (Z.abs_eq n); last omega.
      rewrite Z.mul_comm. 
      rewrite <- Zopp_mult_distr_r.
      rewrite Zopp_mult_distr_l.
      rewrite <- Z.mod_eq; last omega.
      pose (Hlt:= Z_mod_lt n (-m)); omega.
    }
    { omega. }
  }      
Qed.

Lemma cmod_le_m : ∀ l u n m (Hm: m ≠ 0) (Hl: l <= m) (Hu: m <= u),
  - (Z.max (Z.abs l) (Z.abs u) - 1) <= 
  Cmod n m <= 
  (Z.max (Z.abs l) (Z.abs u) - 1).
Proof.
  intros; 
    pose (Hcmod:= cmod_abs_m n m Hm); destruct Hcmod as [Hcmodl Hcmodu];
      pose (Hzabs:= zabs_le l m u Hl Hu); split.
  
  apply Zle_trans with (- (Z.abs m - 1)); omega.
  apply Zle_trans with (Z.abs m - 1); omega.
Qed.

Lemma cmod_opp : ∀ n m, - Cmod n m = Cmod (-n) m.
Proof.
  intros; unfold Cmod; unfold Cdiv; repeat destif.
  {
    move/orP:C => [C1|C1]; move/orP:C0 => [C2|C2];
    move/andP:C1 => [C11 C12]; move/andP:C2 => [C21 C22];
    move/lebP:C11 => /= C11; move/lebP:C12 => /= C12;
    move/lebP:C21 => /= C21; move/lebP:C22 => /= C22;
    rewrite/le in C11 C12 C21 C22; simpl in C11,C12,C21,C22.
    { assert (n = 0); first omega; subst; auto. }
    { assert (m = 0); first omega; subst; simpl. 
      repeat rewrite Zdiv_0_r; omega. }
    { assert (m = 0); first omega; subst; simpl. 
      repeat rewrite Zdiv_0_r; omega. }
    { assert (n = 0); first omega; subst; auto. }
  }
  { rewrite Z.abs_opp; rewrite <- Zopp_mult_distr_l; omega. }
  { rewrite Z.abs_opp; rewrite <- Zopp_mult_distr_l; omega. }
  {
    destruct (Z_le_dec 0 n); destruct (Z_le_dec 0 m).
    { 
      assert (Hzn: leb 0 n); first (apply/lebP; rewrite/le; simpl; omega).
      assert (Hzm: leb 0 m); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hzn in C; rewrite Hzm in C; inversion C.
    }      
    {
      assert (Hzn: leb (-n) 0); first (apply/lebP; rewrite/le; simpl; omega).
      assert (Hzm: leb m 0); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hzn in C0; rewrite Hzm in C0.
      apply Bool.orb_false_elim in C0; destruct C0 as [Ca Cb].
      inversion Cb.
    }      
    {
      assert (Hzn: leb 0 (-n)); first (apply/lebP; rewrite/le; simpl; omega).
      assert (Hzm: leb 0 m); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hzn in C0; rewrite Hzm in C0; inversion C0.
    }    
    {
      assert (Hzn: leb n 0); first (apply/lebP; rewrite/le; simpl; omega).
      assert (Hzm: leb m 0); first (apply/lebP; rewrite/le; simpl; omega).
      rewrite Hzn in C; rewrite Hzm in C.
      apply Bool.orb_false_elim in C; destruct C as [Ca Cb].
      inversion Cb.
    }
  }      
Qed.

Lemma cmod_pos : ∀ n m (Hzn: 0 <= n), 0 <= Cmod n m.
Proof.
  intros; unfold Cmod; unfold Cdiv; destif.
  {
    rewrite (Z.abs_eq n); last omega.
    rewrite Z.mul_comm.
    destruct (Z_le_dec 0 m) as [Hm|Hm].
    {
      apply Z_le_lt_eq_dec in Hm; destruct Hm as [Hm|Hm].
      {
        rewrite (Z.abs_eq m); last omega.
        pose(Z_mult_div_ge n m); omega.
      }  
      { rewrite <- Hm; omega. }
    }
    {
      rewrite (Z.abs_neq m); last omega.
      apply Zle_minus_le_0; apply Zle_trans with 0; last omega.
      apply Z.mul_nonpos_nonneg; first omega.
      apply Z_div_pos_strong; omega.
    }
  }
  {
    rewrite (Z.abs_eq n); last omega.
    rewrite Z.mul_comm.
    destruct (Z_le_dec 0 m) as [Hm|Hm].
    {
      apply Z_le_lt_eq_dec in Hm; destruct Hm as [Hm|Hm].
      {
        rewrite (Z.abs_eq m); last omega.
        apply Zle_minus_le_0; apply Zle_trans with 0; last omega.
        apply Z.mul_nonneg_nonpos; first omega.
        assert (0 <= n / m); last omega.
        apply Z_div_pos_strong; omega.
      }  
      { rewrite <- Hm; omega. }
    }
    {
      rewrite (Z.abs_neq m); last omega.
      apply Zle_minus_le_0.
      rewrite <- Zopp_mult_distr_r.
      rewrite Zopp_mult_distr_l.
      pose(Z_mult_div_ge n (-m)); omega.
    }
  }
Qed.      

Lemma cmod_neg : ∀ n m (Hnz: n <= 0), Cmod n m <= 0.
Proof.
  intros; apply Z.opp_le_mono; rewrite cmod_opp.
  apply cmod_pos; omega.
Qed.

Lemma cmod_n_zn : ∀ n m (Hzn: 0 <= n), Cmod n m <= n.
Proof.
  intros; unfold Cmod.
  assert (0 <= Cdiv n m * m); last omega.
  destruct (Z_le_dec 0 m).
  {
    apply Z.mul_nonneg_nonneg; last omega.
    apply Cdiv_pos_pos; omega.
  }
  {
    apply Z.mul_nonpos_nonpos; last omega.
    apply Z.opp_le_mono; rewrite <- Cdiv_opp.
    apply Cdiv_pos_pos; omega.
  }
Qed.

Lemma cmod_n_nz : ∀ n m (Hnz: n <= 0), n <= Cmod n m.
Proof. 
  intros; apply Z.opp_le_mono; rewrite cmod_opp.
  apply cmod_n_zn; omega.
Qed.

Lemma cmod_n_u : ∀ n u m (Hnu: n <= u) (Hzu: 0 <= u), Cmod n m <= u.
Proof.
  intros; destruct (Z_le_dec 0 n).
  apply Zle_trans with n; last omega.
  apply cmod_n_zn; auto.
  apply Zle_trans with 0; last omega.
  apply cmod_neg; omega.
Qed.

Lemma cmod_l_n : ∀ l n m (Hln: l <= n) (Hlz: l <= 0), l <= Cmod n m.
Proof.
  intros; apply Z.opp_le_mono; rewrite cmod_opp.
  apply cmod_n_u; omega.
Qed.

(*****************************************************************)
(* basic operations on Z which is used in the concrete semantics *)
(*****************************************************************)

Definition con_op_not val := if eqb val 0 then 1 else 0.
Parameter con_op_bnot : Z -> Z.

Definition con_op_lt (lhs rhs:Z) := if ltb lhs rhs then 1 else 0.
Definition con_op_le (lhs rhs:Z) := if leb lhs rhs then 1 else 0.
Definition con_op_gt (lhs rhs:Z) := if ltb rhs lhs then 1 else 0.
Definition con_op_ge (lhs rhs:Z) := if leb rhs lhs then 1 else 0.
Definition con_op_eq (lhs rhs:Z) := if eqb lhs rhs then 1 else 0.
Definition con_op_neq (lhs rhs:Z) := if eqb lhs rhs then 0 else 1.
Definition con_op_and (lhs rhs:Z) := 
  if eqb lhs 0 then 0
    else if eqb rhs 0 then 0
      else 1.
Definition con_op_or (lhs rhs:Z) :=
  if eqb lhs 0 && eqb rhs 0 
    then 0
    else 1.
Parameter con_op_band : Z -> Z -> Z.
Parameter con_op_bor : Z -> Z -> Z.
Parameter con_op_bxor : Z -> Z -> Z.
Parameter con_op_lshift : Z -> Z -> Z.
Parameter con_op_rshift : Z -> Z -> Z.

Close Scope Z.
