Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomZ.
Require Import DomMap.
Require Import DomPidx.
Require Import DomPidxSen.
Require Import DomAbsValue.
Require Import DomConValue.
Require Import DomConMem.

Require Import Syntax.

Set Implicit Arguments.

Inductive eval_lit (m:memT) : literal -> valueT -> Prop :=
| eval_lit_int : forall n,
  eval_lit m (lit_int n) (vint n)
| eval_lit_int_top : forall n,
  eval_lit m lit_int_top (vint n)
| eval_lit_top : forall v,
  eval_lit m lit_top v.

Lemma eval_lit_proper : Proper (eq ==> eq ==> eq ==> le) eval_lit.
Proof.
  move=> x y Hxy a b Hab z w Hzw.
  usual_tac Hab.
  usual_tac Hzw.
  move:b x y Hxy w. destruct b; intros; intro Hsem.
  { inversion Hsem. constructor. }
  { inversion Hsem. constructor. }
  { inversion Hsem. constructor. }
  { inversion Hsem. }
Qed.

Inductive cast_type_in : ctype -> Z -> Prop :=
| cast_in_uchar : forall z
  (Hz:(0 <= z <= 255)%Z),
  cast_type_in typ_uchar z
| cast_in_ushort : forall z
  (Hz:(0 <= z <= 65535)%Z),
  cast_type_in typ_ushort z
| cast_in_uint : forall z
  (Hz:(0 <= z <= 4294967295)%Z),
  cast_type_in typ_uint z
| cast_in_ulong : forall z
  (Hz:(0 <= z <= 4294967295)%Z),
  cast_type_in typ_ulong z
| cast_in_ulonglong : forall z
  (Hz:(0 <= z <= 18446744073709551615)%Z),
  cast_type_in typ_ulonglong z

| cast_in_char : forall z
  (Hz:(-128 <= z <= 127)%Z),
  cast_type_in typ_char z
| cast_in_short : forall z
  (Hz:(-32768 <= z <= 32767)%Z),
  cast_type_in typ_short z
| cast_in_int : forall z
  (Hz:(-2147483648 <= z <= 2147483647)%Z),
  cast_type_in typ_int z
| cast_in_long : forall z
  (Hz:(-2147483648 <= z <= 2147483647)%Z),
  cast_type_in typ_long z
| cast_in_longlong : forall z
  (Hz:(-9223372036854775808 <= z <= 9223372036854775807)%Z),
  cast_type_in typ_longlong z.

Lemma cast_check_proper : Proper (eq ==> eq ==> le) cast_type_in.
Proof.
  move=> x y Hxy a b Hab.
  usual_tac Hab.
  usual_tac Hxy.
  firstorder.
Qed.

Inductive cast_stride_of : ctype -> strideT -> Prop :=
| cast_st_uchar : cast_stride_of typ_uchar (1%Z)
| cast_st_ushort : cast_stride_of typ_ushort (2%Z)
| cast_st_uint : cast_stride_of typ_uint (4%Z)
| cast_st_ulong : cast_stride_of typ_ulong (4%Z)
| cast_st_ulonglong : cast_stride_of typ_ulonglong (8%Z)

| cast_st_char : cast_stride_of typ_char (1%Z)
| cast_st_short : cast_stride_of typ_short (2%Z)
| cast_st_int : cast_stride_of typ_int (4%Z)
| cast_st_long : cast_stride_of typ_long (4%Z)
| cast_st_longlong : cast_stride_of typ_longlong (8%Z).

Lemma cast_stride_of_proper : Proper (eq ==> eq ==> le) cast_stride_of.
Proof.
  move=> x y Hxy a b Hab.
  usual_tac Hab.
  usual_tac Hxy.
  firstorder.
Qed.

Inductive sem_expr (pidx:pidxT) : exprT -> memT -> valueT -> Prop :=
| sem_const : forall m l v 
  (Hlit:eval_lit m l v), 
  sem_expr pidx (expr_const l) m v

| sem_op_add_zz : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_add lex rex) m (vint (lz + rz)%Z)
| sem_op_add_za : forall m lex rex z al uq ofs sz st
  (Hl:sem_expr pidx lex m (vint z)) (Hr:sem_expr pidx rex m
                                                 (vheap al uq (ofs,sz,st))),
  sem_expr pidx (expr_op2 op_add lex rex) m (vheap al uq ((ofs+z)%Z,sz,st))
| sem_op_add_az : forall m lex rex z al uq ofs sz st
  (Hr:sem_expr pidx rex m (vint z)) (Hl:sem_expr pidx lex m
                                                 (vheap al uq (ofs,sz,st))),
  sem_expr pidx (expr_op2 op_add lex rex) m (vheap al uq ((ofs+z)%Z,sz,st))

| sem_op_sub_zz : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_sub lex rex) m (vint (lz - rz)%Z)
| sem_op_sub_az : forall m lex rex z al uq ofs sz st 
  (Hr:sem_expr pidx rex m (vint z)) (Hl:sem_expr pidx lex m
                                                 (vheap al uq (ofs,sz,st))),
  sem_expr pidx (expr_op2 op_sub lex rex) m (vheap al uq ((ofs-z)%Z,sz,st))

| sem_op_mul : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_mul lex rex) m (vint (lz * rz)%Z)

| sem_op_div : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_div lex rex) m (vint (Cdiv lz rz)%Z)

| sem_op_mod : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_mod lex rex) m (vint (Cmod lz rz)%Z)

| sem_op_lt : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_lt lex rex) m (vint (con_op_lt lz rz)%Z)

| sem_op_le : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_le lex rex) m (vint (con_op_le lz rz)%Z)

| sem_op_gt : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_gt lex rex) m (vint (con_op_gt lz rz)%Z)

| sem_op_ge : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_ge lex rex) m (vint (con_op_ge lz rz)%Z)

| sem_op_eq : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_eq lex rex) m (vint (con_op_eq lz rz)%Z)

| sem_op_neq : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_neq lex rex) m (vint (con_op_neq lz rz)%Z)

| sem_op_and : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_and lex rex) m (vint (con_op_and lz rz)%Z)

| sem_op_or : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_or lex rex) m (vint (con_op_or lz rz)%Z)

| sem_op_band : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_band lex rex) m (vint (con_op_band lz rz)%Z)

| sem_op_bor : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_bor lex rex) m (vint (con_op_bor lz rz)%Z)

| sem_op_bxor : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_bxor lex rex) m (vint (con_op_bxor lz rz)%Z)

| sem_op_lshift : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_lshift lex rex) m (vint (con_op_lshift lz rz)%Z)

| sem_op_rshift : forall m lex rex lz rz
  (Hl:sem_expr pidx lex m (vint lz)) (Hr:sem_expr pidx rex m (vint rz)),
  sem_expr pidx (expr_op2 op_rshift lex rex) m (vint (con_op_rshift lz rz)%Z)

| sem_op_not : forall m ex z
  (He:sem_expr pidx ex m (vint z)),
  sem_expr pidx (expr_op1 op_not ex) m (vint (con_op_not z)%Z)

| sem_op_bnot : forall m ex z
  (He:sem_expr pidx ex m (vint z)),
  sem_expr pidx (expr_op1 op_bnot ex) m (vint (con_op_bnot z)%Z)

| sem_value_of : forall m l lval v
  (Hlv:sem_lv pidx l m lval) (Hlookup:m[lval]! = Some v),
  sem_expr pidx (expr_value_of l) m v

| sem_addr_of : forall m l lval
  (Hlv:sem_lv pidx l m lval),
  sem_expr pidx (expr_addr_of l) m (vloc lval)

| sem_cast_z_in : forall m t e z
  (He:sem_expr pidx e m (vint z))
  (Ht:cast_type_in t z),
  sem_expr pidx (expr_cast t e) m (vint z)
| sem_cast_z_out : forall m t e z z'
  (He:sem_expr pidx e m (vint z))
  (Ht:not (cast_type_in t z)) (Hm:cast_type_in t z'),
  sem_expr pidx (expr_cast t e) m (vint z')
| sem_cast_array : forall m t e al uq ofs sz st st'
  (He:sem_expr pidx e m (vheap al uq (ofs,sz,st)))
  (Hst:cast_stride_of t st'),
  sem_expr pidx (expr_cast (typ_pointer t) e) m 
  (vheap al uq ((Cdiv (ofs * st) st'),(Cdiv (sz * st) st'),st'))

| sem_size_of : forall m e z,
  sem_expr pidx (expr_size_of e) m (vint z)

with sem_lv (pidx:pidxT) : lv -> memT -> locT -> Prop :=
| sem_lv_id_global : forall m (x:idT),
  sem_lv pidx (lv_id (global x)) m (lglobal x)
| sem_lv_id_local : forall m (x:idT),
  sem_lv pidx (lv_id (local x)) m (llocal pidx x)

| lv_deref_loc : forall m e l
  (He:sem_expr pidx e m (vloc l)),
  sem_lv pidx (lv_deref e) m l
| lv_deref_array : forall m e al uq ofs sz st
  (He:sem_expr pidx e m (vheap al uq (ofs,sz,st))),
  sem_lv pidx (lv_deref e) m (lheap al uq ofs)

| lv_array : forall m arr idx al uq ofs sz st z
  (Harr:sem_expr pidx arr m (vheap al uq (ofs,sz,st)))
  (Hidx:sem_expr pidx idx m (vint z)),
  sem_lv pidx (lv_array arr idx) m (lheap al uq (z+ofs)%Z)

| lv_strct : forall m lv loc pos
  (Hlv: sem_lv pidx lv m loc),
  sem_lv pidx (lv_struct lv pos) m (loc.1, pos::loc.2).

Definition stateT := (pidxT * memT)%type.

Instance memT_maplike : maplike (Key:=locT) (Elt:=option valueT) (M:=memT) := _.

Lemma sem_expr_proper' :
  ∀ x y : pidxT, x ≡ y
  → ∀ a b : exprT, a ≡ b
  → ∀ z w : memT, z ≡ w 
  → ∀ c d : valueT, c ≡ d 
  → sem_expr x a z c
  → sem_expr y b w d
with sem_lv_proper' :
  ∀ x y : pidxT, x ≡ y
  → ∀ a b : lv, a ≡ b
  → ∀ z w : memT, z ≡ w 
  → ∀ c d : locT, c ≡ d 
  → sem_lv x a z c
  → sem_lv y b w d.
Proof.
  move=> x y Hxy a b Hab z w Hzw c d Hcd Hsem.
  usual_tac Hxy.
  usual_tac Hab.
  usual_tac Hcd.
  move:w Hzw. induction Hsem; intros.

  (* expr_const *)
  { constructor.
    eapply eval_lit_proper.
    apply Hzw. reflexivity. reflexivity. assumption. }

  (* op_add *)
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { apply sem_op_add_za.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { apply sem_op_add_az.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }

  (* op_sub *)
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { apply sem_op_sub_az.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }

  (* op_others in op2 *)
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }
  { constructor.
    eapply IHHsem1. apply Hzw. 
    eapply IHHsem2. apply Hzw. }

  (* expr_op1 *)
  { constructor.
    eapply IHHsem. apply Hzw. }
  { constructor.
    eapply IHHsem. apply Hzw. }

  (* expr_value_of *)
  { apply sem_value_of with lval.
    apply sem_lv_proper' with y l m lval.
    reflexivity. reflexivity. apply Hzw. reflexivity. assumption.
    rewrite <- Hzw. assumption. }

  (* expr_addr_of *)
  { apply sem_addr_of.
    apply sem_lv_proper' with y l m lval.
    reflexivity. reflexivity. apply Hzw. reflexivity. assumption. }

  (* expr_cast *)
  { apply sem_cast_z_in.
    apply IHHsem. assumption.
    assumption. }
  { apply sem_cast_z_out with z.
    apply IHHsem. assumption.
    assumption.
    assumption. }
  { apply sem_cast_array.
    apply IHHsem. assumption. assumption. }

  (* expr_size_of *)
  { constructor. }

  move=> x y Hxy a b Hab z w Hzw c d Hcd Hsem.
  usual_tac Hxy.
  usual_tac Hab.
  usual_tac Hcd.
  move:w Hzw. induction Hsem; intros.

  (* lv_id *)
  { constructor. }

  (* lv_deref *)
  { constructor. } 
  { constructor.
    by eapply sem_expr_proper'; eauto; reflexivity.
  }
  { apply lv_deref_array with sz st.
    by eapply sem_expr_proper'; eauto; reflexivity.
  }
    
  (* lv_array *)
  { apply lv_array with sz st.
    by eapply sem_expr_proper'; last exact Harr; eauto; reflexivity.
    by eapply sem_expr_proper'; last exact Hidx; eauto; reflexivity.
  }

  (* lv_struct *)
  { apply lv_strct.
    by apply IHHsem.
  }
Qed.

Lemma sem_expr_proper : Proper (eq ==> eq ==> eq ==> eq ==> le) sem_expr.
Proof.
  move=> x y Hxy z w Hzw a b Hab c d Hcd Hsem.
  eapply sem_expr_proper'; eauto.
Qed.

Lemma sem_lv_proper : Proper (eq ==> eq ==> eq ==> eq ==> le) sem_lv.
Proof.
  move=> x y Hxy z w Hzw a b Hab c d Hcd Hsem.
  eapply sem_lv_proper'; eauto.
Qed.

Add Parametric Morphism : sem_expr with
    signature eq ==> eq ==> eq ==> eq ==> iff as sem_expr_mor.
Proof.
  move=> x y Hxy z w Hzw a b Hab c d Hcd.
  by split=> Heq; eapply sem_expr_proper; eauto; apply/eq_sym.
Qed.

Inductive sem_exprs (pidx:pidxT) : list exprT -> memT -> list valueT -> Prop :=
| sem_exprs_nil : forall mem, sem_exprs pidx nil mem nil
| sem_exprs_cons :
    forall mem e v exprs vals
           (Hexpr: sem_expr pidx e mem v)
           (Hexprs: sem_exprs pidx exprs mem vals),
      sem_exprs pidx (e::exprs) mem (v::vals).

Lemma sem_exprs_proper : Proper (eq ==> eq ==> eq ==> eq ==> le) sem_exprs.
Proof.
  move=> lp rp Hp lle rle Hle lm rm Hm llv rlv Hlv.
  usual_tac Hle; usual_tac Hlv.
  elim: rle rlv => [|e es IH] [|v vs] Hsem.
  apply sem_exprs_nil.
  inversion Hsem.
  inversion Hsem.
  inversion Hsem; subst; apply sem_exprs_cons.
  by rewrite <- Hp; rewrite <- Hm.
  exact/IH.
Qed.

Add Parametric Morphism : sem_exprs with
    signature eq ==> eq ==> eq ==> eq ==> iff as sem_exprs_proper_added.
Proof.
  move=> x y Hxy z w Hzw a b Hab c d Hcd.
    by split=> Heq; eapply sem_exprs_proper; eauto; apply/eq_sym.
Qed.    

Inductive assume (pidx:pidxT) : exprT -> memT -> Prop :=
| assume_intro : forall mem v e
                        (Hsem: sem_expr pidx e mem (vint v))
                        (Hnonzero: ~ eq v 0%Z),
                   assume pidx e mem.

Lemma assume_proper : Proper (eq ==> eq ==> eq ==> le) assume.
Proof.
  move=> lp rp Hp le re Hle lm rm Hm.
  usual_tac Hle.
  move=> Hassume; inversion Hassume; subst; eapply assume_intro; eauto.
  by rewrite <- Hp; rewrite <- Hm.
Qed.    

Add Parametric Morphism : assume with
    signature eq ==> eq ==> eq ==> iff as assume_mor.
Proof.
  move=> x y Hxy z w Hzw a b Hab.
    by split=> Heq; eapply assume_proper; eauto; apply/eq_sym.
Qed.

Definition var_loc (var:varT) (pidx:pidxT) : locT :=
  match var with
    | inl gvar => lglobal gvar
    | inr lvar => llocal pidx lvar
  end.

Global Add Parametric Morphism : var_loc with
    signature eq ==> eq ==> eq as var_loc_mor.
Proof.
  move=> x y Hxy; usual_tac Hxy.
  move=> lp rp Hp; usual_tac Hp.
  reflexivity.
Qed.

Definition fresh_loc (u:uniqueT) (mem:memT): Prop := 
  forall a o, mem[lheap a u o]! = None.

Global Add Parametric Morphism : fresh_loc with
  signature eq ==> eq ==> eq as fresh_loc_mor.
Proof.
  move => x u Hxy; usual_tac Hxy.
  move => x y Hxy. rewrite/fresh_loc; split; intros.
  by rewrite <- Hxy. by rewrite -> Hxy. 
Qed.  


Section CorLocs.
  Definition t_loc (loc:locT) : DomAbsValue.locT :=
    match loc with
      | (inl (inl (a, _, _)), fs) => DomAbsValue.lheap a fs
      | (inl (inr gvar), fs) => DomAbsValue.lglobal gvar fs
      | (inr (fid, bid, depth), fs) => DomAbsValue.llocal fid bid fs
    end.

  Add Parametric Morphism : t_loc with
    signature eq ==> eq as t_loc_mor.
  Proof.
    move=> x y Hxy.
    usual_tac Hxy.
    reflexivity.
  Qed.

  Definition Tloc (loc:locT) (loc':DomAbsValue.locT) := eq (t_loc loc) loc'.

  Add Parametric Morphism : Tloc with
    signature eq ==> eq ==> eq as Tloc_mor.
  Proof.
    intros. 
    usual_tac H.
    usual_tac H0.
    reflexivity.
  Qed.

  Program Definition gamma_loc (aloc:DomAbsValue.locT) : \powerset locT :=
    fun loc => eqb (t_loc loc) aloc.
  Next Obligation.
    by move=> x y Hxy; usual_tac Hxy.
  Qed.
  
  Program Definition t_locs (locs : \powerset DomAbsValue.locT) : \powerset locT :=
    fun loc=> (t_loc loc) !\in locs.
  Next Obligation.
    move=> x y Hxy.
    usual_tac Hxy.
    exact: eq_refl.
  Qed.

  Definition Tlocs (locs : \powerset locT) (locs' : \powerset DomAbsValue.locT) := eq locs (t_locs locs').

  Lemma t_locs_spec (loc:locT) (locs: \powerset DomAbsValue.locT) :
    eq (loc !\in (t_locs locs)) (t_loc loc !\in locs).
  Proof.
    rewrite/t_locs/=.
    reflexivity.
  Qed.
End CorLocs.


Inductive libcall (pidx:pidxT) : option lv -> lcall -> list exprT -> 
  allocT -> memT -> memT -> Prop := 

(* airac_print *)
| sem_libcall_airac_print : forall ret args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx ret airac_print args site mem newmem

(* airac_dump *)
| sem_libcall_airac_dump : forall ret args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx ret airac_dump args site mem newmem

(* rand *)
| sem_libcall_rand_some : forall ret ret_loc z args site mem newmem
  (Hret:sem_lv pidx ret mem ret_loc)
  (Hz:(0 <= z)%Z)
  (Heq:eq mem[ret_loc <- Some (vint z)]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) rand args site mem newmem
| sem_libcall_rand_none  : forall args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None rand args site mem newmem

(* strlen *)
| sem_libcall_strlen_some : forall ret ret_loc z args site mem newmem
  (Hret:sem_lv pidx ret mem ret_loc)
  (Hz:(0 <= z)%Z)
  (Heq:eq mem[ret_loc <- Some (vint z)]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) strlen args site mem newmem
| sem_libcall_strlen_none  : forall args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None strlen args site mem newmem

(* strcpy *)
| sem_libcall_strcpy_some : forall dst_loc src_loc v
  ret ret_loc args site mem mem' newmem
  (Hdst:sem_expr pidx (hd (expr_const lit_bot) args) mem 
    (vloc dst_loc))
  (Hsrc:sem_expr pidx (hd (expr_const lit_bot) (tl args)) mem 
    (vloc src_loc))
  (Hsrc: mem[src_loc ]! = Some v)
  (Hcopy: eq (mem[dst_loc <- (Some v : option valueT) ]! ) mem')
  (Hwf_mem': mem_wf pidx mem')  
  (Hret:sem_lv pidx ret mem' ret_loc)
  (Heq: eq mem'[ret_loc <- Some (vloc dst_loc)]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) strcpy args site mem newmem
| sem_libcall_strcpy_none : forall dst_loc src_loc v args site mem newmem
  (Hdst:sem_expr pidx (hd (expr_const lit_bot) args) mem 
    (vloc dst_loc))
  (Hsrc:sem_expr pidx (hd (expr_const lit_bot) (tl args)) mem 
    (vloc src_loc))
  (Hsrc: mem[src_loc ]! = Some v)
  (Hcopy: eq (mem[dst_loc <- (Some v : option valueT) ]! ) newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None strcpy args site mem newmem

(* scanf *)
| sem_libcall_scanf_nil : forall site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None scanf nil site mem newmem
| sem_libcall_scanf_cons : forall x x_loc z args site mem mem' newmem
  (IH:libcall pidx None scanf args site mem mem')
  (Hwf_mem': mem_wf pidx mem')
  (Hx:sem_expr pidx x mem (vloc x_loc))
  (Heq: eq mem'[x_loc <- Some (vint z%Z)]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None scanf (cons x args) site mem newmem

(* errno_location *)
| sem_libcall_errno_location_some : forall ret ret_loc args site u mem newmem
  (Hret:sem_lv pidx ret mem ret_loc)
  (Heq: eq mem[ret_loc <- Some (vheap site u (0%Z,1%Z,4%Z))]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) errno_location args site mem newmem
| sem_libcall_errno_location_none : forall args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None errno_location args site mem newmem

(* builtin_alloca *)
| sem_libcall_builtin_alloca_some : forall ret ret_loc args sz site u mem newmem
  (Hret:sem_lv pidx ret mem ret_loc)
  (Hsz:sem_expr pidx (hd (expr_const lit_bot) args) mem (vint sz))
  (Hfresh:fresh_loc u mem)
  (Heq: eq mem[ret_loc <- Some (vheap site u (0%Z,sz%Z,1%Z))]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) builtin_alloca args site mem newmem
| sem_libcall_builtin_alloca_none : forall args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None builtin_alloca args site mem newmem

(* realloc *)
(* skcho: we give up writing semantics of realloc.  So the semantics
of realloc is just the idle function which is written in the unknown
case below. *)

(* builtin_choose_expr *)
| sem_libcall_builtin_choose_expr_none : forall args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx None builtin_choose_expr args site mem newmem
| sem_libcall_builtin_choose_expr_some_true : forall ret ret_loc 
  args cnd v site mem newmem
  (Hdst:sem_expr pidx (hd (expr_const lit_bot) args) mem (vint cnd))
  (Hsrc:sem_expr pidx (hd (expr_const lit_bot) (tl args)) mem v)
  (Hcnd:cnd <> 0%Z)
  (Hret:sem_lv pidx ret mem ret_loc)
  (Heq: eq mem[ret_loc <- Some v]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) builtin_choose_expr args site mem newmem
| sem_libcall_builtin_choose_expr_some_false : forall ret ret_loc 
  args cnd v site mem newmem
  (Hdst:sem_expr pidx (hd (expr_const lit_bot) args) mem (vint cnd))
  (Hsrc:sem_expr pidx (hd (expr_const lit_bot) (tl (tl args))) mem v)
  (Hcnd:cnd = 0%Z)
  (Hret:sem_lv pidx ret mem ret_loc)
  (Heq: eq mem[ret_loc <- Some v]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx (Some ret) builtin_choose_expr args site mem newmem

(* unknown *)
| sem_libcall_unknown : forall ret args site mem newmem
  (Heq: eq mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  libcall pidx ret unknown args site mem newmem.

Global Add Parametric Morphism : libcall with
  signature eq ==> eq ==> eq ==> eq ==> eq ==> eq ==> eq ==> le as libcall_le_mor.
Proof.
  move => px pidx Hpidx; usual_tac Hpidx.
  move => lx ret Hoption_lv; usual_tac Hoption_lv.
  move => lcx lib Hlcall; usual_tac Hlcall.
  move => ex args Hexpr; usual_tac Hexpr.
  move => stx site Hsite; usual_tac Hsite.
  move => m m' Hmem newm newm' Hnewm Hlib.
  move : m' newm' Hmem Hnewm. induction Hlib; intros; subst.
  { apply sem_libcall_airac_print. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_airac_dump.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_rand_some with ret_loc z; auto.
    move:Hret. apply sem_lv_proper; try reflexivity. assumption.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_rand_none.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_strlen_some with ret_loc z; auto.
    move:Hret. apply sem_lv_proper; try reflexivity. assumption.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_strlen_none.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_strcpy_some with dst_loc src_loc v ret_loc mem'; auto.
    move:Hdst. apply sem_expr_proper; try reflexivity. assumption.
    move:Hsrc. apply sem_expr_proper; try reflexivity. assumption.
    rewrite <- Hmem. assumption. 
    rewrite <- Hmem. assumption.
    rewrite <- Hnewm. assumption.
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_strcpy_none with dst_loc src_loc v.
    move:Hdst. apply sem_expr_proper; try reflexivity. assumption.
    move:Hsrc. apply sem_expr_proper; try reflexivity. assumption.
    rewrite <- Hmem. assumption. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_scanf_nil.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_scanf_cons with x_loc z mem'.
    apply IHHlib; auto. reflexivity.
    assumption.
    rewrite <- Hmem. assumption.
    rewrite <- Hnewm. assumption.
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_errno_location_some with ret_loc u.
    move:Hret. apply sem_lv_proper; try reflexivity. assumption.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_errno_location_none.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_builtin_alloca_some with ret_loc sz u.
    move:Hret. apply sem_lv_proper; try reflexivity. assumption.
    rewrite <- Hmem. assumption. 
    rewrite <- Hmem. assumption.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_builtin_alloca_none. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_builtin_choose_expr_none. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_builtin_choose_expr_some_true with ret_loc cnd v; auto. 
    move:Hdst. apply sem_expr_proper; auto; try reflexivity. 
    move:Hsrc. apply sem_expr_proper; auto; try reflexivity. 
    move:Hret. apply sem_lv_proper; auto; try reflexivity. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_builtin_choose_expr_some_false with ret_loc 0%Z v; auto.
    move:Hdst. apply sem_expr_proper; auto; try reflexivity. 
    move:Hsrc. apply sem_expr_proper; auto; try reflexivity. 
    move:Hret. apply sem_lv_proper; auto; try reflexivity. 
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
  { apply sem_libcall_unknown.
    rewrite <- Hmem. rewrite <- Hnewm. assumption. 
    rewrite <- Hnewm. assumption. }
Qed.

Global Add Parametric Morphism : libcall with
  signature eq ==> eq ==> eq ==> eq ==> eq ==> eq ==> eq ==> eq as libcall_eq_mor.
Proof.
  intros. split. apply libcall_le_mor; auto.
  apply libcall_le_mor; symmetry; auto.
Qed.

Inductive sem_instr (pidx:pidxT) : instrT -> memT -> memT -> Prop :=
| sem_instr_assign :
    forall mem l e lc val newmem
           (Hl:sem_lv pidx l mem lc)
           (Hexpr: sem_expr pidx e mem val)
           (Hsubst: eq mem[lc <- Some val]! newmem)
           (Hwf_newmem: mem_wf pidx newmem),
      sem_instr pidx (instr_assign l e) mem newmem
| sem_instr_alloc_arr : forall l lc sze sz ste st site u mem newmem
  (Hl:sem_lv pidx l mem lc)
  (Hsz:sem_expr pidx sze mem (vint sz))
  (Hst:sem_expr pidx ste mem (vint st))
  (Hfresh:fresh_loc u mem)
  (Hsubst:eq mem[lc <- Some (vheap site u (0%Z, sz, st))]! newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  sem_instr pidx (instr_alloc l (alloc_arr sze ste) site) mem newmem
| sem_instr_assume :
    forall mem newmem e
           (Hassume: assume pidx e mem)
           (Heq: eq mem newmem)
           (Hwf_newmem: mem_wf pidx newmem),
      sem_instr pidx (instr_assume e) mem newmem
| sem_instr_libcall : forall ret lib args site mem newmem
  (Hlcall:libcall pidx ret lib args site mem newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  sem_instr pidx (instr_libcall ret lib args site) mem newmem
| sem_instr_forget : forall aloc mem newmem
  (Hforget:eq (mrestrict (K:=locT) (E:=option valueT) (M:=memT) (powerset_neg (gamma_loc aloc)) mem) newmem)
  (Hwf_newmem: mem_wf pidx newmem),
  sem_instr pidx (instr_forget aloc) mem newmem
| sem_instr_skip :
    forall mem newmem (Heq: eq mem newmem)
      (Hwf_newmem: mem_wf pidx newmem),
      sem_instr pidx instr_skip mem newmem.

Lemma sem_instr_proper : Proper (eq ==> eq ==> eq ==> eq ==> le) sem_instr.
Proof.
  move=> x y Hxy a b Hab z w Hzw c d Hcd.
  usual_tac Hxy. usual_tac Hab.
  move=> Hsem; inversion Hsem; subst.
  { apply sem_instr_assign with lc val.
    move :Hl. apply sem_lv_proper; try reflexivity. assumption.
    move :Hexpr. apply sem_expr_mor; try reflexivity.
    by symmetry.
    rewrite <- Hzw. rewrite <- Hcd. assumption.
    rewrite <- Hcd. assumption. }
  { apply sem_instr_alloc_arr with lc sz st u.
    move :Hl. apply sem_lv_proper; try reflexivity. 
    assumption.
    by rewrite <- Hzw. 
    by rewrite <- Hzw. 
    by rewrite <- Hzw. 
    rewrite <- Hzw. rewrite <- Hcd. assumption.
    rewrite <- Hcd. assumption. }
  { apply sem_instr_assume.
    by rewrite <- Hzw.
    rewrite <- Hzw. rewrite <- Hcd. assumption. 
    rewrite <- Hcd. assumption. }
  { apply sem_instr_libcall.
    move:Hlcall. apply libcall_le_mor; auto; reflexivity.
    rewrite <- Hcd. assumption. }
  { apply sem_instr_forget.
    rewrite <- Hzw.
    by rewrite <- Hcd.
    by rewrite <- Hcd. }
  { apply sem_instr_skip.
    rewrite <- Hzw. rewrite <- Hcd. assumption.
    rewrite <- Hcd. assumption. }
Qed.

Add Parametric Morphism : sem_instr with
    signature eq ==> eq ==> eq ==> eq ==> iff as sem_instr_mor.
Proof.
  move=> x y Hxy a b -> z w Hzw c d Hcd.
  by split=> Heq; eapply sem_instr_proper; eauto; apply/eq_sym.
Qed.

Inductive sem_instrs (pidx:pidxT) : list instrT -> memT -> memT -> Prop :=
| sem_instrs_nil :
    forall mem mequiv
           (Heq: eq mem mequiv)
           (Hwf_mequiv: mem_wf pidx mequiv),
      sem_instrs pidx nil mem mequiv
| sem_instrs_cons :
    forall mem mmid mfin i li
           (Hinstr: sem_instr pidx i mem mmid)
           (Hwf_mmid: mem_wf pidx mmid)
           (Hinstrs: sem_instrs pidx li mmid mfin)
           (Hwf_mfin: mem_wf pidx mfin),
      sem_instrs pidx (i::li) mem mfin.

Lemma sem_instrs_proper : Proper (eq ==> eq ==> eq ==> eq ==> le) sem_instrs.
Proof.
  move=> lp rp Hp lle rle Hle lm rm Hm.
  usual_tac Hle.
  elim: rle lm rm Hm => [|e es IH] lm rm Hm x y Hxy Hsem; inversion Hsem; subst; clear Hsem.
  constructor. 
  by rewrite <- Hm; rewrite -> Heq.
  by rewrite <- Hp; rewrite <- Hxy.
  econstructor; first by rewrite <- Hp; rewrite <- Hm; eauto.
  by rewrite <- Hp.
  by apply/IH; eauto; apply/eq_refl.
  by rewrite <- Hp; rewrite <- Hxy.
Qed.

Add Parametric Morphism : sem_instrs with
    signature eq ==> eq ==> eq ==> eq ==> iff as sem_instrs_mor.
Proof.
  move=> x y Hxy z w Hzw a b Hab c d Hcd.
  by split=> Heq; eapply sem_instrs_proper; eauto; apply/eq_sym.
Qed.

Inductive sem_bind_params (pidx:pidxT) :
  list valueT -> list idT -> memT -> memT -> Prop :=
| sem_bind_params_nil :
    forall mem mequiv
           (Heq: eq mem mequiv)
           (Hwf_mequiv: mem_wf pidx mequiv),
      sem_bind_params pidx nil nil mem mequiv
| sem_bind_params_cons :
    forall mem arg param mmid args params mfin
           (Hsubst: eq mem[var_loc (local param) pidx <- Some arg]! mmid)
           (Hwf_mmid: mem_wf pidx mmid)
           (Hbind: sem_bind_params pidx args params mmid mfin)
           (Hwf_mfin: mem_wf pidx mfin),
      sem_bind_params pidx (arg::args) (param::params) mem mfin.

Lemma sem_bind_params_proper : Proper (eq ==> eq ==> eq ==> eq ==> eq ==> impl)
                                      sem_bind_params.
Proof.
  move=> lp rp Hp llv rlv Hlv lli rli Hli lm rm Hm lnm rnm Hnm.
  usual_tac Hp.
  usual_tac Hlv.
  usual_tac Hli.
  elim: rlv rli lm rm Hm lnm rnm Hnm=> [|rv rlv IH] rli lm rm Hm lnm rnm Hnm Hsem.
  { inv Hsem.
    constructor.
    by rewrite <- Hm; rewrite Heq.
    by rewrite <- Hnm.
  }
  { inv Hsem.
    econstructor; last first.
    { by rewrite <- Hnm. }
    { eapply IH; last first.
      by apply Hbind.
      done.
      by apply eq_sym.
    }
    { by rewrite Hsubst. }
    { rewrite Hm.
      exact: eq_refl.
    }
  }
Qed.

Add Parametric Morphism : sem_bind_params with
    signature eq ==> eq ==> eq ==> eq ==> eq ==> iff
      as sem_bind_params_mor.
Proof.
  move=> lp rp Hp llv rlv Hlv lli rli Hli lm rm Hm lnm rnm Hnm.
  by split=> Hsem; eapply sem_bind_params_proper; eauto; apply: eq_sym.
Qed.

Definition sem_bind_ret (pidx:pidxT) (result: valueT) (retl: option varT) (mem mnew: memT) : Prop :=
  (if retl is Some retl'
   then eq mem[var_loc retl' pidx <- Some result]! mnew
   else eq mem mnew) /\
  mem_wf pidx mnew.

Lemma sem_bind_ret_proper : Proper (eq ==> eq ==> eq ==> eq ==> eq ==> impl)
                                   sem_bind_ret.
Proof.
  move=> lp rp Hp lv rv Hv lvar rvar Hvar lm rm Hm lnm rnm Hnm Hsem.
  usual_tac Hp.
  usual_tac Hv.
  usual_tac Hvar.
  inv Hsem.
  split; last by rewrite <- Hnm.
  move: rvar H=> [rvar|] H.
  { by rewrite <- Hm; rewrite H. }
  { by rewrite <- Hm; rewrite H. }
Qed.  

Add Parametric Morphism : sem_bind_ret with
    signature eq ==> eq ==> eq ==> eq ==> eq ==> iff
      as sem_bind_ret_proper_added.
Proof.
  move=> lp rp Hp lv rv Hv lvar rvar Hvar lm rm Hm lnm rnm Hnm.
  by split=> Hsem; eapply sem_bind_ret_proper; eauto; apply: eq_sym.
Qed.

Program Definition filter_nth (n:nat) : \powerset locT :=
  fun loc =>
    if loc is (inr (_,_,m), _)
    then leb m n
    else true.
Next Obligation.
  move=> [[[x|x]|x] xfs] [[[y|y]|y] yfs] // [Hxy Hxyfs];
  rewrite/fst/snd in Hxy Hxyfs;
  try by inv Hxy.
  move: x y Hxy=> [[xf xv] xd] [[yf yv] yd] Hxy.
  by usual_tac Hxy; inv Hxy.
Qed.

Inductive sem_step (prog:progT) : stateT -> stateT -> Prop :=
| sem_step_instrs :
    forall fid bid stack nstack instrs mold mnew
           (Hpidx: piinstrs prog (fid, bid, false, stack) = Some instrs)
           (Hstack: eq stack nstack)
           (Hinstrs: sem_instrs (fid, bid, false, stack) instrs mold mnew)
           (Hwf_mnew: mem_wf (fid, bid, true, nstack) mnew),
      sem_step prog (fid, bid, false, stack, mold)
               (fid, bid, true, nstack, mnew)

| sem_step_jump :
    forall nbids nbid fid bid stack mem mequiv
           (Htmn: pitmn prog (fid, bid, true, stack) = Some (tmn_jump nbids))
           (Hnbid: SetInterface.mem nbid nbids)
           (Heq: eq mem mequiv)
           (Hwf_mnew: mem_wf (fid, nbid, false, stack) mequiv),
      sem_step prog (fid, bid, true, stack, mem)
               (fid, nbid, false, stack, mequiv)

| sem_step_call :
    forall nfid nfids nbid args argvals retto nfunc stack fid bid mold mnew
           (Htmn: pitmn prog (fid, bid, true, stack)
                  = Some (tmn_call nbid nfids args retto))
           (Hargs: sem_exprs (fid, bid, true, stack) args mold argvals)
           (Hnfid: nfid !\in nfids)
           (Hfunc: MapsTo nfid nfunc (pfuncs prog))
           (Hbind: sem_bind_params
                     (nfid, fentry nfunc, false, (fid, bid)::stack) argvals (fparams nfunc) mold mnew)
           (Hwf_mnew: mem_wf (nfid, fentry nfunc, false, (fid,bid)::stack) mnew),
      sem_step prog (fid, bid, true, stack, mold)
               (nfid, fentry nfunc, false, (fid, bid)::stack, mnew)

| sem_step_exit :
    forall pfid pbid stack ret retval pfunc fid fids bid nbid args retto mold mmid mnew
           (Hcmd: pitmn prog (fid, bid, true, (pfid, pbid)::stack)
                  = Some (tmn_exit ret))
           (Hret: sem_expr (fid, bid, true, (pfid, pbid)::stack) ret mold retval)
           (Hpfunc: MapsTo pfid pfunc (pfuncs prog))
           (Hcall: pitmn prog (pfid, pbid, true, stack)
                   = Some (tmn_call nbid fids args retto))
           (Hnfid: fid !\in fids)
           (Hmmid: mmid = mrestrict (K:=locT) (E:=option valueT) (M:=memT) (filter_nth (length stack)) mold)
           (Hwf_mmid: mem_wf (pfid, nbid, false, stack) mmid)
           (Hbind: sem_bind_ret (pfid, pbid, false, stack) retval retto mmid mnew)
           (Hwf_mnew: mem_wf (pfid, nbid, false, stack) mnew),
      sem_step prog (fid, bid, true, (pfid, pbid)::stack, mold)
               (pfid, nbid, false, stack, mnew).

Lemma sem_step_proper (prog:progT) : Proper (eq ==> eq ==> le) (sem_step prog).
Proof.
  move=> [lpidx lmem] [rpidx rmem] [Hpidx Hmem]
                      [lnpidx lnmem] [rnpidx rnmem] [Hnpidx Hnmem].
  repeat match goal with
           | [H: context[(?a,?b).1] |- _] => rewrite -[(a,b).1]/a in H
           | [H: context[(?a,?b).2] |- _] => rewrite -[(a,b).2]/b in H
         end.
  usual_tac Hpidx; usual_tac Hnpidx.
  move=> Hsem; inversion Hsem; subst.
  { eapply sem_step_instrs; eauto.
    by rewrite <- Hmem; rewrite <- Hnmem; eauto.
    by rewrite <- Hnmem.
  }
  { eapply sem_step_jump; eauto.
    by rewrite <- Hmem; rewrite <- Hnmem.
    by rewrite <- Hnmem.
  }
  { eapply sem_step_call; eauto.
    by rewrite <- Hmem; eauto.
    by rewrite <- Hmem; rewrite <- Hnmem.
    by rewrite <- Hnmem.
  }
  { eapply sem_step_exit; eauto.
    by rewrite <- Hmem; eauto.
    by rewrite <- Hmem.
    by rewrite <- Hmem; rewrite <- Hnmem.
    by rewrite <- Hnmem.
  }
Qed.

Add Parametric Morphism (prog:progT) : (sem_step prog) with
    signature eq ==> eq ==> iff as sem_step_mor.
Proof.
  move=> x y Hxy z w Hzw; split=> Hsem.
    by apply/sem_step_proper; last by eauto.
    apply/sem_step_proper; last by eauto.
    exact/eq_sym.
    exact/eq_sym.
Qed.

Inductive sem (prog:progT) : nat -> stateT -> Prop :=
| sem_initial:
    forall pidx pidxi mem
           (Hentry: (piinitial prog) = Some pidxi)
           (Hpieq: eq pidx pidxi)
           (Hmemeq: eq mem bot),
      sem prog 0 (pidx, mem)

| sem_progress:
    forall step midstate newstate
           (Hprog: sem prog step midstate)
           (Hwf_mid: mem_wf midstate.1 midstate.2)
           (Hstep: sem_step prog midstate newstate)
           (Hwf: mem_wf newstate.1 newstate.2),
      sem prog (S step) newstate.

Lemma sem_proper (prog:progT) : Proper (eq ==> eq ==> le) (sem prog).
Proof.
  move=> x y Hxy z w Hzw.
  rewrite -> Hxy; move=> {x Hxy}.
  move=> Hsem; inversion Hsem; subst.
  { case: w Hzw => pidx' mem' [Hpidx Hmem].
    eapply sem_initial; eauto.
    by rewrite <- Hpidx.
    by rewrite <- Hmem.
  }
  { eapply sem_progress; eauto.
    by rewrite <- Hzw.
    move: Hzw=> [Hpidx Hmem].
    by rewrite <- Hpidx; rewrite <- Hmem.
  }
Qed.

Add Parametric Morphism (prog:progT) : (sem prog) with
    signature eq ==> eq ==> iff as sem_mor.
Proof.
  move=> x y Hxy z w Hzw.
  split=> Hsem.
    by apply/sem_proper; last by eauto.
    apply/sem_proper; last by eauto.
    exact/eq_sym.
    exact/eq_sym.
Qed.  

Lemma sem_consistent (prog:progT) (step:nat) (state:stateT)
  (Hsem : sem prog step state) :
  pidx_consistent prog (pifid state.1) (state.1.2).
Proof.
  elim: Hsem.
  { rewrite/piinitial/=.
    case Hentry: (pfuncs prog)[pentry prog]=> [func|] pidx' pidx mem Hinitial Hpidx Hmem //=.
    usual_tac Hpidx.
    move: pidx Hinitial=> [[[fid bid] epilogue] stack] //= Hinitial.
    inversion Hinitial; subst; clear Hinitial.
    exact/eqbP/eq_refl.
  }
  { move=> depth [midpidx midmem] [newpidx newmem] Hsem IH Hwf_mid Hstep Hwf.
    inversion Hstep; subst; clear Hstep; rewrite/= in IH *; auto.
    { by rewrite <- Hstack. }
    { apply/andP; split; last done.
      move: Htmn; rewrite/pitmn/picmd/pifunc/fcmds/ctmn/=.
      case: (pfuncs prog)[fid]=> [func|]; last done.
      case: (func.1.2)[bid]=> [[instrs tmn] Hcmd|]; last done.
      by inv Hcmd.
    }
    { by move/andP: IH=> [_ ?]. }
  }
Qed.
