Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import PArith.

Require Import DomType.
Require Import DomBasic.
Require Import DomZ.
Require Import DomSet.
Require Import DomMap.
Require Import DomAbsValue.

Set Implicit Arguments.

Definition global := nosimpl (fun (id:idT) => inl id : varT).
Definition local := nosimpl (fun (id:idT) => inr id : varT).

Inductive literal :=
| lit_int (n:Z)
| lit_int_top
| lit_top
| lit_bot.

Definition literal_eqb (lhs rhs : literal) : bool :=
  match lhs, rhs with
    | lit_int ln, lit_int rn => eqb ln rn
    | lit_int_top, lit_int_top => true
    | lit_top, lit_top => true
    | lit_bot, lit_bot => true
    | _, _ => false
  end.

Program Instance literal_setoid : setoid literal := {}.
Program Instance literal_usual_setoid : @usual_setoid literal _ := {}.
Program Instance literal_setoidDec : @setoidDec literal _ := {
  eqb_ := literal_eqb
}.
Next Obligation.
  case: x; case: y; try by constructor.
  simpl; intros.
  remember (eqb n0 n)%Z as b; destruct b; symmetry in Heqb.
  move/eqbP:Heqb => Heqb; rewrite Heqb; by constructor.
  move/eqbP:Heqb => Heqb. constructor. intro. elim Heqb. by inversion H.
Qed.

Inductive ctype :=
| typ_uint | typ_uchar | typ_ushort | typ_ulong | typ_ulonglong(* unsigned *)
| typ_int | typ_char | typ_short | typ_long | typ_longlong (* integer *)
| typ_struct (s:varT) (* named structure *)
| typ_union  (s:varT)  (* named union *)
| typ_pointer (t:ctype) (* pointer *)
| typ_array  (t:ctype) (idx:nat) (* array *)
| typ_function (t:ctype)  (* function, return type *)
| typ_void
| typ_va_list   (* builtin_va_list *)
| typ_unknown.   (* ellipses, etc.. *)

Fixpoint ctype_eqb (lhs rhs : ctype) : bool := 
  match lhs, rhs with
    | typ_uint, typ_uint => true
    | typ_uchar, typ_uchar => true
    | typ_ushort, typ_ushort => true
    | typ_ulong, typ_ulong => true
    | typ_ulonglong, typ_ulonglong => true
    | typ_int, typ_int => true
    | typ_char, typ_char => true
    | typ_short, typ_short => true
    | typ_long, typ_long => true
    | typ_longlong, typ_longlong => true
    | typ_struct ls, typ_struct rs => eqb ls rs
    | typ_union ls, typ_union rs => eqb ls rs
    | typ_pointer lct, typ_pointer rct => ctype_eqb lct rct
    | typ_array lct ln, typ_array rct rn => 
      ctype_eqb lct rct && eqb ln rn
    | typ_function lct, typ_function rct => ctype_eqb lct rct
    | typ_void, typ_void => true
    | typ_va_list, typ_va_list => true
    | typ_unknown, typ_unknown => true
    | _, _ => false
  end.

Program Instance ctype_setoid : setoid ctype := {}.
Program Instance ctype_usual_setoid : @usual_setoid ctype _ := {}.
Program Instance ctype_setoidDec : @setoidDec ctype _ := {
  eqb_ := ctype_eqb
}.
Next Obligation.
  move: y. elim x. 
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.

  simpl; intros; case: y; (try by constructor); intros.
  remember (eqb s s0) as b; destruct b; symmetry in Heqb.
  move/eqbP:Heqb => Heqb. constructor.
  unfold eq; simpl. unfold Equivalence.equiv; f_equal. by rewrite Heqb.
  move/eqbP:Heqb => Heqb. constructor. intro. elim Heqb. 
  unfold eq in H; simpl in H. 
  unfold Equivalence.equiv in H; inversion H; reflexivity.

  simpl; intros; case: y; (try by constructor); intros.
  remember (eqb s s0) as b; destruct b; symmetry in Heqb.
  move/eqbP:Heqb => Heqb. constructor.
  unfold eq; simpl. unfold Equivalence.equiv; f_equal. by rewrite Heqb.
  move/eqbP:Heqb => Heqb. constructor. intro. elim Heqb. 
  unfold eq in H; simpl in H. 
  unfold Equivalence.equiv in H; inversion H; reflexivity.

  simpl; intros; case: y; (try by constructor); intros.
  remember (H t0). clear Heqr.
  remember (ctype_eqb t t0) as b; destruct b; symmetry in Heqb.
  inversion r. constructor. rewrite H0. reflexivity.
  inversion r. constructor. intro. elim H0. inversion H1; reflexivity.

  simpl; intros; case: y; (try by constructor); intros.
  remember (H t0). clear Heqr.
  remember (ctype_eqb t t0) as b; destruct b; symmetry in Heqb.
  remember (eqb idx idx0) as b; destruct b; symmetry in Heqb.
  inversion r. constructor. 
  rewrite H0. symmetry in Heqb0. move/eqbP:Heqb0 => Heqb0. rewrite Heqb0.
  reflexivity.
  symmetry in Heqb0. move/eqbP:Heqb0 => Heqb0. 
  constructor. intro. elim Heqb0. inversion H0; reflexivity.
  inversion r. constructor. intro. elim H0. inversion H1; reflexivity.

  simpl; intros; case: y; (try by constructor); intros.
  remember (H t0). clear Heqr.
  remember (ctype_eqb t t0) as b; destruct b; symmetry in Heqb.
  inversion r. constructor. rewrite H0. reflexivity.
  inversion r. constructor. intro. elim H0. inversion H1; reflexivity.

  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
  move => y; case: y; try by constructor.
Qed.

Inductive exprT := 
| expr_const (l:literal)
| expr_op2 (op:op2) (lhs rhs:exprT)
| expr_op1 (op:op1) (e:exprT)
| expr_value_of (l:lv)
| expr_addr_of (l:lv)
| expr_cast (t:ctype) (e:exprT)
| expr_size_of (e:exprT)

with lv :=
| lv_id (x:varT)
| lv_deref (e:exprT)
| lv_array (arr:exprT) (idx:exprT)
| lv_struct (strct:lv) (pos:idT).

Fixpoint exprT_eqb (lhs rhs : exprT) : bool :=
  match lhs, rhs with
    | expr_const ln, expr_const rn => eqb ln rn
    | expr_op2 lop ll lr, expr_op2 rop rl rr =>
      eqb lop rop && exprT_eqb ll rl && exprT_eqb lr rr
    | expr_op1 lop le, expr_op1 rop re =>
      eqb lop rop && exprT_eqb le re
    | expr_value_of ll, expr_value_of rl => lv_eqb ll rl
    | expr_addr_of ll, expr_addr_of rl => lv_eqb ll rl
    | expr_cast lct lex, expr_cast rct rex =>
      eqb lct rct && exprT_eqb lex rex
    | expr_size_of lex, expr_size_of rex => exprT_eqb lex rex
    | _, _ => false
  end

with lv_eqb (lhs rhs : lv) : bool :=
  match lhs, rhs with
    | lv_id lx, lv_id rx => eqb lx rx
    | lv_deref lex, lv_deref rex => exprT_eqb lex rex
    | lv_array larr lidx, lv_array rarr ridx =>
      exprT_eqb larr rarr && exprT_eqb lidx ridx
    | lv_struct lstrct lpos, lv_struct rstrct rpos =>
      lv_eqb lstrct rstrct && eqb lpos rpos
    | _, _ => false
  end.

Program Instance exprT_setoid : setoid exprT := {}.
Program Instance exprT_usual_setoid : @usual_setoid exprT _ := {}.
Program Instance lv_setoid : setoid lv := {}.
Program Instance lv_usual_setoid : @usual_setoid lv _ := {}.

Lemma exprT_reflect (x y: exprT) : reflect (x ≡ y) (exprT_eqb x y)
with lv_reflect (x y: lv) : reflect (x ≡ y) (lv_eqb x y).
Proof.
  move: x y.
  induction x; destruct y; try by constructor.
  { simpl. remember (eqb l l0) as b; symmetry in Heqb; destruct b.
    move/eqbP:Heqb => Heqb. constructor. rewrite Heqb. reflexivity.
    move/eqbP:Heqb => Heqb. constructor. intro. elim Heqb. inversion H.
    reflexivity. }
  { simpl. remember (eqb op op0) as b; symmetry in Heqb; destruct b.
    move/eqbP:Heqb => Heqb. rewrite Heqb.
    remember (IHx1 y1); clear Heqr.
    remember (exprT_eqb x1 y1) as b; symmetry in Heqb0; destruct b.
    remember (IHx2 y2); clear Heqr0.
    remember (exprT_eqb x2 y2) as b; symmetry in Heqb1; destruct b.
    constructor. inversion r. inversion r0. rewrite H. rewrite H0. reflexivity.
    constructor. intro. inversion r0. elim H0. inversion H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity.
    constructor. intro. move/eqbP:Heqb => Heqb. elim Heqb. inversion H.
    reflexivity. }
  { simpl. remember (eqb op op0) as b; symmetry in Heqb; destruct b.
    move/eqbP:Heqb => Heqb. rewrite Heqb.
    remember (IHx y); clear Heqr.
    remember (exprT_eqb x y) as b; symmetry in Heqb0; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity.
    constructor. intro. move/eqbP:Heqb => Heqb. elim Heqb. inversion H.
    reflexivity. }
  { simpl. remember (lv_reflect l l0); clear Heqr.
    remember (lv_eqb l l0) as b; symmetry in Heqb; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }
  { simpl. remember (lv_reflect l l0); clear Heqr.
    remember (lv_eqb l l0) as b; symmetry in Heqb; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }
  { simpl. remember (eqb t t0) as b; symmetry in Heqb; destruct b.
    move/eqbP:Heqb => Heqb. rewrite Heqb.
    remember (IHx y); clear Heqr.
    remember (exprT_eqb x y) as b; symmetry in Heqb0; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity.
    constructor. intro. move/eqbP:Heqb => Heqb. elim Heqb. inversion H.
    reflexivity. }
  { simpl. remember (IHx y); clear Heqr.
    remember (exprT_eqb x y) as b; symmetry in Heqb; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }

  induction x; destruct y; try by constructor.
  { simpl. remember (eqb x x0) as b; symmetry in Heqb; destruct b.
    move/eqbP:Heqb => Heqb. constructor. 
    unfold eq; simpl. unfold Equivalence.equiv. f_equal.
    rewrite Heqb. reflexivity. 
    move/eqbP:Heqb => Heqb. constructor. 
    intro. elim Heqb. inversion H. reflexivity. }
  { simpl. remember (exprT_reflect e e0); clear Heqr.
    remember (exprT_eqb e e0) as b; symmetry in Heqb; destruct b.
    constructor. inversion r. rewrite H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }
  { simpl. remember (exprT_reflect arr arr0); clear Heqr.
    remember (exprT_eqb arr arr0) as b; symmetry in Heqb; destruct b.
    remember (exprT_reflect idx idx0); clear Heqr0.
    remember (exprT_eqb idx idx0) as b; symmetry in Heqb0; destruct b.    
    constructor. inversion r. inversion r0. rewrite H. rewrite H0.
    reflexivity.
    constructor. intro. inversion r0. elim H0. inversion H. reflexivity.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }
  { simpl. remember (lv_reflect x y); clear Heqr.
    remember (lv_eqb x y) as b; symmetry in Heqb; destruct b.
    remember (eqb pos pos0) as b'; symmetry in Heqb'; destruct b'.
    move/eqbP in Heqb'.
    constructor. inversion r. rewrite H. rewrite Heqb'.
    reflexivity.
    constructor. intro. inversion H. subst. move/eqbP in Heqb'. contradict Heqb'. done.
    constructor. intro. inversion r. elim H0. inversion H. reflexivity. }
Qed.

Program Instance exprT_setoidDec : @setoidDec exprT _ := {
  eqb_ := exprT_eqb
}.
Next Obligation with eauto with typeclass_instances.
  apply exprT_reflect.
Qed.

Program Instance lv_setoidDec : @setoidDec lv _ := {
  eqb_ := lv_eqb
}.
Next Obligation with eauto with typeclass_instances.
  apply lv_reflect.
Qed.

Inductive alloc :=
| alloc_arr (e:exprT) (s:exprT). (* [e, stride] *)

Definition alloc_eqb (lhs rhs : alloc) : bool :=
  match lhs, rhs with
    | alloc_arr lex ls, alloc_arr rex rs => 
      eqb lex rex && eqb ls rs
  end.

Program Instance alloc_setoid : setoid alloc := {}.
Program Instance alloc_usual_setoid : @usual_setoid alloc _ := {}.
Program Instance alloc_setoidDec : @setoidDec alloc _ := {
  eqb_ := alloc_eqb
}.
Next Obligation.
  move: x y=> [xe xs] [ye ys];
    rewrite/eq/=/Equivalence.equiv/alloc_eqb;
    try by constructor.
  { apply/equivP; first by apply andP.
    constructor.
    { move=> [He Hs].
      move/eqbP in He; usual_tac He.
      by move/eqbP in Hs; usual_tac Hs.
    }
    { move=> Heq.
      inversion Heq; subst; clear Heq.
      by split; apply/eqbP.
    }
  }
Qed.

Inductive lcall :=
| airac_print 
| airac_dump  
| rand 
| strlen 
| strcpy
| scanf 
| errno_location
| builtin_alloca
| realloc
| builtin_choose_expr
| unknown.

Definition lcall_eqb (lhs rhs : lcall) : bool :=
  match lhs, rhs with
    | airac_print, airac_print => true
    | airac_dump, airac_dump => true
    | rand, rand => true
    | strlen, strlen => true
    | strcpy, strcpy => true
    | scanf, scanf => true
    | errno_location, errno_location => true
    | builtin_alloca, builtin_alloca => true
    | realloc, realloc => true
    | builtin_choose_expr, builtin_choose_expr => true
    | unknown, unknown => true
    | _, _ => false
  end.

Program Instance lcall_setoid : setoid lcall := {}.
Program Instance lcall_usual_setoid : @usual_setoid lcall _ := {}.
Program Instance lcall_setoidDec : @setoidDec lcall _ := {
  eqb_ := lcall_eqb
}.
Next Obligation.
  case x; case y; by constructor.
Qed.

Inductive instrT :=
| instr_assign (l:lv) (e:exprT)
| instr_alloc (l:lv) (a:alloc) (site:allocT)
| instr_assume (test:exprT)
| instr_libcall (ol:option lv) (lc:lcall) (le:list exprT) (site:allocT)
| instr_forget (l:DomAbsValue.locT)
| instr_skip.

Definition instrT_eqb (lhs rhs : instrT) : bool :=
  match lhs, rhs with
    | instr_assign lx lex, instr_assign rx rex => eqb lx rx && eqb lex rex
    | instr_alloc ll la ls, instr_alloc rl ra rs =>
      eqb ll rl && eqb la ra && eqb ls rs
    | instr_assume lt, instr_assume rt => eqb lt rt
    | instr_libcall lol llc lle ls, instr_libcall rol rlc rle rs =>
      eqb lol rol && eqb llc rlc && eqb lle rle && eqb ls rs
    | instr_forget ll, instr_forget rl =>
      eqb ll rl
    | instr_skip, instr_skip => true
    | _, _ => false
  end.

Program Instance instrT_setoid : setoid instrT := {}.
Program Instance instrT_usual_setoid : @usual_setoid instrT _ := {}.
Program Instance instrT_setoidDec : @setoidDec instrT _ := {
  eqb_ := instrT_eqb
}.
Next Obligation.
  move: x y => [lv le|ll la ls|le|lol llc lle ls|ll|] 
               [rv re|rl ra rs|re|rol rlc rle rs|rl|] //; try constructor=> //=.
  apply/equivP; first apply andP.
  split; move=> [Hv He].
  by congr instr_assign; apply/usual_eqP'/eqbP.
  by subst; split; apply/eqbP/eq_refl.

  apply/equivP. apply andP. split; move=> [Hv He].
  move/andP:Hv => [Hl Ha].
  by congr instr_alloc; apply/usual_eqP'/eqbP.
  intros; subst; split; last apply/eqbP/eq_refl.
  apply/andP; split; apply/eqbP/eq_refl.

  apply/equivP; first apply eqbP.
  rewrite usual_eqP...
  by split=> Heq; inversion Heq.

  apply/equivP. apply andP. split; move=> [Hv Hs].
  move/andP:Hv => [Hlce He].
  move/andP:Hlce => [Hol Hc].
  by congr instr_libcall; apply/usual_eqP'/eqbP.
  intros; subst; split; last apply/eqbP/eq_refl. 
  apply/andP; split; last by apply/eqbP/eq_refl.
  apply/andP; split; apply/eqbP/eq_refl.

  apply/equivP; first apply eqbP.
  rewrite usual_eqP...
  by split=> Heq; inversion Heq.
Qed.

Inductive tmnT :=
| tmn_jump: forall (nextbids:Set![bidT]), tmnT
| tmn_call: forall (nextbid:bidT) (nextfids:Set![fidT]) (args:list exprT) (retto:option varT), tmnT (* retto = return to *)
| tmn_exit: forall (ret:exprT), tmnT.

Definition tmnT_eqb (lhs rhs : tmnT) : bool :=
  match lhs, rhs with
    | tmn_jump lnbids, tmn_jump rnbids => eqb lnbids rnbids
    | tmn_call lbid lfid largs lretto, tmn_call rbid rfid rargs rretto =>
      eqb lbid rbid && eqb lfid rfid && eqb largs rargs && eqb lretto rretto
    | tmn_exit lret, tmn_exit rret => eqb lret rret
    | _, _ => false
  end.

Program Instance tmnT_setoid : setoid tmnT := {
  eq_ := tmnT_eqb
}.
Next Obligation.
  case=> [nbids|bid fid args retto|ret].
  exact/eqbP/eq_refl.
  apply/andP; split; last exact/eqbP/eq_refl.
  apply/andP; split; last exact/eqbP/eq_refl.
  apply/andP; split; last exact/eqbP/eq_refl.
  exact/eqbP/eq_refl.
  exact/eqbP/eq_refl.
Qed.
Next Obligation.
  move=> [lnbids|lbid lfid largs lretto|lret] [rnbids|rbid rfid rargs rretto|rret] // Hlr.
  exact/eqbP/eq_sym/eqbP.
  move: Hlr; do 3! case/andP; move=> Hbid Hfid Hargs Hretto.
  apply/andP; split; last exact/eqbP/eq_sym/eqbP.
  apply/andP; split; last exact/eqbP/eq_sym/eqbP.
  apply/andP; split; last exact/eqbP/eq_sym/eqbP.
  exact/eqbP/eq_sym/eqbP.
  exact/eqbP/eq_sym/eqbP.
Qed.
Next Obligation.
  move=> [lnbids|lbid lfid largs lretto|lret] [mnbids|mbid mfid margs mretto|mret] [rnbids|rbid rfid rargs rretto|rret] // Hlm Hmr.
  by apply/eqbP; apply eq_trans with mnbids; apply/eqbP.
  move: Hlm; do 3! case/andP; move=> Hlmb Hlmf Hlma Hlmr.
  move: Hmr; do 3! case/andP; move=> Hmrb Hmrf Hmra Hmrr.
  apply/andP; split; last by apply/eqbP; apply eq_trans with mretto; apply/eqbP.
  apply/andP; split; last by apply/eqbP; apply eq_trans with margs; apply/eqbP.
  apply/andP; split; last by apply/eqbP; apply eq_trans with mfid; apply/eqbP.
  by apply/eqbP; apply eq_trans with mbid; apply/eqbP.
  by apply/eqbP; apply eq_trans with mret; apply/eqbP.
Qed.

Program Instance tmnT_setoidDec : @setoidDec tmnT _ := {
  eqb_ := tmnT_eqb
}.
Next Obligation. exact/idP. Qed.

Definition cmdT := (list instrT * tmnT)%type.
Definition cinstrs := nosimpl (fun (cmd:cmdT) => cmd.1).
Definition ctmn := nosimpl (fun (cmd:cmdT) => cmd.2).

Definition funcT := (bidT * Map![bidT, cmdT] * list idT)%type.
Definition fentry  := nosimpl (fun (func:funcT) => func.1.1).
Definition fcmds := nosimpl (fun (func:funcT) => func.1.2).
Definition fparams := nosimpl (fun (func:funcT) => func.2).

Definition progT := (fidT * Map![fidT, funcT])%type.
Definition pentry := nosimpl (fun (prog:progT) => prog.1).
Definition pfuncs := nosimpl (fun (prog:progT) => prog.2).

Ltac dom_tac :=
  do ? [
    rewrite [eq (Elt _) (Elt _)]/eq/eq_/withtop_setoid/withtop_eq |
    rewrite [le (Elt _) (Elt _)]/le/le_/withtop_poset/withtop_le |
    rewrite [join _ bot] join_bot |
    rewrite [join bot _] join_commutative join_bot |
    rewrite [join (Elt _) (Elt _)]/join/join_/withtop_hasJoin/withtop_join |
    rewrite [join (Top _) _]/join/join_/withtop_hasJoin/withtop_join |
    rewrite [join _ (Top _)] join_commutative |
    rewrite [mrestrict _ (Top _)]/mrestrict/withtop_restricting_maplike |
    rewrite [mrestrict _ (Elt _)]/mrestrict/withtop_restricting_maplike |
    rewrite [mrestrict_join _ (Top _) _]/mrestrict_join |
    rewrite [mrestrict_join _ _ (Top _)]/mrestrict_join |
    rewrite [mrestrict_join _ (Elt _) (Elt _)]/mrestrict_join |
    rewrite powerset_smem_spec |
    rewrite morphism_mlookup_spec |
    rewrite morphism_madd_spec |
    rewrite morphism_mrestrict_spec |
    rewrite morphism_join_spec |
    rewrite set_smem_spec |
    rewrite map_mlookup_spec |
    rewrite map_madd_spec |
    rewrite map_join_spec |
    rewrite map_mrestrict_spec |
    rewrite map_mrestrict_spec' |
    rewrite powerset_neg_spec |
    rewrite set_to_powerset_spec |
  ].

Ltac eqb_tac :=
  do ? [
    rewrite [eqb (inl _) (inl _)]/eqb |
    rewrite [eqb_ (inl _) (inl _)]/= |
    rewrite [eqb (inr _) (inr _)]/eqb |
    rewrite [eqb_ (inr _) (inr _)]/= |
    rewrite [eqb (inl _) (inr _)]/eqb |
    rewrite [eqb_ (inl _) (inr _)]/= |
    rewrite [eqb (inr _) (inl _)]/eqb |
    rewrite [eqb_ (inr _) (inl _)]/= |
    rewrite [eqb (_, _) (_, _)]/eqb |
    rewrite [eqb_ (_, _) (_, _)]/= |
    rewrite [eqb (Some _) (Some _)]/eqb |
    rewrite [eqb_ (Some _) (Some _)]/= |
    rewrite [pair_eqb _ _]/pair_eqb |
    rewrite [option_eqb _ _]/option_eqb |
    rewrite eqb_refl |
    rewrite/andb/negb/fst/snd |
  ].
