exception TODO

exception PpError of string

module Log = struct
  type t = int * string

  let empty = (0, "")

  let rec get_spaces (n:int) : string =
    if n = 0 then ""
    else " " ^ (get_spaces (n - 1))

  let log: string -> t -> t =
    fun str log ->
      (fst log,
       snd log ^
         (get_spaces (fst log)) ^
         str ^ "\n")

  let increase (log:t) = (fst log + 2, snd log)
  let decrease (log:t) = (fst log - 2, snd log)
  let inner: string -> (t -> t) -> t -> t =
    fun title inner l ->
      let l = log ("[" ^ title ^ "]") l in
      let l = increase l in
      let l = inner l in
      let l = decrease l in
      l

  let to_string (log:t) = snd log

  let prerr_endline_log (f:t -> t) : unit = prerr_endline (to_string (f empty))
end

(*****************************)
(* pretty printer for syntax *)
(*****************************)

let pprint_idT 
    (pos2str: (int, string) PMap.t) 
    (x: DomBasic.idT) : string = 
  try 
    PMap.find x pos2str
  with Not_found -> raise (PpError ("pprint_idT: no matching string."))

let pprint_idT_native (x: DomBasic.idT) : string = string_of_int x

let pprint_var 
    (pos2str: (int, string) PMap.t)
    (x: DomBasic.varT) : string = 
  match x with
  | Datatypes.Coq_inl a -> ("gvar "^(pprint_idT pos2str a))
  | Datatypes.Coq_inr b -> ("lvar "^(pprint_idT pos2str b))

let pprint_op1 (o: DomBasic.op1) : string = 
  match o with
  | DomBasic.Coq_op_not -> "!"
  | DomBasic.Coq_op_bnot -> "~"

let pprint_op2 (o: DomBasic.op2) : string = 
  match o with
  | DomBasic.Coq_op_add    -> "+" 
  | DomBasic.Coq_op_sub    -> "-"
  | DomBasic.Coq_op_mul    -> "*"
  | DomBasic.Coq_op_div    -> "/"
  | DomBasic.Coq_op_mod    -> "%"
  | DomBasic.Coq_op_lt     -> "<"
  | DomBasic.Coq_op_le     -> "<="
  | DomBasic.Coq_op_gt     -> ">"
  | DomBasic.Coq_op_ge     -> ">="
  | DomBasic.Coq_op_eq     -> "=="
  | DomBasic.Coq_op_neq    -> "!="
  | DomBasic.Coq_op_and    -> "&&"
  | DomBasic.Coq_op_or     -> "||"
  | DomBasic.Coq_op_band   -> "&"
  | DomBasic.Coq_op_bor    -> "|"
  | DomBasic.Coq_op_bxor   -> "^" 
  | DomBasic.Coq_op_lshift -> "<<"
  | DomBasic.Coq_op_rshift -> ">>"
    
let pprint_literal (l: Syntax .literal) : string =
  match l with
  | Syntax.Coq_lit_int z -> (string_of_int z)
  | Syntax.Coq_lit_int_top -> "inttop"
  | Syntax.Coq_lit_top -> "top"
  | Syntax.Coq_lit_bot -> "bot"

let rec pprint_ctype 
    (pos2str: (int, string) PMap.t)
    (ct: Syntax.ctype) : string = 
  match ct with
  | Syntax.Coq_typ_uint          -> "uint"  
  | Syntax.Coq_typ_uchar         -> "uchar" 
  | Syntax.Coq_typ_ushort        -> "ushort" 
  | Syntax.Coq_typ_ulong         -> "ulong" 
  | Syntax.Coq_typ_ulonglong     -> "ulonglong" 
  | Syntax.Coq_typ_int           -> "int" 
  | Syntax.Coq_typ_char          -> "char" 
  | Syntax.Coq_typ_short         -> "short" 
  | Syntax.Coq_typ_long          -> "long" 
  | Syntax.Coq_typ_longlong      -> "longlong" 
  | Syntax.Coq_typ_struct x      -> ("struct "^(pprint_var pos2str x))
  | Syntax.Coq_typ_union  x      -> ("union "^(pprint_var pos2str x))
  | Syntax.Coq_typ_pointer ct1   -> ("*"^(pprint_ctype pos2str ct1))
  | Syntax.Coq_typ_array (ct1,i) -> 
    ((pprint_ctype pos2str ct1)^"["^(string_of_int i)^"]")
  | Syntax.Coq_typ_function ct1  -> ("func "^(pprint_ctype pos2str ct1))
  | Syntax.Coq_typ_void          -> "void" 
  | Syntax.Coq_typ_va_list       -> "va_list" 
  | Syntax.Coq_typ_unknown       -> "unknown"  

let rec pprint_exprT 
    (pos2str: (int, string) PMap.t)
    (e: Syntax.exprT) : string = 
  match e with
  | Syntax.Coq_expr_const l         -> pprint_literal l
  | Syntax.Coq_expr_op2 (o, e1, e2) -> 
    ("("^(pprint_exprT pos2str e1)^(pprint_op2 o)^
        (pprint_exprT pos2str e2)^")")
  | Syntax.Coq_expr_op1 (o, e1)     -> 
    ((pprint_op1 o)^"("^(pprint_exprT pos2str e1)^")")
  | Syntax.Coq_expr_value_of l      -> ("value_of ("^(pprint_lv pos2str l)^")")
  | Syntax.Coq_expr_addr_of l       -> ("& "^(pprint_lv pos2str l))
  | Syntax.Coq_expr_cast (ct, e1)   -> 
    ("("^(pprint_ctype pos2str ct)^")("^(pprint_exprT pos2str e1)^")")
  | Syntax.Coq_expr_size_of e1      -> 
    ("size_of ("^(pprint_exprT pos2str e1)^")")

and pprint_lv
    (pos2str: (int, string) PMap.t)
    (l: Syntax.lv) : string = 
  match l with
  | Syntax.Coq_lv_id x -> pprint_var pos2str x
  | Syntax.Coq_lv_deref e -> ("*("^(pprint_exprT pos2str e)^")")
  | Syntax.Coq_lv_array (e1,e2) -> 
    ("("^(pprint_exprT pos2str e1)^")["^(pprint_exprT pos2str e2)^"]")
  | Syntax.Coq_lv_struct (lv,x) ->
    ("("^(pprint_lv pos2str lv)^")->"^(pprint_idT pos2str x))

let pprint_alloc
    (pos2str: (int, string) PMap.t)
    (a: Syntax.alloc) : string = 
  match a with
  | Syntax.Coq_alloc_arr (e1,e2) ->
    ("array("^(pprint_exprT pos2str e1)^","^(pprint_exprT pos2str e2)^")")

let pprint_lcall
    (lc: Syntax.lcall) : string = 
  match lc with
  | Syntax.Coq_airac_print         -> "airac_print"
  | Syntax.Coq_airac_dump          -> "airac_dump"
  | Syntax.Coq_rand                -> "rand"
  | Syntax.Coq_strlen              -> "strlen"
  | Syntax.Coq_strcpy              -> "strcpy"
  | Syntax.Coq_scanf               -> "scanf"
  | Syntax.Coq_errno_location      -> "errno_location"
  | Syntax.Coq_builtin_alloca      -> "builtin_alloca"
  | Syntax.Coq_realloc             -> "realloc"
  | Syntax.Coq_builtin_choose_expr -> "builtin_choose_expr"
  | Syntax.Coq_unknown             -> "unknown"

let pprint_allocT
    (pos2str: (int, string) PMap.t)
    (a: DomBasic.allocT) : string = 
  pprint_idT pos2str a

let rec pprint_list 
    (pprint_elem: 'a -> string)
    (elem_list: 'a list) : string =
  match elem_list with
  | [] -> ""
  | [x] -> (pprint_elem x)
  | hd::tl -> ((pprint_elem hd)^","^(pprint_list pprint_elem tl))

let rec pprint_idT_list 
    (pos2str: (int, string) PMap.t)
    (x_list: DomBasic.idT list) : string = 
  pprint_list (pprint_idT pos2str) x_list

let rec pprint_exprT_list 
    (pos2str: (int, string) PMap.t)
    (e_list: Syntax.exprT list) : string = 
  pprint_list (pprint_exprT pos2str) e_list

let pprint_bidT (b: DomBasic.bidT) : string = 
  if b = 1
  then "ENTRY"
  else if b = 2
  then "EXIT"
  else ("NODE"^(string_of_int(b-3)))

let pprint_fidT
    (pos2str: (int, string) PMap.t)
    (f: DomBasic.fidT) : string = pprint_idT pos2str f

let pprint_fidT_native (f: DomBasic.fidT) : string = pprint_idT_native f

let pprint_idT_set 
    (pos2str: (int, string) PMap.t)
    (x_set: DomBasic.idT SetInterface.set) : string = 
  "("^
    (Orders.intSetFold
       (fun x s -> 
         if s = "" 
         then pprint_idT pos2str x
         else (s^" "^(pprint_idT pos2str x)))
       x_set "")^
    ")"

let pprint_bidT_set (b_set: DomBasic.bidT SetInterface.set) : string = 
  "("^
    (Orders.intSetFold
       (fun b s -> 
         if s = "" 
         then pprint_bidT b
         else (s^" "^(pprint_bidT b)))
       b_set "")^
    ")"

let pprint_fidT_set 
    (pos2str: (int, string) PMap.t)
    (f_set: DomBasic.fidT SetInterface.set) : string = 
  pprint_idT_set pos2str f_set

let pprint_tmnT
    (pos2str: (int, string) PMap.t)
    (t: Syntax.tmnT) : string = 
  match t with
  | Syntax.Coq_tmn_jump b_set -> ("jump ("^(pprint_bidT_set b_set)^")")
  | Syntax.Coq_tmn_call (b,f_set,e_list,retto_opt) ->
    (match retto_opt with
    | Some retto -> (pprint_var pos2str retto)^"="
    | None -> "")^
      "call("^(pprint_fidT_set pos2str f_set)^")("^
      (pprint_exprT_list pos2str e_list)^"),"^
      (pprint_bidT b)
  | Syntax.Coq_tmn_exit e -> ("exit ("^(pprint_exprT pos2str e)^")")

let pprint_start_bidT (b: DomBasic.bidT) (l: Log.t) : Log.t = 
  Log.log ("start block: "^(pprint_bidT b)) l

let pprint_params 
    (pos2str: (int, string) PMap.t)
    (ps: DomBasic.idT list) (l: Log.t) : Log.t = 
  Log.log ("params: "^(pprint_idT_list pos2str ps)) l

(****************************)
(* pretty printer for state *)
(****************************)


let pprint_bidBool (bb: DomBasic.bidT * bool) : string =
  match bb with
  | (bi,bo) -> "("^(pprint_bidT bi)^","^(string_of_bool bo)^")"

let pprint_heap 
    (pos2str: (int, string) PMap.t) 
    (h: DomBasic.allocT) : string = 
  "Dyn("^(pprint_idT pos2str h)^")"

let pprint_heap_native
    (h: DomBasic.allocT) : string = 
  "Dyn("^(pprint_idT_native h)^")"

let pprint_global
    (pos2str: (int, string) PMap.t) 
    (g: DomBasic.idT) : string = 
  "Gvar("^(pprint_idT pos2str g)^")"

let pprint_global_native
    (g: DomBasic.idT) : string = 
  "Gvar("^(pprint_idT_native g)^")"

let pprint_local
    (pos2str: (int, string) PMap.t) 
    (l: DomBasic.fidT * DomBasic.idT) : string = 
  match l with
  | (f,x) -> "Lvar("^(pprint_fidT pos2str f)^","^(pprint_idT pos2str x)^")"

let pprint_local_native
    (l: DomBasic.fidT * DomBasic.idT) : string = 
  match l with
  | (f,x) -> "Lvar("^(pprint_fidT_native f)^","^(pprint_idT_native x)^")"

let pprint_locT'
    (pos2str: (int, string) PMap.t)
    (lc: ((DomBasic.allocT, DomBasic.idT) Datatypes.sum, DomBasic.fidT*DomBasic.idT) Datatypes.sum) : string =
  match lc with
  | Datatypes.Coq_inl hg ->
    (match hg with
    | Datatypes.Coq_inl heap -> pprint_heap pos2str heap
    | Datatypes.Coq_inr global -> pprint_global pos2str global
    )
  | Datatypes.Coq_inr local -> pprint_local pos2str local

let pprint_locT'_native
    (lc: ((DomBasic.allocT, DomBasic.idT) Datatypes.sum, DomBasic.fidT*DomBasic.idT) Datatypes.sum) : string =
  match lc with
  | Datatypes.Coq_inl hg ->
    (match hg with
    | Datatypes.Coq_inl heap -> pprint_heap_native heap
    | Datatypes.Coq_inr global -> pprint_global_native global
    )
  | Datatypes.Coq_inr local -> pprint_local_native local

let rec pprint_fields 
    (pos2str: (int, string) PMap.t)
    (fs: DomBasic.idT list) : string =
  match fs with
  | [] -> ""
  | hd::tl -> "->"^(pprint_idT pos2str hd)^(pprint_fields pos2str tl)

let rec pprint_fields_native
    (fs: DomBasic.idT list) : string =
  match fs with
  | [] -> ""
  | hd::tl -> "->"^(pprint_idT_native hd)^(pprint_fields_native tl)

let pprint_locT 
    (pos2str: (int, string) PMap.t)
    (lc: DomAbsValue.locT) : string =
  match lc with
  | (l,fs) -> (pprint_locT' pos2str l)^(pprint_fields pos2str (List.rev fs))

let pprint_locT_native
    (lc: DomAbsValue.locT) : string =
  match lc with
  | (l,fs) -> (pprint_locT'_native l)^(pprint_fields_native (List.rev fs))

let pprint_itvT' (itv: DomItv.itvT') : string = 
  match itv with
  | DomItv.Coq_itv'_top -> "[-oo,+oo]"
  | DomItv.Coq_itv'_inf_z z -> 
    "[-oo,"^(string_of_int z)^"]"
  | DomItv.Coq_itv'_z_inf z -> 
    "["^(string_of_int z)^",+oo]"
  | DomItv.Coq_itv'_z_z (lz,uz) ->
    "["^(string_of_int lz)^","^
      (string_of_int uz)^"]"

let pprint_itvT (itv: DomItv.itvT) : string =
  match itv with
  | Some itv' -> pprint_itvT' itv'
  | None -> "bot_itv"

let pprint_locT_set 
    (pos2str: (int, string) PMap.t)
    (ls: DomAbsValue.locT SetInterface.set) : string =
  "("^
    (Orders.locTSetFold
       (fun l s -> 
         if s = "" 
         then pprint_locT pos2str l
         else (s^" "^(pprint_locT pos2str l)))
       ls "")^
    ")"

let pprint_locsT 
    (pos2str: (int, string) PMap.t)
    (ls: DomAbsValue.locsT) : string =
  match ls with
  | DomBasic.Elt ls' -> pprint_locT_set pos2str ls'  
  | DomBasic.Top -> "top_locs"

let pprint_oss (oss: DomAbsValue.ossT) : string =
  match oss with
  | ((o,sz),st) -> 
    "("^(pprint_itvT o)^","^(pprint_itvT sz)^","^(pprint_itvT st)^")"

let pprint_arrvT'
    (pos2str: (int, string) PMap.t)
    (a: DomAbsValue.arrvT') : string =
  "["^
    (Orders.intMapFold
       (fun a oss s ->
         if s = ""
         then ((pprint_heap pos2str a)^"->"^(pprint_oss oss))
         else (s^","^(pprint_heap pos2str a)^"->"^(pprint_oss oss))
       )
       a "")^
    "]"

let pprint_arrvT
    (pos2str: (int, string) PMap.t)
    (a: DomAbsValue.arrvT) : string =
  match a with
  | DomBasic.Elt a' -> pprint_arrvT' pos2str a'
  | DomBasic.Top -> "top_arrv"

let pprint_valueT
    (pos2str: (int, string) PMap.t)
    (v: DomAbsValue.valueT) : string =
  "("^(pprint_itvT (DomAbsValue.itv_of v))^","^
    (pprint_locsT pos2str (DomAbsValue.locs_of v))^","^
    (pprint_arrvT pos2str (DomAbsValue.arrv_of v))^")"

let pprint_memT'
    (pos2str: (int, string) PMap.t)
    (m: (DomAbsValue.locT,DomAbsValue.valueT) MapInterface.dict) 
    (l: Log.t) : Log.t = 
  Orders.locTMapFold
    (fun lc v l ->
      Log.log ((pprint_locT pos2str lc)^":"^(pprint_valueT pos2str v)) l
    )
    m l

let pprint_memT
    (pos2str: (int, string) PMap.t)
    (m: DomAbsMem.memT) (l: Log.t) : Log.t = 
  match m with
  | DomBasic.Elt m' -> pprint_memT' pos2str m' l
  | DomBasic.Top -> Log.log ("top_mem") l

let pprint_func_stateT
    (pos2str: (int, string) PMap.t)
    (fs: DomAbsState.func_stateT) (l: Log.t) : Log.t = 
  Orders.bidBoolMapFold
    (fun bb m l ->
      Log.inner
        ("pos: "^(pprint_bidBool bb))
        (pprint_memT pos2str m)
        l
    )
    fs l

let pprint_stateT'
    (pos2str: (int, string) PMap.t)
    (s: DomAbsState.stateT') (l: Log.t) : Log.t = 
  Orders.intMapFold
    (fun f fs l ->
      Log.inner
        ("function: "^(PMap.find f pos2str))
        (pprint_func_stateT pos2str fs)
        l
    )
    s l

let pprint_instrT
    (pos2str: (int, string) PMap.t)
    (i: Syntax.instrT) : string = 
  match i with
  | Syntax.Coq_instr_assign (l,e) -> 
    ((pprint_lv pos2str l)^"="^(pprint_exprT pos2str e))
  | Syntax.Coq_instr_alloc (l,a,at) ->
    ((pprint_lv pos2str l)^"=alloc("^(pprint_alloc pos2str a)^"),"^
        (pprint_allocT pos2str at))
  | Syntax.Coq_instr_assume e -> ("assume("^(pprint_exprT pos2str e)^")")
  | Syntax.Coq_instr_libcall (lv_opt,lc,e_list,at) ->
    ((match lv_opt with
      | Some l -> ((pprint_lv pos2str l)^"=")
      | None -> "")^
        (pprint_lcall lc)^
        "("^(pprint_exprT_list pos2str e_list)^"),"^
        (pprint_allocT pos2str at))
  | Syntax.Coq_instr_forget aloc ->
     "forget("^(pprint_locT pos2str aloc)^")"
  | Syntax.Coq_instr_skip -> "skip"

let rec pprint_instrT_list 
    (pos2str: (int, string) PMap.t)
    (i_list: Syntax.instrT list) (l: Log.t) : Log.t = 
  match i_list with
  | [] -> l
  | hd::tl -> 
    let l = Log.log (pprint_instrT pos2str hd) l in
    let l = pprint_instrT_list pos2str tl l in
    l

let pprint_cmdT
    (pos2str: (int, string) PMap.t)
    (c: Syntax.cmdT) (l: Log.t) : Log.t = 
  let l = pprint_instrT_list pos2str (Syntax.cinstrs c) l in
  let l = Log.log (pprint_tmnT pos2str (Syntax.ctmn c)) l in
  l

let pprint_bidT_cmdT_dict
    (pos2str: (int, string) PMap.t)
    (bc_map: (DomBasic.bidT,Syntax.cmdT) MapInterface.dict) (l: Log.t) : Log.t = 
  Orders.intMapFold
    (fun b c l -> 
      Log.inner 
        ("block: "^(pprint_bidT b))
        (pprint_cmdT pos2str c)
        l
    )
    bc_map l

let pprint_funcT
    (pos2str: (int, string) PMap.t)
    (f: Syntax.funcT) (l: Log.t) : Log.t = 
  let l = pprint_params pos2str (Syntax.fparams f) l in
  let l = pprint_start_bidT (Syntax.fentry f) l in
  let l = pprint_bidT_cmdT_dict pos2str (Syntax.fcmds f) l in
  l

let pprint_start_fidT
    (pos2str: (int, string) PMap.t)
    (f: DomBasic.fidT) (l: Log.t) : Log.t = 
  Log.log ("start function: "^(pprint_fidT pos2str f)) l

let pprint_fidT_funcT_dict
    (pos2str: (int, string) PMap.t)
    (ff_map: (DomBasic.fidT,Syntax.funcT) MapInterface.dict) (l: Log.t) : Log.t = 
  Orders.intMapFold
    (fun f func l -> 
      Log.inner 
        ("function: "^(pprint_fidT pos2str f))
        (pprint_funcT pos2str func)
        l
    )
    ff_map l

let pprint_progT
    (pos2str: (int, string) PMap.t)
    (p: Syntax.progT) (l: Log.t) : Log.t = 
  match p with
  | (f,funcs) ->
    let l = pprint_start_fidT pos2str f l in
    let l = pprint_fidT_funcT_dict pos2str funcs l in
    l

(************************************)
(* pretty printer for accessibility *)
(************************************)

let pprint_fidT_bidT 
    (pos2str: (int, string) PMap.t)
    (node: DomBasic.fidT*DomBasic.bidT) : string = 
  match node with
  | (f,b) -> "("^(pprint_fidT pos2str f)^","^(pprint_bidT b)^")"

let pprint_fidT_bidT_set
    (pos2str: (int, string) PMap.t)
    (froms: (DomBasic.fidT*DomBasic.bidT) SetInterface.set) : string = 
  "("^
    (Orders.intintSetFold
       (fun f s ->
         if s = ""
         then (pprint_fidT_bidT pos2str f)
         else (s^" "^(pprint_fidT_bidT pos2str f))
       )
       froms "")^
    ")"

let pprint_callfroms 
    (pos2str: (int, string) PMap.t)
    (cf: (DomBasic.fidT, (DomBasic.fidT*DomBasic.bidT) SetInterface.set) MapInterface.dict)
    (l: Log.t) : Log.t =
  Orders.intMapFold
    (fun f froms l ->
      Log.log
        ((pprint_fidT pos2str f)^"<-"^(pprint_fidT_bidT_set pos2str froms))
        l
    )
    cf l

let pprint_callfroms_star
    (pos2str: (int, string) PMap.t)
    (cfs: (DomBasic.fidT, DomBasic.fidT SetInterface.set) MapInterface.dict)
    (l: Log.t) : Log.t =
  Orders.intMapFold
    (fun f froms l ->
      Log.log
        ((pprint_fidT pos2str f)^"<-*"^(pprint_fidT_set pos2str froms))
        l
    )
    cfs l

let pprint_acclocs
    (pos2str: (int, string) PMap.t)
    (acc_locs: (DomBasic.fidT, DomAbsValue.locT SetInterface.set) MapInterface.dict)
    (l: Log.t) : Log.t =
  Orders.intMapFold
    (fun f froms l ->
      Log.log
        ((pprint_fidT pos2str f)^": "^(pprint_locT_set pos2str froms))
        l
    )
    acc_locs l

let pprint_access
    (pos2str: (int, string) PMap.t)
    (acc: DomAcc.accT) (l: Log.t) : Log.t =
  let l = Log.inner "callfroms" 
    (pprint_callfroms pos2str (DomAcc.callfroms acc)) l in
  let l = Log.inner "callfroms_star" 
    (pprint_callfroms_star pos2str (DomAcc.callfroms_star acc)) l in
  let l = Log.inner "acclocs" 
    (pprint_acclocs pos2str (DomAcc.acclocs acc)) l in
  l

(***************************)
(* pretty printer to prerr *)
(***************************)

let prerr_progT
    (pos2str: (int, string) PMap.t)
    (p: Syntax.progT) : unit = 
  Log.prerr_endline_log 
    (Log.inner "translated prog" (pprint_progT pos2str p))

let prerr_cfg
    (pos2str: (int, string) PMap.t)
    (f: Syntax.funcT) : unit = 
  Log.prerr_endline_log 
    (Log.inner "translated CFG" (pprint_funcT pos2str f))

let prerr_instrT_list
    (pos2str: (int, string) PMap.t)
    (i_list: Syntax.instrT list) : unit = 
  Log.prerr_endline_log 
    (Log.inner "translated statements" (pprint_instrT_list pos2str i_list))

let prerr_access
    (pos2str: (int, string) PMap.t)
    (acc: DomAcc.accT) : unit = 
  Log.prerr_endline_log
    (Log.inner "translated ACCESS info" (pprint_access pos2str acc))

let prerr_memT
    (pos2str: (int, string) PMap.t)
    (m: DomAbsMem.memT) : unit = 
  Log.prerr_endline_log
    (Log.inner "translated MEM" (pprint_memT pos2str m))

let prerr_func_stateT
    (pos2str: (int, string) PMap.t)
    (fs: DomAbsState.func_stateT) : unit = 
  Log.prerr_endline_log
    (Log.inner "translated FUNC_STATET" (pprint_func_stateT pos2str fs))

let prerr_stateT'
    (pos2str: (int, string) PMap.t)
    (s: DomAbsState.stateT') : unit = 
  Log.prerr_endline_log 
    (Log.inner "translated STATET'" (pprint_stateT' pos2str s))
