(***********************)
(* program translation *)
(***********************)

exception TrError of string

let tr_literal (lit:Syn.literal) : Syntax.literal =
  match lit with
  | Syn.Cint i   -> Syntax.Coq_lit_int i
  | Syn.Creal _  -> raise (TrError "tr_literal: Creal case")
  | Syn.Cinttop     -> Syntax.Coq_lit_int_top
  | Syn.Ctop     -> Syntax.Coq_lit_top
  | Syn.Cbot     -> Syntax.Coq_lit_bot
  | Syn.Crange _ -> raise (TrError "tr_literal: Crange case")

let tr_bop (b:Syn.bop) : DomBasic.op2 =
  match b with 
  | Syn.Plus   -> DomBasic.Coq_op_add
  | Syn.Minus  -> DomBasic.Coq_op_sub
  | Syn.Mult   -> DomBasic.Coq_op_mul
  | Syn.Div    -> DomBasic.Coq_op_div
  | Syn.Mod    -> DomBasic.Coq_op_mod
  | Syn.Lt     -> DomBasic.Coq_op_lt
  | Syn.Gt     -> DomBasic.Coq_op_gt
  | Syn.Eq     -> DomBasic.Coq_op_eq
  | Syn.Neq    -> DomBasic.Coq_op_neq
  | Syn.Ge     -> DomBasic.Coq_op_ge
  | Syn.Le     -> DomBasic.Coq_op_le
  | Syn.And    -> DomBasic.Coq_op_and
  | Syn.Or     -> DomBasic.Coq_op_or
  | Syn.Band   -> DomBasic.Coq_op_band
  | Syn.Bor    -> DomBasic.Coq_op_bor
  | Syn.Bxor   -> DomBasic.Coq_op_bxor
  | Syn.Lshift -> DomBasic.Coq_op_lshift
  | Syn.Rshift -> DomBasic.Coq_op_rshift

let tr_uop (u:Syn.uop) : DomBasic.op1 =
  match u with
  | Syn.Not  -> DomBasic.Coq_op_not
  | Syn.Bnot -> DomBasic.Coq_op_bnot

(* We assume that all local variables has a different names.  The
   "locals" argument shows the mapping between local variables and
   functions.
*)
let rec tr_var
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (x:Syn.vid) : Syntax.lv =
  try 
    (match Str0.splits x "->" with 
    | a::b ->
      let lv = 
        if PMap.mem a locals
        then Syntax.Coq_lv_id (Datatypes.Coq_inr (PMap.find a str2pos))
        else Syntax.Coq_lv_id (Datatypes.Coq_inl (PMap.find a str2pos))
      in
      let lv_struct = 
        List.fold_left
          (fun acc field -> 
            Syntax.Coq_lv_struct (acc, PMap.find field str2pos)
          )
          lv b
      in
      lv_struct
    | _ -> raise (TrError "tr_addr1: empty string of location:var"))
  with Not_found -> raise (TrError ("tr_var: "^x^" is not found in str2pos"))

and tr_exp
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (e:Syn.exp) : Syntax.exprT =
  match e with
  | Syn.Econst l -> Syntax.Coq_expr_const (tr_literal l)
  | Syn.Ebop (b,e1,e2) -> 
    Syntax.Coq_expr_op2 (tr_bop b, 
                         tr_exp locals str2pos e1, 
                         tr_exp locals str2pos e2)
  | Syn.Euop (u,e1) -> Syntax.Coq_expr_op1 (tr_uop u, tr_exp locals str2pos e1)
  | Syn.EaddrOf lv -> Syntax.Coq_expr_addr_of (tr_lv locals str2pos lv)
  | Syn.EvalueOf lv -> Syntax.Coq_expr_value_of (tr_lv locals str2pos lv)
  | Syn.Ecast (ct,e1,e2) -> 
    Syntax.Coq_expr_cast (tr_ctype locals str2pos ct, tr_exp locals str2pos e2)
  | Syn.EsizeOf e1 -> Syntax.Coq_expr_size_of (tr_exp locals str2pos e1)

and tr_lv
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (lv:Syn.lv) : Syntax.lv =
  match lv with
  | Syn.Lid x -> tr_var locals str2pos x
  | Syn.Lderef e -> Syntax.Coq_lv_deref (tr_exp locals str2pos e)
  | Syn.Larray (e1,e2) ->
    Syntax.Coq_lv_array (tr_exp locals str2pos e1,tr_exp locals str2pos e2)
  | Syn.Lfield (Syn.EvalueOf lv,s) -> 
    Syntax.Coq_lv_struct (tr_lv locals str2pos lv, PMap.find s str2pos)
  | Syn.Lfield (e1,s) -> 
    Syntax.Coq_lv_deref (Syntax.Coq_expr_const Syntax.Coq_lit_bot)

and tr_ctype
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (ct:Syn.ctypeinfo) : Syntax.ctype = 
  match ct with
  | (_,Syn.UINT)      -> Syntax.Coq_typ_uint
  | (_,Syn.UCHAR)     -> Syntax.Coq_typ_uchar
  | (_,Syn.USHORT)    -> Syntax.Coq_typ_ushort
  | (_,Syn.ULONG)     -> Syntax.Coq_typ_ulong
  | (_,Syn.ULONGLONG) -> Syntax.Coq_typ_ulonglong
  | (_,Syn.INT)       -> Syntax.Coq_typ_int
  | (_,Syn.CHAR)      -> Syntax.Coq_typ_char
  | (_,Syn.SHORT)     -> Syntax.Coq_typ_short
  | (_,Syn.LONG)      -> Syntax.Coq_typ_long
  | (_,Syn.LONGLONG)  -> Syntax.Coq_typ_long
  | (_,Syn.FLOAT) | (_,Syn.DOUBLE) | (_,Syn.LONGDOUBLE) ->
    raise (TrError "tr_ctype: type casting to float")
  | (_,Syn.STRUCT s)  -> 
    (match tr_var locals str2pos s with
    | Syntax.Coq_lv_id sid -> Syntax.Coq_typ_struct sid
    | _ -> raise (TrError "tr_ctype: struct name of (a->b)")
    )
  | (_,Syn.ASTRUCT _) -> Syntax.Coq_typ_unknown
  | (_,Syn.UNION s)   -> 
    (match tr_var locals str2pos s with
    | Syntax.Coq_lv_id sid -> Syntax.Coq_typ_union sid
    | _ -> raise (TrError "tr_ctype: union name of (a->b)")
    )
  | (_,Syn.AUNION _)  -> Syntax.Coq_typ_unknown
  | (_,Syn.POINTER ct1)   -> 
    Syntax.Coq_typ_pointer (tr_ctype locals str2pos ct1)
  | (_,Syn.ARRAY (ct1,e)) -> 
    Syntax.Coq_typ_array (tr_ctype locals str2pos ct1, 0)
  (* The syntax of array type is a little bit different.  However it
     does not matter to validate the analysis since the analysis does
     not follow semantics of array type casting. *)
  | (_,Syn.FUNCTION ct1)  -> 
    Syntax.Coq_typ_function (tr_ctype locals str2pos ct1)
  | (_,Syn.VOID)      -> Syntax.Coq_typ_void
  | (_,Syn.VA_LIST)   -> Syntax.Coq_typ_va_list
  | (_,Syn.UNKNOWN)   -> Syntax.Coq_typ_unknown

let tr_lv_opt 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (lv_opt:Syn.lv option) : Syntax.lv option =
  match lv_opt with
  | Some l -> Some (tr_lv locals str2pos l)
  | None -> None

let tr_lv_opt_for_call
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (lv_opt:Syn.lv option) : DomBasic.varT option =
  match lv_opt with
  | Some (Syn.Lid f) -> 
    (match tr_var locals str2pos f with
    | Syntax.Coq_lv_id ret -> Some ret
    | _ -> raise (TrError "tr_lv_opt_for_call: return address is complicated1")
    )
  | None -> None
  | Some l -> 
    ( prerr_endline (Syn.slv l "");
      raise (TrError "tr_lv_opt_for_call: return address is complicated2")
    )

let tr_exp_list 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (e_list:Syn.exp list) : Syntax.exprT list =
  List.map (fun x -> tr_exp locals str2pos x) e_list

let tr_rel (r:Syn.rel) : DomBasic.op2 =
  match r with 
  | Syn.Peq  -> DomBasic.Coq_op_eq
  | Syn.Pneq -> DomBasic.Coq_op_neq
  | Syn.Plt  -> DomBasic.Coq_op_lt
  | Syn.Ple  -> DomBasic.Coq_op_le
  | Syn.Pgt  -> DomBasic.Coq_op_gt
  | Syn.Pge  -> DomBasic.Coq_op_ge

let rec tr_fact
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (f:Syn.fact) : Syntax.exprT = 
  match f with
  | Syn.Pbin (r,e1,e2) ->
    Syntax.Coq_expr_op2 (tr_rel r, 
                         tr_exp locals str2pos e1,
                         tr_exp locals str2pos e2)
  | Syn.Pnot f1 ->
    Syntax.Coq_expr_op1 (DomBasic.Coq_op_not,
                         tr_fact locals str2pos f1)

(* 
   node number translator
   ENTRY  -> 1
   EXIT   -> 2
   NODE n -> n+3
*)
let tr_cfg_node (n:Cfg.Node.t) : int = 
  match n with
  | Cfg.Node.ENTRY  -> 1
  | Cfg.Node.EXIT   -> 2
  | Cfg.Node.NODE k -> (k+3)

let tr_cfg_node_rev (p:int) : Cfg.Node.t =
  if p = 1
  then Cfg.Node.ENTRY
  else if p = 2
  then Cfg.Node.EXIT
  else Cfg.Node.NODE (p-3)

let tr_cfg_params 
    (str2pos: (string, int) PMap.t)
    (params:string list) : int list =
  List.map (fun x -> PMap.find x str2pos) params

let tr_alloc 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (a:Syn.alloc) : Syntax.alloc = 
  match a with
  | Syn.Aarr (e1,e2) -> Syntax.Coq_alloc_arr (tr_exp locals str2pos e1,
                                              tr_exp locals str2pos e2)
  | Syn.Arec f_list -> raise (TrError ("tr_alloc: structure case"))

let tr_lcall (e:Syn.exp) : Syntax.lcall =
  match e with
  | Syn.EvalueOf (Syn.Lid "airac_print")      -> Syntax.Coq_airac_print
  | Syn.EvalueOf (Syn.Lid "airac_dump")       -> Syntax.Coq_airac_dump
  | Syn.EvalueOf (Syn.Lid "rand")             -> Syntax.Coq_rand
  | Syn.EvalueOf (Syn.Lid "strlen")           -> Syntax.Coq_strlen
  | Syn.EvalueOf (Syn.Lid "strcpy")           -> Syntax.Coq_strcpy
  | Syn.EvalueOf (Syn.Lid "scanf")            -> Syntax.Coq_scanf
  | Syn.EvalueOf (Syn.Lid "__errno_location") -> Syntax.Coq_errno_location
  | Syn.EvalueOf (Syn.Lid "__builtin_alloca") -> Syntax.Coq_builtin_alloca
  | Syn.EvalueOf (Syn.Lid "__builtin_choose_expr") -> 
    Syntax.Coq_builtin_choose_expr
  | _                                         -> Syntax.Coq_unknown

let tr_cmd
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (cmd: Cfg.bcmd) (tag:Syn.tag) : Syntax.instrT =
  match cmd with
  | Cfg.SET (lv,e) -> 
    Syntax.Coq_instr_assign (tr_lv locals str2pos lv,
                             tr_exp locals str2pos e)
  | Cfg.WEAKSET (lv,e) ->
    Syntax.Coq_instr_assign (tr_lv locals str2pos lv,
                             tr_exp locals str2pos e)
  | Cfg.ESCAPE e -> Syntax.Coq_instr_skip
  | Cfg.ASSUME p -> Syntax.Coq_instr_assume (tr_fact locals str2pos p)
  | Cfg.ASSERT p -> Syntax.Coq_instr_assume (tr_fact locals str2pos p)
  | Cfg.ALLOC (lv, a, sp) ->
    Syntax.Coq_instr_alloc (tr_lv locals str2pos lv, 
                            tr_alloc locals str2pos a,
                            PMap.find (Syn.labeloftag tag) str2pos)
  | Cfg.CHECK _ -> Syntax.Coq_instr_skip
  | Cfg.LIBCALL (lv_opt, lc, e_list) ->
    Syntax.Coq_instr_libcall (tr_lv_opt locals str2pos lv_opt,
                              tr_lcall lc,
                              tr_exp_list locals str2pos e_list,
                              PMap.find (Syn.labeloftag tag) str2pos)
  | Cfg.FREE _ -> Syntax.Coq_instr_skip
  | Cfg.REFINE _ -> Syntax.Coq_instr_skip
  | Cfg.REFINEO _ -> Syntax.Coq_instr_skip
  | Cfg.NOP -> Syntax.Coq_instr_skip

let tr_cmds
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (cmds : Cfg.bcmds) : Syntax.instrT list = 
  List.map (fun (cmd,tag) -> tr_cmd locals str2pos cmd tag) cmds

let tr_closure 
    (str2pos: (string, int) PMap.t)
    (cls:Dom.Closure.t) : DomBasic.fidT SetInterface.set =
  match cls with
  | Dom.Closure.Top -> Orders.intSetEmpty
  | Dom.Closure.V proc_pset ->
    let pids = PSet.map Dom.Proc.pidof proc_pset in
    let fidTs = PSet.map (fun x -> PMap.find x str2pos) pids in
    PSet.fold (fun n acc -> Orders.intSetAdd n acc) fidTs Orders.intSetEmpty

let tr_call 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (mem: Dom.Mem.t) (g: Cfg.t) (nd: Cfg.Node.t)
    ((lv_opt:Syn.lvalue option), (e:Syn.exp), (e_list:Syn.exp list)) 
    : Syntax.tmnT = 
  let ret_blk = 
    try 
      tr_cfg_node (PSet.choose (Cfg.succ g nd)) 
    with Not_found -> raise (TrError "tr_call: no successor of call node")
  in
  let funcs   = tr_closure str2pos (Dom.Value.closure_of (Eval.eval e mem)) in
  let args    = tr_exp_list locals str2pos e_list in
  let ret_to  = tr_lv_opt_for_call locals str2pos lv_opt in
  Syntax.Coq_tmn_call (ret_blk, funcs, args, ret_to)

let rec get_escape_bcmds (cmds: (Cfg.bcmd * Syn.tag) list) : Syn.exp option =
  match cmds with
  | (Cfg.ESCAPE e, _)::_ -> Some e
  | _::tl            -> get_escape_bcmds tl
  | []               -> None

let get_escape (blk: Cfg.block) : Syn.exp option =
  match blk with
  | Cfg.CALL (_, _, cmds) -> get_escape_bcmds cmds
  | Cfg.BLOCK cmds        -> get_escape_bcmds cmds

(* In the translated program, return value is bottom when there is no
   escape statement in the program.
*)
let tr_exit
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (blk: Cfg.block) : Syntax.tmnT =
  let escape_exp = get_escape blk in
  match escape_exp with
  | Some e -> 
    Syntax.Coq_tmn_exit (tr_exp locals str2pos e)
  | None -> 
    Syntax.Coq_tmn_exit (Syntax.Coq_expr_const Syntax.Coq_lit_bot)

let tr_jump (g: Cfg.t) (nd: Cfg.Node.t) : Syntax.tmnT =
  let succ = Cfg.succ g nd in
  let succ_coq = 
    PSet.fold 
      (fun n acc -> Orders.intSetAdd n acc)
      (PSet.map tr_cfg_node succ)
      Orders.intSetEmpty 
  in
  Syntax.Coq_tmn_jump succ_coq

(* is_exit finds a note right before Cfg.Node.EXIT.  In many case,
   those nodes include ESCAPE statement. *)
let is_exit (g: Cfg.t) (nd:Cfg.Node.t) : bool =
  let succ = Cfg.succ g nd in
  PSet.exists Cfg.is_exit succ

let tr_cfg_block
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (mem: Dom.Mem.t) (g: Cfg.t) (nd: Cfg.Node.t) (blk: Cfg.block) 
    : Syntax.cmdT = 
  match blk with
  | Cfg.CALL ((lv_opt,e,e_list),t,cmds) -> 
    (* cmds in CALL block are not a normal statement as in Cfg.BLOCK
       node.  So, we translate the cmds to an empty statement. *)
    let instrs = [] in
    let tmn    = tr_call locals str2pos mem g nd (lv_opt,e,e_list) in
    (instrs, tmn)
  | Cfg.BLOCK cmds ->
    let instrs = tr_cmds locals str2pos cmds in
    let tmn    = 
      if is_exit g nd
      then tr_exit locals str2pos blk
      else tr_jump g nd in
    (instrs, tmn)

let tr_cfg_blocks 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (fname: Syn.pid) (inputof: Dom.Table.t)
    (g: Cfg.t) (blks: (Cfg.Node.t, Cfg.block) PMap.t) :
    (int, Syntax.cmdT) MapInterface.dict =
  PMap.foldi 
    (fun bid blk acc -> 
      if Cfg.is_exit bid
      then acc
      else 
        let bid_coq = tr_cfg_node bid in
        let input_mem = Dom.Table.find inputof (Dom.Work.make0 (fname,bid)) in
        let blk_coq = tr_cfg_block locals str2pos input_mem g bid blk in
	Orders.intMapAdd bid_coq blk_coq acc
    )
    blks 
    Orders.intMapEmpty

let tr_cfg 
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (fname: Syn.pid) (inputof: Dom.Table.t)
    (c:Cfg.t) : Syntax.funcT = 
  let initBidT = tr_cfg_node Cfg.Node.ENTRY in
  let entry_node = Cfg.BLOCK [] in
  let c        = Cfg.addnode Cfg.Node.ENTRY entry_node c in
  let blocks   = tr_cfg_blocks locals str2pos fname inputof
    c c.Cfg.blocks in
  let params   = tr_cfg_params str2pos c.Cfg.params in
  ((initBidT, blocks), params)

let tr_global
    ((locals: (Syn.vid,Syn.fid) PMap.t), (str2pos: (string, int) PMap.t),
     (inputof: Dom.Table.t),
     (g:Dom.global)) : Syntax.progT = 
  let init_fid = PMap.find "_G_" str2pos in
  let prog = 
    PMap.foldi
      (fun p cfg acc ->
        let fid_coq = PMap.find p str2pos in
        let cfg_coq = tr_cfg locals str2pos p inputof cfg in
        Orders.intMapAdd fid_coq cfg_coq acc
      )
      g.Dom.graph.M5graph.cfgs
      Orders.intMapEmpty
  in
  (init_fid, prog)

(******************************)
(* abstract value translation *)
(******************************)

let tr_addr1
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (a:Dom.Addr1.t) : DomAbsValue.locT = 
  match a with
  | Dom.Addr1.Variable x -> 
    (match Str0.splits x "->" with 
    | a::b ->
      if PMap.mem a locals
      then DomAbsValue.llocal 
        (PMap.find (PMap.find a locals) str2pos) 
        (PMap.find a str2pos)
        (List.map (fun x -> PMap.find x str2pos) (List.rev b))
      else DomAbsValue.lglobal 
        (PMap.find a str2pos) 
        (List.map (fun x -> PMap.find x str2pos) (List.rev b))
    | _ -> raise (TrError "tr_addr1: empty string of location:var"))
  | Dom.Addr1.Dynamic x -> 
    (match Str0.splits x "->" with 
    | a::b ->
      DomAbsValue.lheap 
        (PMap.find a str2pos)
        (List.map (fun x -> PMap.find x str2pos) (List.rev b))
    | _ -> raise (TrError "tr_addr1: empty string of location:dyn"))
  | Dom.Addr1.Null -> raise (TrError "tr_addr1: null case")

let tr_address
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (a:Dom.Address.t) : DomAbsValue.locsT = 
  match a with
  | Dom.Address.Top -> DomBasic.Top
  | Dom.Address.V a_set -> DomBasic.Elt 
    (PSet.fold
       (fun a acc -> Orders.locTSetAdd (tr_addr1 locals str2pos a) acc) 
       a_set 
       Orders.locTSetEmpty
    )

let tr_inter (pi:Inter.interval) : (DomItv.itvT) =
  match pi with
  | Inter.Interval (lb,ub) ->
    (match lb,ub with
    | Inter.MinusInf, Inter.PlusInf -> Some DomItv.Coq_itv'_top
    | Inter.MinusInf, Inter.Int uz  -> 
      Some (DomItv.Coq_itv'_inf_z uz)
    | Inter.Int lz,   Inter.PlusInf -> Some 
      (DomItv.Coq_itv'_z_inf lz)
    | Inter.Int lz,   Inter.Int uz  -> 
      Some (DomItv.Coq_itv'_z_z (lz, uz))
    | _, _ -> None)
  | Inter.IBot -> None

let tr_oss (oss:Dom.ArrayBlock.blk) : DomAbsValue.ossT =
  ((tr_inter oss.Dom.ArrayBlock.offset,
    tr_inter oss.Dom.ArrayBlock.size),
   tr_inter oss.Dom.ArrayBlock.stride)

let tr_arrayblock'
    (str2pos: (string, int) PMap.t)
    (ab:(Syn.id,Dom.ArrayBlock.blk) PMap.t) : DomAbsValue.arrvT' =
  PMap.foldi
    (fun a oss acc ->
       Orders.intMapAdd (PMap.find a str2pos) (tr_oss oss) acc)
    ab 
    Orders.intMapEmpty
 
let tr_arrayblock
    (str2pos: (string, int) PMap.t)
    (ab:Dom.ArrayBlock.t) : DomAbsValue.arrvT =
  match ab with 
  | Dom.ArrayBlock.Top -> DomBasic.Top
  | Dom.ArrayBlock.V a_map -> DomBasic.Elt (tr_arrayblock' str2pos a_map)

let tr_value
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (v: Dom.Value.t) : DomAbsValue.valueT =
  match v with
  | (itv, _, addrs, arrblk, _) ->
    let (itv,addrs) = 
      match addrs with
      | Dom.Address.Top -> (itv,addrs)
      | Dom.Address.V a_set -> 
        if (PSet.mem Dom.Addr1.Null a_set)
        then (Inter.join itv (Inter.Interval (Inter.Int 0,Inter.Int 0)),
	      Dom.Address.V (PSet.remove Dom.Addr1.Null a_set))
        else (itv,addrs)
    in
      ((tr_inter itv, 
	tr_address locals str2pos addrs), 
       tr_arrayblock str2pos arrblk)
	
let tr_mem
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (v: Dom.Mem.t) : DomAbsMem.memT =
  PMap.foldi 
    (fun k v acc -> 
      if k = Dom.Addr1.Null 
      then acc
      else 
	match acc with
	  | DomBasic.Elt m ->
            DomBasic.Elt (
	      Orders.locTMapAdd 
		(tr_addr1 locals str2pos k)
		(tr_value locals str2pos v)
		m
            ) 
	  | DomBasic.Top -> DomBasic.Top
    ) 
    v
    (DomBasic.Elt Orders.locTMapEmpty)

let tr_m5graph_node_simpl
    (str2pos: (string, int) PMap.t)    
    (n:M5graph.node): DomBasic.fidT*DomBasic.bidT =
  match n with
  | (f,n) -> (PMap.find f str2pos, tr_cfg_node n)

let tr_m5graph_node_bidBool
    (str2pos: (string, int) PMap.t)    
    (is_output: bool) (n:M5graph.node): DomBasic.bidT * bool = 
  match n with
  | (_,cfg_n) -> (tr_cfg_node cfg_n, is_output)

let tr_state
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (s: Dom.State.t) : DomAbsMem.memT = 
  tr_mem locals str2pos (Dom.State.join_all_mems s)

let tr_table_fun
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (f: Syn.fid) (nodes: Cfg.node list) (is_output: bool)
    (tbl: Dom.Table.t) : DomAbsState.func_stateT =
  List.fold_left
    (fun acc n ->
      let work = Dom.Work.make0 (M5graph.make_node f n) in
	  let s = Dom.Table.find tbl work in
	  Orders.bidBoolMapAdd
	    (tr_m5graph_node_bidBool str2pos is_output (f, n))
	    (tr_mem locals str2pos s)
	    acc
    )
    Orders.bidBoolMapEmpty
    nodes

let tr_table_fun_w_init
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (f: Syn.fid) (nodes: Cfg.node list) (is_output: bool)
    (tbl: Dom.Table.t) (init:DomAbsState.func_stateT) : DomAbsState.func_stateT =
  List.fold_left
    (fun acc n ->
      let work = Dom.Work.make0 (M5graph.make_node f n) in
	  let s = Dom.Table.find tbl work in
	  Orders.bidBoolMapAdd
	    (tr_m5graph_node_bidBool str2pos is_output (f, n))
	    (tr_mem locals str2pos s)
	    acc
    )
    init
    nodes

let tr_table
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (is_output: bool)
    (tbl: Dom.Table.t) : DomAbsState.stateT' =
  PMap.foldi
    (fun n s acc ->
      let f = M5graph.ctxof n in
      let fid_coq = PMap.find f str2pos in
      let bb_coq = tr_m5graph_node_bidBool str2pos is_output n in
      let mem_coq = tr_state locals str2pos s in
      let func_state = 
	match Orders.intMapFind fid_coq acc with
	  | Some fs -> fs
	  | None -> Orders.bidBoolMapEmpty
      in
      let func_state = Orders.bidBoolMapAdd bb_coq mem_coq func_state in
      let acc = Orders.intMapAdd fid_coq func_state acc in
	acc
    )
    tbl.Dom.Table.table
    Orders.intMapEmpty

let tr_table_w_init
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (is_output: bool)
    (tbl: Dom.Table.t) (init:DomAbsState.stateT') : DomAbsState.stateT' =
  PMap.foldi
    (fun n s acc ->
      let f = M5graph.ctxof n in
      let fid_coq = PMap.find f str2pos in
      let bb_coq = tr_m5graph_node_bidBool str2pos is_output n in
      let mem_coq = tr_state locals str2pos s in
      let func_state = 
	match Orders.intMapFind fid_coq acc with
	  | Some fs -> fs
	  | None -> Orders.bidBoolMapEmpty
      in
      let func_state = Orders.bidBoolMapAdd bb_coq mem_coq func_state in
      let acc = Orders.intMapAdd fid_coq func_state acc in
	acc
    )
    tbl.Dom.Table.table
    init

(*****************************)
(* accessibility translation *)
(*****************************)

let tr_callfroms 
    (str2pos: (string, int) PMap.t)
    (g:Dom.global) (pre: Pre.preinfo) 
    : (DomBasic.fidT, (DomBasic.fidT*DomBasic.bidT) SetInterface.set) MapInterface.dict =
  (* all functions in the program *)
  let cfgs = g.Dom.graph.M5graph.cfgs in
  let all_user_pids = 
    PMap.foldi (fun p _ acc -> PSet.add p acc) cfgs PSet.empty
  in

  (* filter library call related edges *)
  let (node2pids, _) = pre.Pre.callnode in
  let (node2user_pids:(M5graph.node, Syn.pid PSet.t) PMap.t) = 
    PMap.map 
      (fun pids -> 
        PSet.map Pre.CallNode.nameof
          (PSet.filter Pre.CallNode.is_user_pid pids)
      ) 
      node2pids 
  in

  (* make callfroms in the airac world 
     
     1. make initial callfroms : pid -> empty set 
     
     2. update the initial callfroms with the call nodes
     (node2user_pids) *)
  let init_callfroms = 
    PSet.fold 
      (fun pid acc -> PMap.add pid PSet.empty acc) 
      all_user_pids PMap.empty
  in
  let (callfroms:(Syn.pid,(M5graph.node PSet.t))PMap.t) = 
    PMap.foldi
      (fun (node:M5graph.node) (pids:Syn.pid PSet.t) 
        (acc:(Syn.pid,(M5graph.node PSet.t))PMap.t) ->
          PSet.fold
            (fun pid acc -> 
              let nodes = PMap.find pid acc in
              let nodes = PSet.add node nodes in
              PMap.add pid nodes acc)
            pids acc)
      node2user_pids init_callfroms
  in

  (* translate callfroms *)
  let translated_callfroms = 
    PMap.foldi
      (fun pid nodes acc ->
        let translated_nodes = 
          PSet.fold 
            (fun node acc -> 
              Orders.intintSetAdd
                (tr_m5graph_node_simpl str2pos node) 
                acc
            )
            nodes 
            Orders.intintSetEmpty
        in
        Orders.intMapAdd (PMap.find pid str2pos) translated_nodes acc
      )
      callfroms 
      Orders.intMapEmpty
  in
  translated_callfroms

let tr_callfroms_star 
    (str2pos: (string, int) PMap.t)
    (g:Dom.global) (pre: Pre.preinfo)
    : (DomBasic.fidT, DomBasic.fidT SetInterface.set) MapInterface.dict =

  (* all functions in the program *)
  let cfgs = g.Dom.graph.M5graph.cfgs in
  let all_user_pids = 
    PMap.foldi (fun p _ acc -> PSet.add p acc) cfgs PSet.empty
  in
  
  (* make callfroms_star in the airac world *)
  let callfroms_star = 
    PSet.fold 
      (fun callee acc -> 
        let transitive_callers = 
          Pre.CallGraph.find_transitive_callers callee pre.Pre.callgraph
        in
        PMap.add callee transitive_callers acc
      )
      all_user_pids PMap.empty
  in

  (* translate callfroms_star *)
  let translated_callfroms_star = 
    PMap.foldi
      (fun callee callers acc ->
        let translated_callers = 
          PSet.fold
            (fun caller acc -> Orders.intSetAdd (PMap.find caller str2pos) acc)
            callers 
            Orders.intSetEmpty
        in
	Orders.intMapAdd (PMap.find callee str2pos) translated_callers acc
      )
      callfroms_star
      Orders.intMapEmpty
  in
  translated_callfroms_star

let tr_acclocs
    (locals: (Syn.vid,Syn.fid) PMap.t) (str2pos: (string, int) PMap.t)
    (g:Dom.global) (pre: Pre.preinfo)
    : (DomBasic.fidT, DomAbsValue.locT SetInterface.set) MapInterface.dict =
  (* all functions in the program *)
  let cfgs = g.Dom.graph.M5graph.cfgs in
  let all_user_pids = 
    PMap.foldi (fun p _ acc -> PSet.add p acc) cfgs PSet.empty
  in
  
  (* initial acclocs : pid -> empty set *)
  let init_acclocs = 
    PSet.fold 
      (fun pid acc -> PMap.add pid PSet.empty acc) 
      all_user_pids PMap.empty
  in

  (* make acclocs in the airac world 

     CAUTION: remove null pointer *)
  let access = pre.Pre.access_for_du in
  let acclocs = 
    PMap.foldi
      (fun node location acc ->
        let pid = M5graph.ctxof node in
        let addr1_set = PMap.find pid acc in
        let addr1_set = 
	      PSet.union 
	        addr1_set 
	        (PSet.filter 
	           (fun x -> not (x = Dom.Addr1.Null))
	           location.Pre.Location.access
	        )
	    in
        PMap.add pid addr1_set acc
      )
      access init_acclocs
  in
  
  (* add function arguments to acclocs *)
  let (acclocs : (Syn.pid, Dom.Addr1.t PSet.t) PMap.t) = 
    PMap.mapi
      (fun pid pid_locs ->
        let cfg = PMap.find pid cfgs in
        let params_list = cfg.Cfg.params in
        let params_set = Vocabularies.list2set params_list in
        let params_addr1_set = PSet.map 
          (fun x -> Dom.Addr1.variable x)
          params_set
        in
        PSet.union pid_locs params_addr1_set
      )
      acclocs
  in

  (* add local variables to acclocs *)
  (*
  let acclocs = 
    PMap.mapi
      (fun pid addr1_set ->
        let local_var_set = 
          PMap.foldi
            (fun x f acc ->
              if f = pid 
              then PSet.add (Dom.Addr1.Variable x) acc
              else acc
            )
          locals PSet.empty
        in
        PSet.union addr1_set local_var_set
      )
      acclocs 
  in
  *)

  (* propagate acclocs from transitive callees to caller for
     validation *)
  let (acclocs: (Syn.pid, Dom.Addr1.t PSet.t) PMap.t) = 
    PMap.mapi
      (fun pid orig_acclocs ->
        let reachables = 
          Pre.CallGraph.find_transitive_callees pid pre.Pre.callgraph
        in
        let (reachable_acclocs:Dom.Addr1.t PSet.t) = 
          PSet.fold 
            (fun callee acc -> PSet.union acc (PMap.find callee acclocs))
            reachables orig_acclocs
        in
        reachable_acclocs
      )
      acclocs
  in

  (* translate acclocs *)
  let translated_acclocs = 
    PMap.foldi
      (fun pid addr1_set acc ->
        let translated_addr1_set = 
          PSet.fold 
            (fun addr1 acc -> 
	          Orders.locTSetAdd (tr_addr1 locals str2pos addr1) acc
            )
            addr1_set
            Orders.locTSetEmpty
        in

        Orders.intMapAdd (PMap.find pid str2pos) translated_addr1_set acc
      )
      acclocs
      Orders.intMapEmpty
  in

  translated_acclocs

let tr_access
    ((locals: (Syn.vid,Syn.fid) PMap.t), (str2pos: (string, int) PMap.t),
     (g:Dom.global), (pre: Pre.preinfo))
    : DomAcc.accT = 
  { DomAcc.callfroms      = tr_callfroms str2pos g pre;
    DomAcc.callfroms_star = tr_callfroms_star str2pos g pre;
    DomAcc.acclocs        = tr_acclocs locals str2pos g pre }
